{
  RealThinClient W/A Framework: Server-side components interaction
  Copyright (c) Danijel Tkalcec

  @exclude
}
unit rtcBase;

interface

{$include rtcDefs.inc}

uses
  rtcTrashcan,
  Types, Classes, Graphics, Windows, Messages, SysUtils,
  {$IFNDEF NoForms} Controls {$ELSE} _Controls {$ENDIF},
  rtcSyncObjs, rtcThrPool, rtcInfo, ExtCtrls,

  rtcLog, rtcLink, memObjList,

  rtcGraphics, rtcCache;

const
  RtcFormControlStyle = [csAcceptsControls, csCaptureMouse, csClickEvents, csSetCaption, csDoubleClicks {$IFDEF IDE_7up}, csParentBackground{$ENDIF}];

type
  TRtcBaseForm=class; // forward
  TRtcBaseDataModule=class; // forward
  TRtcBaseApplication=class; // forward

  TRtcBaseFormClass=class of TRtcBaseForm;
  TRtcBaseDataModuleClass=class of TRtcBaseDataModule;

  TRtcBaseEvent=procedure of object;
  TRtcExceptionEvent=procedure(const E:Exception) of object;

  IrtcForm=interface(IrtcfComponent)
    ['{137676FA-5D38-493E-A3BA-5409CCC53FFB}']
    function New_Object(obj:IrtcfComponent):longint;
    function Get_Object(id:longint):IrtcfComponent;
    procedure Old_Object(id:longint);

    function _ID:longint;
    function _App:TRtcBaseApplication;
    function _Owner:TObject;
    procedure _Detach;
    end;

  IrtcMainMenu=interface(IrtcfComponent)
    ['{E158C65B-D1C2-4AC5-96E3-FD2B8387A5D7}']
    procedure Refresh;
    procedure UpdateNow;
    end;

  IrtcPopupMenu=interface(IrtcfComponent)
    ['{72AF3C4D-BD93-496F-B46B-40C3474B15CE}']
    procedure Popup(X,Y:integer);
    procedure Refresh;
    procedure UpdateNow;
    end;

  IrtcBasicAction=interface(IrtcfComponent)
    ['{68B3FC18-45F8-44B3-B7FA-D5333A92A06E}']
    procedure UpdateNow;
    end;

  IrtcWinControl=interface(IrtcfComponent)
    ['{60EBED89-8955-4ABD-86F7-2C686DF2B31B}']
    procedure UpdateNow;
    end;

  TRtcBaseApplicationThread=class(TRtcThread)
  public
    App:TRtcBaseApplication;

    constructor Create; override;
    destructor Destroy; override;

    function Work(Job:TObject):boolean; override;
    procedure Kill(Job:TObject); override;
    end;

  TRtcExePackPtr=^TRtcExePack;
  TRtcExePack=record
    InData:TRtcArray;
    Next:TRtcExePackPtr;
    end;

  (* Server-side Application component, 1 instance needed per user *)
  TRtcBaseApplication=class(TComponent)
  private
    fCS:TRtcCritSec;
    fThr:TRtcBaseApplicationThread;

    fIdle:boolean;
    fPackFirst,
    fPackLast:TRtcExePackPtr;

    LastObjID:longint;
    fObjs:array of IrtcForm;

    // Data prepared for sending to the Client controlling this Application
    xformid,xcompid:longint;
    OutData:TRtcArray;

    // Properties and events pepared for sending
    PropData:TRtcRecord;
    // All other data prepared for sending
    SendData:TRtcArray;

    // Data prepared for local execution (received from Client)
    _loc,_fid,_cid:longint;
    _pname:string;
    _pid:longint;
    InData:TRtcArray;
    FRecv: TRtcReceiveEvent;
    FOnQuit: TNotifyEvent;
    FOnException: TRtcExceptionEvent;

    FExecuteCount:integer;
    FMaxUpdateDelay:integer;
    FUseClientCache:boolean;

    procedure PackSendData;

    procedure Prepare(Data: TRtcArray);
    function PrepareMore:boolean;

  protected
    MainFormID:longint;

    // Send data to the assigned Receiver (Client)
    procedure _Send(Data:TRtcValueObject); virtual;

    { Execute next line sent for execution.
      Return TRUE if there was something to execute,
      FALSE if all lines were already executed. }
    function Execute:boolean;

    // Sending order important, every call is sent out
    function Form_Create(frm:IrtcForm; const Class_Name:string):longint;
    procedure Form_Data(formID,compID:longint; const xName:string; xData:TRtcValueObject); overload;
    procedure Form_Data(formID,compID:longint; const xName:string; const xData:Variant); overload;
    procedure Form_Loaded(id:longint);
    procedure Form_Destroy(id:longint);
    // Sending order not important, new calls with same name overwrite old calls with that name
    procedure Form_Event(formID,compID:longint; const xName:string; xID:integer);

    procedure Form_Prop(formID,compID:longint; const xName:string; xData:TRtcValueObject); overload;
    procedure Form_Prop(formID,compID:longint; const xName:string; const xData:Variant); overload;

    function ToCache(xData:TRtcValueObject):TRtcValueObject; overload;
    function ToCache(const xData:Variant):TRtcValueObject; overload;

    function Get_Object(FormID,ID:longint):IrtcfComponent;

  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;

    // Make this Application the default application for the current thread.
    // All new Forms created within this Thread will be assigned to this Application.
    procedure MakeActive;

    // Add package "Data" to execution queue, so it will be executed from a background thread
    procedure EnterExecute;
    procedure ToExecute(const Data:TRtcValueObject);
    procedure LeaveExecute;

    procedure ExecuteNow(const Data:TRtcValueObject);

    function ExecuteEvent(ev:TRtcBaseEvent):boolean;
    function CreateForm(form:TRtcBaseFormClass; isMainForm:boolean=False):boolean;
    function CreateDataModule(module:TRtcBaseDataModuleClass):boolean;
    // Send "Quit" message to the application, which will free all forms and destroy the application object
    function Quit:boolean; virtual;

    // Send all prepared changes to the Client NOW!
    procedure UpdateNow;

    // Destroy all Forms assigned to this application
    procedure FreeAllForms(detach:boolean=False);

    procedure ProcessMessages;

    // Read access to the application Thread (can be used by components)
    property Thread:TRtcBaseApplicationThread read fThr;

    // Maximum number of updates that can be kept in memory (delayed sending) before they have to be sent to the client
    property MaxUpdateDelay:integer read FMaxUpdateDelay write FMaxUpdateDelay;

    // Use Caching mechanisms to reduce traffic to the Client?
    property UseClientCache:boolean read FUseClientCache write FUSeClientCache;

    // Event called when data is prepared for sending
    property OnSend:TRtcReceiveEvent read FRecv write FRecv;

    // OnQuit will be triggered when application has terminated
    property OnQuit:TNotifyEvent read FOnQuit write FOnQuit;
    // OnException will be triggered when an exception was raised in server-side code
    property OnException:TRtcExceptionEvent read FOnException write FOnException;
    end;

  (* Central class used by all server-side RTC Forms components *)
  TRtc_Com=class(TRtcCom)
  private
    FID:longint;
    FFormID:longint;
    FForm:IrtcForm;
    FApp:TRtcBaseApplication;

    FMe:TControl;
    FSelf:TComponent;
    FView:IrtcfComponentX;
    Bnds:TRect;
    HaveBnds:boolean;
    isControl:boolean;

    fEvents:array of TRtcExecuteEvent;
    fEventNames:array of string;
    fEventCount:integer;

    function Get_Preview: boolean;
    procedure Set_Preview(const Value: boolean);

  protected
    // Call event stored under "xID" with parameters "xData"
    procedure Call_Event(xID:integer; xData:TRtcValue);

    // Detach component from Form and Application components
    procedure Detach;

    procedure CheckMe(const xName:string);
    procedure UpdateMe;

  public
    // Call in RTC Forms Component constructor to prepare the component for remote use
    constructor Create(myOwner:TObject; me:TComponent; class_name:string);

    // Call in RTC Forms Component destructor to remove the component from remote list
    destructor Destroy; override;

    { Calls where order is important and every call is recorded }
    // Update components property or call components method by name
    procedure Send(const xName: string; xData: TRtcValueObject); overload; override;
    procedure Send(const xName: string; const xData:Variant); overload; override;
    // Link component to another component
    procedure Link(const xName: string; xComp: TComponent); override;

    { Calls where order is not important and calls with same name
      will override old calls with that name }
    // Set event accessible remotely (xMethod = NIL to remove)
    procedure Event(const xName: string; xMethod:TRtcExecuteEvent); override;
    // Set property on remote component
    procedure Prop(const xName:string; xData:TRtcValueObject); overload; override;
    procedure Prop(const xName:string; const xData:Variant); overload; override;

    // Set components Name for remote use by name
    procedure Rename(const NewName: TComponentName);

    function Cache(xData:TRtcValueObject):TRtcValueObject; overload; override;
    function Cache(const xData:Variant):TRtcValueObject; overload; override;

    function TabOrder:TTabOrder; override;

    function GetUID:TRtcRecord; override;
    function FindByUID(UID:TRtcRecord):TComponent; override;

    // Return component ID
    property ID:longint read FID;
    // Return (owner) Form ID
    property FormID:longint read FFormID;
    // Paint component on a DC
    procedure Paint(const Can:TCanvas);

    procedure UpdateNow; override;

    procedure Loaded;

    property Preview:boolean read Get_Preview write Set_Preview default False;
    end;

  // New created Forms will automatically be assigned to the default Application for the current thread
  TRtcFormBorderStyle = (bsNone, bsSingle, bsSizeable, bsDialog, bsToolWindow, bsSizeToolWin);
  TRtcFormStyle = (fsNormal, fsMDIChild, fsMDIForm, fsStayOnTop);
  TRtcBorderIcon = (biSystemMenu, biMinimize, biMaximize, biHelp);
  TRtcBorderIcons = set of TRtcBorderIcon;
  TRtcPosition = (poDesigned, poDefault, poDefaultPosOnly, poDefaultSizeOnly, poScreenCenter, poDesktopCenter, poMainFormCenter, poOwnerFormCenter);
  TRtcDefaultMonitor = (dmDesktop, dmPrimary, dmMainForm, dmActiveForm);
  TRtcCloseAction = (caNone, caHide, caFree, caMinimize);
  TRtcCloseEvent = procedure(Sender: TObject; var Action: TRtcCloseAction) of object;
  TRtcCloseQueryEvent = procedure(Sender: TObject; var CanClose: Boolean) of object;
  TRtcWindowState = (wsNormal, wsMinimized, wsMaximized);

  TRtcBaseForm=class(TWinControl, IrtcForm, IrtcfComponent)
  private
    LastObjID:longint;
    fObjs:array of IrtcfComponent;

    fID:longint;
    fApp:TRtcBaseApplication;

    FOnClose: TRtcCloseEvent;
    FOnCloseQuery: TRtcCloseQueryEvent;
    FOnCreate: TNotifyEvent;
    FOnDestroy: TNotifyEvent;
    FOnHide: TNotifyEvent;
    FOnShow: TNotifyEvent;

    FVisible:boolean;
    FLeft,FTop: integer;
    FKeyPreview: Boolean;
    FWindowState: TRtcWindowState;
    FTransparentColorValue: TColor;
    FTransparentColor: Boolean;
    FPosition: TRtcPosition;
    FFormStyle: TRtcFormStyle;
    FBorderIcons: TRtcBorderIcons;
    FBorderStyle: TRtcFormBorderStyle;
    FAlphaBlend: Boolean;
    FAlphaBlendValue: Byte;
    FScreenSnap: Boolean;
    FSnapBuffer: integer;
    FDefaultMonitor: TRtcDefaultMonitor;
    FIDE_Link: boolean;

    procedure PrepareCom;

    function Get_Left: integer;
    function Get_Top: integer;
    procedure Set_Left(const Value: integer);
    procedure Set_Top(const Value: integer);

    function Get_Caption: TCaption;
    procedure Set_Caption(const Value: TCaption);

    function Get_ClientHeight: integer;
    function Get_ClientWidth: integer;
    function Get_Height: integer;
    function Get_Width: integer;

    procedure Set_ClientHeight(const Value: integer);
    procedure Set_ClientWidth(const Value: integer);
    procedure Set_Height(const Value: integer);
    procedure Set_Width(const Value: integer);

    function Get_Preview: boolean;
    procedure Set_Preview(const Value: boolean);

    function Get_Visible: boolean;
    procedure Set_Visible(const Value: boolean);
    function Get_isMain: boolean;
    procedure Set_isMain(const Value: boolean);

    function Get_Font: TRtcFont;
    procedure Set_Font(const Value: TRtcFont);
    procedure Set_IDELink(const Value: boolean);

  protected
    FCom:TRtc_Com;
    FFnt: TRtcFont;

    function New_Object(obj:IrtcfComponent):longint;
    function Get_Object(id:longint):IrtcfComponent;
    procedure Old_Object(id:longint);

    procedure _Process(const xName:string; xData:TRtcValue); virtual;
    function _ClassName:string;

    function _ID:longint;
    function _Owner:TObject;
    function _Self:TComponent;
    function _App:TRtcBaseApplication;
    procedure _Detach;
    function _Com:TRtcCom;

    procedure MyFontChange(Sender:TObject);

  public
    constructor Create(AOwner:TComponent); override;
    constructor CreateNew(AOwner:TComponent; Dummy:integer=0); virtual;

    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;

    destructor Destroy; override;
    procedure BeforeDestruction; override;

    procedure Hide;
    procedure Show;
    procedure Close;
    procedure UpdateNow; virtual;

    procedure SetFocus; override;
    procedure BringToFront;
    procedure SendToBack;

    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;

    property isMain:boolean read Get_isMain write Set_isMain;

    property Caption:TCaption read Get_Caption write Set_Caption;

    property DefaultMonitor: TRtcDefaultMonitor read FDefaultMonitor write FDefaultMonitor;
    property ScreenSnap: Boolean read FScreenSnap write FScreenSnap;
    property SnapBuffer: integer read FSnapBuffer write FSnapBuffer;
    property AlphaBlend: Boolean read FAlphaBlend write FAlphaBlend;
    property AlphaBlendValue: Byte read FAlphaBlendValue write FAlphaBlendValue;
    property KeyPreview: Boolean read FKeyPreview write FKeyPreview default False;
    property WindowState: TRtcWindowState read FWindowState write FWindowState default wsNormal;
    property TransparentColorValue: TColor read FTransparentColorValue write FTransparentColorValue default clBlack;
    property TransparentColor: Boolean read FTransparentColor write FTransparentColor default False;
    property Position: TRtcPosition read FPosition write FPosition default poDesigned;
    property FormStyle: TRtcFormStyle read FFormStyle write FFormStyle default fsNormal;
    property BorderStyle: TRtcFormBorderStyle read FBorderStyle write FBorderStyle default bsSizeable;
    property BorderIcons: TRtcBorderIcons read FBorderIcons write FBorderIcons default [biSystemMenu, biMinimize, biMaximize];

    property Visible:boolean read Get_Visible write Set_Visible default False;

  published
    property Preview: boolean read Get_Preview write Set_Preview stored False;
    property IDE_Link: boolean read FIDE_Link write Set_IDELink stored False;

    property Font: TRtcFont read Get_Font write Set_Font;

    property Left:integer read Get_Left write Set_Left default 0;
    property Top:integer read Get_Top write Set_Top default 0;
    property Width:integer read Get_Width write Set_Width;
    property Height:integer read Get_Height write Set_Height;
    property ClientWidth:integer read Get_ClientWidth write Set_ClientWidth;
    property ClientHeight:integer read Get_ClientHeight write Set_ClientHeight;

    property OnCloseQuery:TRtcCloseQueryEvent read FOnCloseQuery write FOnCloseQuery;
    property OnClose:TRtcCloseEvent read FOnClose write FOnClose;
    property OnCreate:TNotifyEvent read FOnCreate write FOnCreate;
    property OnDestroy:TNotifyEvent read FOnDestroy write FOnDestroy;
    property OnHide:TNotifyEvent read FOnHide write FOnHide;
    property OnShow:TNotifyEvent read FOnShow write FOnShow;
    end;

  // New created DataModules will automatically be assigned to the default Application for the current thread
  TRtcBaseDataModule=class(TDataModule, IrtcForm, IrtcfComponent)
  private
    LastObjID:longint;
    fObjs:array of IrtcfComponent;

    fID:longint;
    fApp:TRtcBaseApplication;

    procedure PrepareCom;

  protected
    FCom:TRtc_Com;

    function New_Object(obj:IrtcfComponent):longint;
    function Get_Object(id:longint):IrtcfComponent;
    procedure Old_Object(id:longint);

    procedure _Process(const xName:string; xData:TRtcValue); virtual; abstract;
    function _ClassName:string;

    function _ID:longint;
    function _Owner:TObject;
    function _Self:TComponent;
    function _App:TRtcBaseApplication;
    procedure _Detach;
    function _Com:TRtcCom;

    procedure Loaded; override;

  public
    constructor CreateNew(AOwner:TComponent; Dummy:integer=0); override;
    destructor Destroy; override;

    procedure UpdateNow; virtual;
    end;

  TRtcBaseComponent=class(TComponent, IrtcfComponent)
  private
    FCom:TRtc_Com;
  protected
    procedure _Process(const xName:string; xData:TRtcValue); virtual;
    function _Com:TRtcCom;
    function _ClassName:string; virtual; abstract;
    function _Self:TComponent;

    procedure SetName(const NewName: TComponentName); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure UpdateNow; virtual;
    end;

  TRtcBaseControl = class(TControl, IrtcfComponent)
  private
    FCom:TRtc_Com;
    FCan: TCanvas;

    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    function Get_Can: TCanvas;
    function Get_Text: TCaption;
    procedure Set_Text(const Value: TCaption);

  protected
    FText: TCaption;

    procedure SetParent(AParent: TWinControl); override;
    procedure SetName(const NewName: TComponentName); override;
    procedure Paint; virtual;
    property Can: TCanvas read Get_Can;

    function IsCaptionStored: Boolean; virtual;

    procedure _Process(const xName:string; xData:TRtcValue); virtual;
    function _Com:TRtcCom;
    function _ClassName:string; virtual; abstract;
    function _Self:TComponent;

  protected
    property Caption: TCaption read Get_Text write Set_Text stored IsCaptionStored;
    property Text: TCaption read Get_Text write Set_Text;

  public
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure UpdateNow; virtual;
  end;

  TRtcBaseWinControl = class(TCustomControl, IrtcfComponent, IrtcWinControl)
  private
    FCom:TRtc_Com;

    function Get_Can: TCanvas;
    function Get_Text: TCaption;
    procedure Set_Text(const Value: TCaption);

  protected
    FText: TCaption;

    procedure SetParent(AParent: TWinControl); override;
    procedure SetName(const NewName: TComponentName); override;
    procedure Paint; override;
    property Can: TCanvas read Get_Can;

    function IsCaptionStored: Boolean; virtual;

    procedure _Process(const xName:string; xData:TRtcValue); virtual;
    function _Com:TRtcCom;
    function _ClassName:string; virtual; abstract;
    function _Self:TComponent;

  protected
    property Caption: TCaption read Get_Text write Set_Text stored IsCaptionStored;
    property Text: TCaption read Get_Text write Set_Text;

  public
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure UpdateNow; virtual;
  end;

implementation

type
  TRtcBaseApplicationJob=class
    end;

var
  fThrApp:TObjList;
  fThrAppCS:TRtcCritSec;

  Message_AppRun,
  Message_AppQuit:TRtcBaseApplicationJob;

// Set the default Application for the current thread.
// This is needed so that newly created forms know to which application they belong
procedure SetApplication(App:TRtcBaseApplication);
  var
    tid:cardinal;
    oapp:TObject;
  begin
  fThrAppCS.Enter;
  try
    tid:=GetCurrentThreadId;
    oapp:=fThrApp.search(tid);
    if oapp<>App then
      begin
      if assigned(oapp) then
        fThrApp.remove(tid);
      if assigned(App) then
        fThrApp.insert(tid,App);
      end;
  finally
    fThrAppCS.Leave;
    end;
  end;

// Get the default Application for the current thread
function GetApplication:TRtcBaseApplication;
  var
    tid:cardinal;
    oapp:TObject;
  begin
  fThrAppCS.Enter;
  try
    tid:=GetCurrentThreadId;
    oapp:=fThrApp.search(tid);
    if assigned(oapp) then
      Result:=TRtcBaseApplication(oapp)
    else
      Result:=nil;
  finally
    fThrAppCS.Leave;
    end;
  end;

{ TRtcApplication }

constructor TRtcBaseApplication.Create(AOwner:TComponent);
  begin
  inherited;
  fCS:=TRtcCritSec.Create;

  MainFormID:=0;

  fThr:=TRtcBaseApplicationThread.Create;
  fThr.App:=self;

  fIdle:=True;
  fPackFirst:=nil;
  fPackLast:=nil;

  LastObjID:=0;
  SetLength(fObjs,0);

  xformid:=-1;
  xcompid:=-1;

  OutData:=nil;
  PropData:=nil;
  SendData:=nil;

  FExecuteCount:=0;

  MakeActive;
  end;

destructor TRtcBaseApplication.Destroy;
  var
    a:integer;
    io:IRtcForm;
    o,o2:TObject;
    t:TRtcBaseApplicationThread;
  begin
  if assigned(FOnQuit) then
    FOnQuit(self);

  if assigned(fThr) then
    begin
    t:=fThr; fThr:=nil;

    t.App:=nil;
    TRtcThread.Stop(t);
    end;

  OnSend:=nil;

  if GetApplication=self then
    SetApplication(nil);

    if assigned(fObjs) then
    for a:=length(fObjs)-1 downto 0 do
      if assigned(fObjs[a]) then
        begin
        io:=fObjs[a];
        fObjs[a]:=nil;
        o:=io._Owner;
        o2:=io._Self;
        io._Detach;
        io:=nil;
        try
          if assigned(o2) then FreeAndNil(o2);
          if assigned(o) then FreeAndNil(o);
        except
          on E:Exception do
            Log('ERROR! '+E.Message);
          end;
        end;

  LastObjID:=0;
  SetLength(fObjs,0);

  // Clean up "InData" ...
  while PrepareMore do;

  if assigned(OutData) then FreeAndNil(OutData);
  PropData:=nil;
  SendData:=nil;
  FreeAndNil(fCS);
  inherited;
  end;

procedure TRtcBaseApplication.FreeAllForms(detach:boolean=False);
  var
    a:integer;
    o,o2:TObject;
    io:IRtcForm;
  begin
  if assigned(fObjs) then
    for a:=length(fObjs)-1 downto 0 do
      if assigned(fObjs[a]) then
        begin
        io:=fObjs[a];
        fObjs[a]:=nil;
        o:=io._Owner;
        o2:=io._Self;
        if detach then io._Detach;
        io:=nil;
        try
          if assigned(o2) then FreeAndNil(o2);
          if assigned(o) then FreeAndNil(o);
        except
          on E:Exception do
            Log('ERROR! '+E.Message);
          end;
        end;

  LastObjID:=0;
  SetLength(fObjs,0);
  end;

function TRtcBaseApplication.Form_Create(frm: IrtcForm; const Class_Name:string): longint;
  begin
  Inc(LastObjID);
  Result:=LastObjID;

  Setlength(fObjs,LastObjID);
  fObjs[LastObjID-1]:=frm;

  Form_Data(LastObjID,0,'new',TRtcTextValue.Create(Class_Name));
  end;

procedure TRtcBaseApplication.Form_Destroy(id: Integer);
  begin
  fObjs[id-1]:=nil;
  Form_Data(id,0,'free',nil);
  end;

procedure TRtcBaseApplication.PackSendData;
  begin
  if assigned(SendData) then
    begin
    if SendData.Count=1 then
      begin
      with OutData.asRecord[OutData.Count-1] do
        begin
        asObject[IntToStr(xformid)]:=nil;
        asObject[IntToStr(xformid)]:=SendData.asObject[0];
        SendData.asObject[0]:=nil;
        SendData.Free;
        SendData:=nil;
        PropData:=nil;
        end;
      end
    else
      begin
      SendData:=nil;
      PropData:=nil;
      end;
    end;
  end;

procedure TRtcBaseApplication.Form_Loaded(id: Integer);
  var
    s:TRtcValueObject;
  begin
  Form_Data(id,0,'ok',nil);
  if FUseClientCache then
    begin
    PackSendData;
    s:=OutData.asRecord[OutData.Count-1].asObject[IntToStr(xformid)];
    with OutData.asRecord[OutData.Count-1] do
      begin
      asObject[IntToStr(xformid)]:=nil;
      asObject[IntToStr(xformid)]:=ToCache(s);
      end;
    xformid:=-1;
    xcompid:=-1;
    end;
  end;

procedure TRtcBaseApplication.Form_Data(formid, compid: Integer; const xName: string; xData: TRtcValueObject);
  begin
  if not assigned(OutData) then
    OutData:=TRtcArray.Create;

  PropData:=nil;

  if (formid<>xformid) then
    begin
    PackSendData;
    SendData:=OutData.newRecord(OutData.Count).newArray(IntToStr(formid));
    xformid:=formid;
    xcompid:=0;
    end;

  with SendData.newRecord(SendData.Count) do
    begin
    if assigned(xData) then
      asObject[xName]:=xData
    else
      asBoolean[xName]:=False;
    if compid<>xcompid then
      begin
      asInteger['.']:=compid;
      xcompid:=compid;
      end;
    end;
  end;

procedure TRtcBaseApplication.Form_Data(formID, compID: Integer; const xName: string; const xData: Variant);
  begin
  if not assigned(OutData) then
    OutData:=TRtcArray.Create;

  PropData:=nil;

  if (formid<>xformid) then
    begin
    PackSendData;
    SendData:=OutData.newRecord(OutData.Count).newArray(IntToStr(formid));
    xformid:=formid;
    xcompid:=0;
    end;

  with SendData.NewRecord(SendData.Count) do
    begin
    asValue[xName]:=xData;
    if compid<>xcompid then
      begin
      asInteger['.']:=compid;
      xcompid:=compid;
      end;
    end;
  end;

procedure TRtcBaseApplication.Form_Prop(formID, compID: Integer; const xName: string; xData: TRtcValueObject);
  var
    cid:string;
  begin
  if not assigned(OutData) then
    OutData:=TRtcArray.Create;

  if (formid<>xformid) then
    begin
    PackSendData;
    SendData:=OutData.newRecord(OutData.Count).newArray(IntToStr(formid));
    xformid:=formid;
    xcompid:=0;
    end;

  if not assigned(PropData) then
    PropData:=SendData.NewRecord(SendData.Count);

  cid:=IntToStr(compID);
  if PropData.isNull[cid] then
    PropData.NewRecord(cid)
  else
    PropData.asRecord[cid].isNull[xName]:=True;

  if xData=nil then
    PropData.asRecord[cid].asBoolean[xName]:=False
  else
    PropData.asRecord[cid].asObject[xName]:=xData;
  end;

procedure TRtcBaseApplication.Form_Prop(formID, compID: Integer; const xName: string; const xData:Variant);
  var
    cid:string;
  begin
  if not assigned(OutData) then
    OutData:=TRtcArray.Create;

  if (formid<>xformid) then
    begin
    PackSendData;
    SendData:=OutData.newRecord(OutData.Count).newArray(IntToStr(formid));
    xformid:=formid;
    xcompid:=0;
    end;

  if not assigned(PropData) then
    PropData:=SendData.NewRecord(SendData.Count);

  cid:=IntToStr(compID);
  if PropData.isNull[cid] then
    PropData.NewRecord(cid)
  else
    PropData.asRecord[cid].isNull[xName]:=True;

  PropData.asRecord[cid].asValue[xName]:=xData;
  end;

function TRtcBaseApplication.ToCache(xData: TRtcValueObject): TRtcValueObject;
  var
    id:String;
    s:AnsiString;
  begin
  if FUseClientCache then
    begin
    s:=xData.toCode;
    if length(s)<RTC_CACHE_MINSIZE then
      Result:=xData
    else
      begin
      xData.Free;
      id:=rtcCache_Hash(s);
      if rtcCache_Write(s,id) then
        Result:=TRtcVariableName.Create(id)
      else
        raise Exception.Create('Error writing to cache');
      end;
    end
  else
    Result:=xData;
  end;

function TRtcBaseApplication.ToCache(const xData: Variant): TRtcValueObject;
  var
    x:TRtcValue;
    id:string;
    s:AnsiString;
  begin
  if FUseClientCache then
    begin
    x:=TRtcValue.Create;
    x.asValue:=xData;
    s:=x.toCode;
    if length(s)<RTC_CACHE_MINSIZE then
      Result:=x
    else
      begin
      x.Free;
      id:=rtcCache_Hash(s);
      if rtcCache_Write(s,String(id)) then
        Result:=TRtcVariableName.Create(id)
      else
        raise Exception.Create('Error writing to cache');
      end;
    end
  else
    begin
    x:=TRtcValue.Create;
    x.asValue:=xData;
    Result:=x;
    end;
  end;

procedure TRtcBaseApplication.Form_Event(formID, compID: Integer; const xName: string; xID: longint);
  begin
  Form_Prop(formID,compID,'.'+IntToStr(xID),xName);
  end;

procedure TRtcBaseApplication.MakeActive;
  begin
  SetApplication(self);
  end;

procedure TRtcBaseApplication.UpdateNow;
  var
    s:TRtcValueObject;
  begin
  if assigned(OutData) then
    begin
    PackSendData;
    // Send all prepared Data to the Client
    if OutData.Count=1 then
      begin
      s:=OutData.asObject[0];
      OutData.asObject[0]:=nil;
      OutData.Free;
      OutData:=nil;
      _Send(s);
      end
    else
      begin
      _Send(OutData);
      OutData:=nil;
      end;

    SendData:=nil;
    PropData:=nil;
    xcompid:=-1;
    xformid:=-1;
    end;
  end;

procedure TRtcBaseApplication.ToExecute(const Data: TRtcValueObject);
  var
    fTemp:TRtcExePackPtr;
    fData:TRtcArray;
  begin
  if not (Data is TRtcArray) then
    begin
    fData:=TRtcArray.Create;
    fData.asObject[0]:=Data;
    end
  else
    fData:=TRtcArray(Data);

  New(fTemp);
  fTemp^.Next:=nil;
  fTemp^.InData:=fData;
  if assigned(fPackLast) then
    fPackLast^.Next:=fTemp
  else
    fPackFirst:=fTemp;
  fPackLast:=fTemp;
  end;

procedure TRtcBaseApplication.EnterExecute;
  begin
  fCS.Enter;
  end;

procedure TRtcBaseApplication.LeaveExecute;
  begin
  try
    if fIdle then // not working, post job
      begin
      fIdle:=False; // first job, leaving IDLE mode.
      TRtcThread.PostJob(fThr, Message_AppRun);
      end;
  finally
    fCS.Leave;
    end;
  end;

procedure TRtcBaseApplication.ExecuteNow(const Data: TRtcValueObject);
  begin
  EnterExecute;
  try
    ToExecute(Data);
  finally
    LeaveExecute;
    end;
  end;

function TRtcBaseApplication.PrepareMore: boolean;
  var
    fTemp:TRtcExePackPtr;
  begin
  fCS.Enter;
  try
    fTemp:=fPackFirst;
    if not assigned(fTemp) then
      begin
      Result:=False;
      Prepare(nil);
      fIdle:=True; // nothing more to do, going to IDLE mode.
      end
    else
      begin
      Result:=True;
      Prepare(fTemp.InData);

      fPackFirst:=fTemp^.Next;
      if fPackFirst=nil then fPackLast:=nil;
      Dispose(fTemp);
      end;
  finally
    fCS.Leave;
    end;
  end;

procedure TRtcBaseApplication.Prepare(Data: TRtcArray);
  begin
  if assigned(InData) then InData.Free;
  InData:=Data;
  _loc:=0;
  _fid:=-1;
  _cid:=-1;
  _pname:='';
  _pid:=-1;
  end;

function TRtcBaseApplication.Execute:boolean;
  var
    r:TRtcRecord;
  procedure ExecRec(rec:TRtcRecord);
    var
      a:integer;
      fn:string;
      Val:TRtcValue;
      haveEvent,
      haveProp:boolean;
    procedure ExecEvent(fid,cid,pid:longint; val:TRtcValue);
      var
        myob:IrtcfComponent;
        myform:IrtcForm;
        c:TRtcCom;
      begin
      if (fid>0) and (LastObjID>=fid) then
        if assigned(fObjs[fid-1]) then
          begin
          myform:=fObjs[fid-1];
          myob:=myform.Get_Object(cid);
          myform:=nil;
          if assigned(myob) then
            begin
            c:=myob._Com;
            myob:=nil;
            if assigned(c) then
              TRtc_Com(c).Call_Event(pid,Val);
            end;
          end;
      end;
    procedure ExecProp(fid,cid:longint;const pname:string; val:TRtcValue);
      var
        myob:IrtcfComponent;
      begin
      if (fid>0) and (LastObjID>=fid) then
        if assigned(fObjs[fid-1]) then
          begin
          myob:=fObjs[fid-1].Get_Object(cid);
          if assigned(myob) then
            myob._Process(pname,Val);
          end;
      end;
    procedure UnwrapProp(rec:TRtcRecord);
      var
        a,b:integer;
        po,fid,cid,pid:longint;
        fn:string;
        val:TRtcValue;
        rr:TRtcRecord;
      begin
      val:=nil;
      try
        for a:=0 to rec.Count-1 do
          begin
          fn:=rec.FieldName[a];
          po:=Pos('.',fn);
          if (rec.isType[fn]=rtc_Record) and (po>0) then
            begin
            fid:=StrToIntDef(Copy(fn,1,po-1),0);
            cid:=StrToIntDef(Copy(fn,po+1,length(fn)-po),0);
            rr:=rec.asRecord[fn];
            for b:=0 to rr.Count-1 do
              begin
              fn:=rr.FieldName[b];
              if (fn[1]>='1') and (fn[1]<='9') then // Event call
                begin
                if assigned(Val) then
                  Val.Clear
                else
                  Val:=TRtcValue.Create;
                pid:=StrToIntDef(fn,0);
                if not rr.isNull[fn] then
                  begin
                  Val.asObject:=rr.asObject[fn];
                  rr.asObject[fn]:=nil;
                  end;
                ExecEvent(fid,cid,pid,val);
                end
              else
                begin
                if assigned(Val) then
                  Val.Clear
                else
                  Val:=TRtcValue.Create;
                if not rr.isNull[fn] then
                  begin
                  Val.asObject:=rr.asObject[fn];
                  rr.asObject[fn]:=nil;
                  end;
                ExecProp(fid,cid,fn,val);
                end;
              end;
            end;
          end;
      finally
        if assigned(val) then val.Free;
        end;
      end;
    procedure UnwrapCache(rec:TRtcArray);
      var
        a:integer;
        fn:string;
        s:AnsiString;
        CacheData:TRtcArray;
      begin
      // Send out anything waiting in cache
      UpdateNow;
      // Prepare our new response with cache data
      CacheData:=TRtcArray.Create;
      try
        for a:=0 to rec.Count-1 do
          if rec.isType[a]=rtc_Variable then
            begin
            fn:=rec.asVarName[a];
            s:=rtcCache_Read(fn);
            CacheData.asVarName[CacheData.Count]:=fn;
            CacheData.asString[CacheData.Count]:=s;
            end;
      except
        CacheData.Free;
        raise;
        end;
      // Send out cache data response ...
      _Send(CacheData);
      end;
    begin
    Val:=nil;
    try
      haveEvent:=False;
      haveProp:=False;
      for a:=0 to rec.Count-1 do
        begin
        fn:=rec.FieldName[a];
        if fn='!' then
          begin
          if rec.isType[fn]=rtc_Array then
            UnwrapCache(rec.asArray[fn]);
          end
        else if fn='*' then
          begin
          if rec.isType[fn]=rtc_Record then
            UnwrapProp(rec.asRecord[fn]);
          end
        else if (fn[1]>='0') and (fn[1]<='9') then
          begin
          _fid:=StrToIntDef(fn,0);
          _cid:=rec.asInteger[fn];
          end
        else if (fn[1]='.') and (length(fn)>1) and (fn[2]>='1') and (fn[2]<='9') then // Event call
          begin
          haveEvent:=True;
          if assigned(Val) then
            Val.Clear
          else
            Val:=TRtcValue.Create;
          _pid:=StrToIntDef(Copy(fn,2,length(fn)-1),0);
          if not rec.isNull[fn] then
            begin
            Val.asObject:=rec.asObject[fn];
            rec.asObject[fn]:=nil;
            end;
          end
        else
          begin
          haveProp:=True;
          if assigned(Val) then
            Val.Clear
          else
            Val:=TRtcValue.Create;
          _pname:=fn;
          if not rec.isNull[fn] then
            begin
            Val.asObject:=rec.asObject[fn];
            rec.asObject[fn]:=nil;
            end;
          end;
        end;
      if haveEvent then
        ExecEvent(_fid,_cid,_pid,Val)
      else if haveProp then
        ExecProp(_fid,_cid,_pname,Val);
    finally
      if assigned(Val) then Val.Free;
      end;
    end;
  begin
  Result:=False;
  if not assigned(InData) then // nothing prepared
    PrepareMore
  else if _loc>=InData.Count then // last prepared line executed
    PrepareMore;

  if assigned(InData) and (_loc<InData.Count) then // data waiting for execution
    begin
    Inc(FExecuteCount);
    Result:=True;
    Inc(_loc);
    if InData.isType[_loc-1]=rtc_Record then
      begin
      r:=InData.asRecord[_loc-1];
      InData.Extract(_loc-1);
      try
        try
          ExecRec(r);
        finally
          r.Free;
          end;
      except
        on E:Exception do
          begin
          if assigned(FOnException) then
            try
              FOnException(E);
              UpdateNow;
            except
            end;
          end;
        end;
      end;
    if FExecuteCount>FMaxUpdateDelay then
      begin
      FExecuteCount:=0;
      UpdateNow;
      end;
    end
  else
    begin
    FExecuteCount:=0;
    UpdateNow;
    end;
  end;

procedure TRtcBaseApplication.ProcessMessages;
  begin
  UpdateNow;
  while Execute do;
  end;

procedure TRtcBaseApplication._Send(Data: TRtcValueObject);
  begin
  if assigned(FRecv) then
    FRecv(Data)
  else
    Data.Free;
  end;

function TRtcBaseApplication.Quit:boolean;
  begin
  Result:=TRtcThread.PostJob(fThr, Message_AppQuit);
  end;

type
  TRtcBaseAppCreateModuleJob=class(TRtcJob)
  public
    App:TRtcBaseApplication;
    Form:TComponentClass;
    isMain:boolean;
    procedure Run(Thr:TRtcThread); override;
    procedure Kill; override;
    end;

function TRtcBaseApplication.CreateForm(form: TRtcBaseFormClass; isMainForm:boolean=False):boolean;
  var
    job:TRtcBaseAppCreateModuleJob;
  begin
  job:=TRtcBaseAppCreateModuleJob.Create;
  job.App:=self;
  job.Form:=form;
  job.isMain:=isMainForm;
  Result:=TRtcThread.PostJob(fThr,job);
  if not Result then job.Free;
  end;

function TRtcBaseApplication.CreateDataModule(module: TRtcBaseDataModuleClass):boolean;
  var
    job:TRtcBaseAppCreateModuleJob;
  begin
  job:=TRtcBaseAppCreateModuleJob.Create;
  job.App:=self;
  job.Form:=module;
  job.isMain:=False;
  Result:=TRtcThread.PostJob(fThr,job);
  if not Result then job.Free;
  end;

function TRtcBaseApplication.Get_Object(FormID, ID: Integer): IrtcfComponent;
  begin
  if (FormID>0) and (LastObjID>=FormID) then
    begin
    if assigned(fObjs[FormID-1]) then
      Result:=fObjs[FormID-1].Get_Object(ID)
    else
      Result:=nil;
    end
  else
    Result:=nil;
  end;

{ TRtcBaseAppCreateModuleJob }

procedure TRtcBaseAppCreateModuleJob.Kill;
  begin
  App:=nil;
  Form:=nil;
  Free;
  end;

procedure TRtcBaseAppCreateModuleJob.Run(Thr: TRtcThread);
  var
    o:TComponent;
  begin
  App.MakeActive;
  try
    o:=form.Create(nil);
    if o is TRtcBaseForm then
      begin
      TRtcBaseForm(o).isMain:=isMain;
      TRtcBaseForm(o).Show;
      end;
  except
    on E:Exception do
      if assigned(App.OnException) then
        try
          App.OnException(E);
          App.UpdateNow;
        except
        end;
    end;
  App:=nil;
  Form:=nil;
  Free;
  end;

type
  TRtcBaseAppJob=class(TRtcJob)
  public
    App:TRtcBaseApplication;
    Ev:TRtcBaseEvent;
    procedure Run(Thr:TRtcThread); override;
    procedure Kill; override;
    end;

function TRtcBaseApplication.ExecuteEvent(ev:TRtcBaseEvent):boolean;
  var
    job:TRtcBaseAppJob;
  begin
  job:=TRtcBaseAppJob.Create;
  job.App:=self;
  job.Ev:=ev;
  Result:=TRtcThread.PostJob(fThr,job);
  if not Result then job.Free;
  end;

{ TRtcBaseAppCreateModuleJob }

procedure TRtcBaseAppJob.Kill;
  begin
  App:=nil;
  Ev:=nil;
  Free;
  end;

procedure TRtcBaseAppJob.Run(Thr: TRtcThread);
  begin
  App.MakeActive;
  try
    Ev;
  except
    on E:Exception do
      if assigned(App.OnException) then
        try
          App.OnException(E);
          App.UpdateNow;
        except
        end;
    end;
  App:=nil;
  Ev:=nil;
  Free;
  end;

{ TRtc_Com }

constructor TRtc_Com.Create(myOwner:TObject; me:TComponent; class_name:string);
  var
    a:integer;
    wc:TWinControl;
    co:TComponent;
    cc:TComponentClass;
  begin
  FID:=0;
  FFormID:=0;
  fApp:=nil;
  FForm:=nil;
  FView:=nil;
  FMe:=nil;
  FSelf:=me;

  if myOwner is TWinControl then
    begin
    wc:=TWinControl(myOwner);
    if Supports(wc,IrtcForm) then
      FForm:=(wc as IrtcForm)
    else
      begin
      for a:=0 to wc.ControlCount-1 do
        if Supports(wc.Controls[a],IrtcForm) then
          FForm:=(wc.Controls[a] as IrtcForm);
      end;
    end;

  if not assigned(FForm) and (myOwner is TComponent) then
    begin
    co:=TComponent(myOwner);
    if Supports(co,IrtcForm) then
      FForm:=(co as IrtcForm)
    else
      begin
      for a:=0 to co.ComponentCount-1 do
        if Supports(co.Components[a],IrtcForm) then
          FForm:=(co.Components[a] as IrtcForm);
      end;
    end;

  if not assigned(FForm) then
    raise Exception.Create('RTC components can only be used on RTC Forms and DataModules')
  else
    begin
    FFormID:=FForm._ID;
    FApp:=FForm._App;
    if assigned(FApp) and assigned(me) and (me<>myOwner) then
      begin
      FID:=FForm.New_Object(me as IrtcfComponent);
      FApp.Form_Data(FFormID,FID,'new',class_name);
      end;
    end;

  isControl:=assigned(me) and (me<>myOwner);

  if not assigned(FApp) then // Design-time
    begin
    cc:=TComponentClass(FindClass('TRtc'+class_name+'X'));
    if assigned(cc) then
      begin
      co:=cc.Create(nil);
      if Supports(co, IrtcfComponentX) then
        begin
        FView:=(co as IrtcfComponentX);
        FView._BeginLoad;
        if assigned(me) and (me is TControl) then
          FMe:=TControl(me);
        end
      else
        co.Free;
      end;
    end;
  end;

destructor TRtc_Com.Destroy;
  var
    ob:TObject;
  begin
  FMe:=nil;
  FSelf:=nil;
  if assigned(FView) then
    begin
    ob:=FView._Self;
    FView:=nil;
    ob.Free;
    end;
  if assigned(FForm) and assigned(FApp) and (FID>0) then
    begin
    FForm.Old_Object(FID);
    FApp.Form_Data(FFormID,FID,'free',nil);
    end;
  FForm:=nil;
  FApp:=nil;
  inherited;
  end;

procedure TRtc_Com.Send(const xName: string; xData: TRtcValueObject);
  procedure DoSend;
    var
      v:TRtcValue;
    begin
    v:=TRtcValue.Create;
    try
      v.asObject:=xData;
      xData:=nil;
      CheckMe(xName);
      FView._Process(xName,v);
      UpdateMe;
    finally
      v.Free;
      end;
    end;
  begin
  if assigned(FApp) then // Run-time
    FApp.Form_Data(FFormID, FID, xName, xData)
  else if assigned(FView) then // Design-time
    DoSend
  else if assigned(xData) then
    xData.Free;
  end;

procedure TRtc_Com.Send(const xName: string; const xData: Variant);
  procedure DoSend;
    var
      v:TRtcValue;
    begin
    v:=TRtcValue.Create;
    try
      v.asValue:=xData;
      CheckMe(xName);
      FView._Process(xName,v);
      UpdateMe;
    finally
      v.Free;
      end;
    end;
  begin
  if assigned(FApp) then // Run-time
    FApp.Form_Data(FFormID, FID, xName, xData)
  else if assigned(FView) then // Design-time
    DoSend;
  end;

procedure TRtc_Com.Prop(const xName: string; xData: TRtcValueObject);
  procedure DoSend;
    var
      v:TRtcValue;
    begin
    v:=TRtcValue.Create;
    try
      v.asObject:=xData;
      xData:=nil;
      CheckMe(xName);
      FView._Process(xName,v);
      UpdateMe;
    finally
      v.Free;
      end;
    end;
  begin
  if assigned(FApp) then // Run-time
    FApp.Form_Prop(FFormID, FID, xName, xData)
  else if assigned(FView) then // Design-time
    DoSend
  else if assigned(xData) then
    xData.Free;
  end;

procedure TRtc_Com.Prop(const xName: string; const xData: Variant);
  procedure DoSend;
    var
      v:TRtcValue;
    begin
    v:=TRtcValue.Create;
    try
      v.asValue:=xData;
      CheckMe(xName);
      if xName<>'visible' then
        FView._Process(xName,v);
      UpdateMe;
    finally
      v.Free;
      end;
    end;
  begin
  if assigned(FApp) then // Run-time
    FApp.Form_Prop(FFormID, FID, xName, xData)
  else if assigned(FView) then // Design-time
    DoSend;
  end;

function TRtc_Com.Cache(xData: TRtcValueObject): TRtcValueObject;
  begin
  if assigned(FApp) then // Run-time
    Result:=FApp.ToCache(xData)
  else
    Result:=xData;
  end;

function TRtc_Com.Cache(const xData: Variant): TRtcValueObject;
  function DoCache:TRtcValue;
    begin
    Result:=TRtcValue.Create;
    Result.asValue:=xData;
    end;
  begin
  if assigned(FApp) then
    Result:=FApp.ToCache(xData)
  else
    Result:=DoCache;
  end;

procedure TRtc_Com.Rename(const NewName: TComponentName);
  begin
  if assigned(FApp) then // Run-time
    FApp.Form_Data(FFormID, FID, 'name', NewName);
  end;

procedure TRtc_Com.Detach;
  begin
  FApp:=nil;
  FForm:=nil;
  FID:=0;
  FFormID:=0;
  end;

procedure TRtc_Com.Link(const xName:string; xComp:TComponent);
  var
    co:IrtcfComponent;
    rec:TRtcRecord;
  begin
  if assigned(FApp) then // Run-time
    begin
    if xComp=nil then
      Send('/'+xName,nil)
    else if Supports(xComp,IrtcfComponent) then
      begin
      co:=xComp as IrtcfComponent;
      if assigned(co._Com) then
        begin
        rec:=TRtcRecord.Create;
        if TRtc_Com(co._Com).FormID<>FFormID then
          rec.asInteger['f']:=TRtc_Com(co._Com).FormID;
        if TRtc_Com(co._Com).ID<>0 then
          rec.asInteger['c']:=TRtc_Com(co._Com).ID;
        Send('/'+xName,rec);
        end;
      end;
    end
  else if assigned(FView) then // Design-time
    begin
    CheckMe(xName);
    if xComp=nil then
      FView._Link(xName,nil)
    else if Supports(xComp,IrtcfComponent) then
      begin
      co:=(xComp as IrtcfComponent);
      if assigned(co._Com) and assigned(TRtc_Com(co._Com).FView) then
        FView._Link(xName,TRtc_Com(co._Com).FView._Self);
      end;
    UpdateMe;
    end;
  end;

procedure TRtc_Com.Call_Event(xID: integer; xData: TRtcValue);
  begin
  if (xID>0) and (xID<=fEventCount) then
    try
      fEvents[xID-1](xData);
    except
      on E:Exception do
        if assigned(FForm) then
          if assigned(FSelf) and (FID>0) then
            raise Exception.Create(FForm._Self.ClassName+'.'+FSelf.Name+'.'+fEventNames[xID-1]+' - '+E.ClassName+#13#10+E.Message)
          else
            raise Exception.Create(FForm._Self.ClassName+'.'+fEventNames[xID-1]+' - '+E.ClassName+#13#10+E.Message);
    end;
  end;

procedure TRtc_Com.Event(const xName: string; xMethod: TRtcExecuteEvent);
  begin
  if assigned(fApp) then // Run-time
    begin
    if assigned(xMethod) then
      begin
      Inc(fEventCount);
      SetLength(fEvents,fEventCount);
      SetLength(fEventNames,fEventCount);
      fEvents[fEventCount-1]:=xMethod;
      fEventNames[fEventCount-1]:=xName;
      FApp.Form_Event(fFormID,fID,xName,fEventCount);
      end
    else
      FApp.Form_Event(FFormID,FID,xName,0);
    end;
  end;

procedure TRtc_Com.Paint(const Can:TCanvas);
  begin
  Loaded;
  if assigned(FView) then // Design-time
    FView._Paint(Can);
  end;

procedure TRtc_Com.Loaded;
  begin
  if assigned(FView) then
    begin
    if FView._Loading then
      begin
      CheckMe('');
      FView._EndLoad;
      UpdateMe;
      end;
    end;
  end;

procedure TRtc_Com.CheckMe(const xName:string);
  begin
  if isControl and assigned(FView) and (FView._Self is TControl) then
    begin
    if (xName='left') or (xName='top') or
       (xName='width') or (xName='height') then
      HaveBnds:=False
    else
      begin
      HaveBnds:=True;
      Bnds.Left:=TControl(FView._Self).Left;
      Bnds.Top:=TControl(FView._Self).Top;
      Bnds.Right:=TControl(FView._Self).Width;
      Bnds.Bottom:=TControl(FView._Self).Height;
      end;
    end
  else
    HaveBnds:=False;
  end;

procedure TRtc_Com.UpdateMe;
  begin
  if haveBnds then
    if assigned(FView) and (FView._Self is TControl) then
      begin
      if ( (Bnds.Left<>TControl(FView._Self).Left) and
           (FMe.Left<>TControl(FView._Self).Left) ) or
         ( (Bnds.Top<>TControl(FView._Self).Top) and
           (FMe.Top<>TControl(FView._Self).Top) ) or
         ( (Bnds.Right<>TControl(FView._Self).Width) and
           (FMe.Width<>TControl(FView._Self).Width) ) or
         ( (Bnds.Bottom<>TControl(FView._Self).Height) and
           (FMe.Height<>TControl(FView._Self).Height) ) then
        FMe.SetBounds(TControl(FView._Self).Left, TControl(FView._Self).Top,
                      TControl(FView._Self).Width, TControl(FView._Self).Height);
      FMe.Refresh;
      end;
  end;

function TRtc_Com.Get_Preview: boolean;
  begin
  if assigned(FView) and (FView._Self is TControl) then
    Result:=TControl(FView._Self).Visible
  else
    Result:=False;
  end;

procedure TRtc_Com.Set_Preview(const Value: boolean);
  var
    v:TRtcValue;
  begin
  if assigned(FView) and (FView._Self is TControl) then
    begin
    v:=TRtcValue.Create;
    try
      v.asBoolean:=False;
      if Value then
        FView._Process('show',v)
      else
        FView._Process('hide',v);
      UpdateMe;
    finally
      v.Free;
      end;
    end;
  end;

function TRtc_Com.TabOrder: TTabOrder;
  begin
  if assigned(FView) and (FView._Self is TWinControl) then
    Result:=TWinControl(FView._Self).TabOrder
  else
    Result:=-1;
  end;

procedure TRtc_Com.UpdateNow;
  begin
  if assigned(FApp) then
    FApp.UpdateNow;
  end;

function TRtc_Com.GetUID: TRtcRecord;
  begin
  Result:=TRtcRecord.Create;
  Result.asInteger['f']:=FormID;
  Result.asInteger['c']:=ID;
  end;

function TRtc_Com.FindByUID(UID: TRtcRecord): TComponent;
  var
    c:IRtcfComponent;
  begin
  if assigned(FApp) then
    begin
    c:=FApp.Get_Object(UID.asInteger['f'],UID.asInteger['c']);
    if assigned(c) then
      Result:=c._Self
    else
      Result:=nil;
    end
  else
    Result:=nil;
  end;

{ TRtcBaseForm }

constructor TRtcBaseForm.Create(AOwner:TComponent);
  begin
  GlobalNameSpace.BeginWrite;
  try
    CreateNew(AOwner);
    if (ClassType <> TRtcBaseForm) and not (csDesigning in ComponentState) then
      begin
      if not InitInheritedComponent(Self, TRtcBaseForm) then
        raise EResNotFound.CreateFmt('Resource %s not found', [ClassName]);
      end;
  finally
    GlobalNameSpace.EndWrite;
    end;

  {$IFDEF IDE_2005up}
  FIDE_Link:=False;
  {$ELSE}
  FIDE_Link:=(csDesigning in ComponentState);
  {$ENDIF}

  FCom.Loaded;
  if assigned(fApp) then
    begin
    fApp.Form_Loaded(fID);
    // First form to be created automatically becomes the main form,
    // unless another form has  already been set as main form
    if fApp.MainFormID=0 then
      fApp.MainFormID:=fID;
    end;
  if assigned(FOnCreate) then
    try
      FOnCreate(self);
    except
      on E:Exception do
        raise Exception.Create(ClassName+'.oncreate - '+E.ClassName+#13#10+E.Message);
      end;
  end;

constructor TRtcBaseForm.CreateNew(AOwner: TComponent; Dummy:integer=0);
  begin
  PrepareCom;
  inherited Create(AOwner);

  FFnt:=TRtcFont.Create(inherited Font);
  FFnt._Com:=_Com;
  FFnt.FullName:='font.';
  FFnt.OnChange:=MyFontChange;

  FVisible:=False;
  FLeft:=0;
  FTop:=0;
  FKeyPreview:=False;
  FWindowState:=wsNormal;
  FTransparentColorValue:=clBlack;
  FTransparentColor:=False;
  FPosition:=poDesigned;
  FFormStyle:=fsNormal;
  FBorderIcons:=[biSystemMenu,biMinimize,biMaximize];
  FBorderStyle:=bsSizeable;
  FAlphaBlend:=False;
  FAlphaBlendValue:=255;
  FScreenSnap:=False;
  FSnapBuffer:=10;
  FDefaultMonitor:=dmActiveForm;

  ControlStyle := RtcFormControlStyle;

  inherited Width:=300;
  inherited Height:=200;
  end;

procedure TRtcBaseForm.PrepareCom;
  begin
  if not (csDesigning in ComponentState) then
    begin
    fApp:=GetApplication;
    if assigned(fApp) then
      fID:=fApp.Form_Create(self,_ClassName);
    end
  else
    fApp:=nil;

  LastObjID:=0;
  SetLength(fObjs,0);

  FCom:=TRtc_Com.Create(self,self,_ClassName);
  end;

procedure TrtcBaseForm.BeforeDestruction;
  begin
  inherited;
  if assigned(FOnDestroy) then
    try
      FOnDestroy(self);
    except on E:Exception do
      raise Exception.Create(ClassName+'.ondestroy - '+E.ClassName+#13#10+E.Message);
    end;
  end;

destructor TRtcBaseForm.Destroy;
  var
    c:TObject;
  begin
  if assigned(fApp) then
    fApp.Form_Destroy(fID);

  _Detach;

  FreeAndNil(FFnt);
  c:=FCom;
  inherited;
  if assigned(c) then FreeAndNil(c);
  end;

procedure TRtcBaseForm._Detach;
  var
    a:integer;
    c:TRtcCom;
  begin
  if assigned(FCom) then FCom.Detach;

  if assigned(fObjs) then
    for a:=length(fObjs)-1 downto 0 do
      if assigned(fObjs[a]) then
        begin
        c:=fObjs[a]._Com;
        if assigned(c) then TRtc_Com(c).Detach;
        fObjs[a]:=nil;
        end;
  LastObjID:=0;
  SetLength(fObjs,0);

  fApp:=nil;
  end;

procedure TRtcBaseForm.UpdateNow;
  begin
  if assigned(fApp) then fApp.UpdateNow;
  end;

function TRtcBaseForm.New_Object(obj:IrtcfComponent): longint;
  begin
  Inc(LastObjID);
  SetLength(fObjs,LastObjID);
  fObjs[LastObjID-1]:=obj;
  Result:=LastObjID;
  end;

procedure TRtcBaseForm.Old_Object(id: Integer);
  begin
  if (id>0) and (id<=LastObjID) then
    fObjs[id-1]:=nil;
  end;

function TRtcBaseForm.Get_Object(id: integer):IrtcfComponent;
  begin
  if id=0 then
    Result:=(self as IrtcfComponent)
  else if (id>0) and (LastObjID>=id) then
    Result:=fObjs[id-1]
  else
    Result:=nil;
  end;

function TRtcBaseForm._ID: longint;
  begin
  Result:=FID;
  end;

function TRtcBaseForm._Owner:TObject;
  begin
  Result:=Owner;
  end;

function TRtcBaseForm._Self:TComponent;
  begin
  Result:=Self;
  end;

function TRtcBaseForm._App: TRtcBaseApplication;
  begin
  Result:=fApp;
  end;

function TRtcBaseForm._Com: TRtcCom;
  begin
  Result:=FCom;
  end;

function TRtcBaseForm.Get_Caption: TCaption;
  begin
  Result:=inherited Caption;
  end;

procedure TRtcBaseForm.Set_Caption(const Value: TCaption);
  begin
  inherited Caption:=Value;
  end;

function TRtcBaseForm.Get_Left: integer;
  begin
  if IDE_Link and assigned(Parent) then
    FLeft:=Parent.Left;
  Result:=FLeft;
  end;

function TRtcBaseForm.Get_Top: integer;
  begin
  if IDE_Link and assigned(Parent) then
    FTop:=Parent.Top;
  Result:=FTop;
  end;

procedure TRtcBaseForm.Set_Left(const Value: integer);
  begin
  if FLeft<>Value then
    begin
    FLeft:=Value;
    if IDE_Link and assigned(Parent) then
      Parent.Left:=FLeft;
    _Com.Prop('left',Value);
    end;
  end;

procedure TRtcBaseForm.Set_Top(const Value: integer);
  begin
  if FTop<>Value then
    begin
    FTop:=Value;
    if IDE_Link and assigned(Parent) then
      Parent.Top:=FTop;
    _Com.Prop('top',Value);
    end;
  end;

function TRtcBaseForm.Get_ClientHeight: integer;
  begin
  Result:=inherited Height;
  end;

function TRtcBaseForm.Get_ClientWidth: integer;
  begin
  Result:=inherited Width;
  end;

function TRtcBaseForm.Get_Height: integer;
  begin
  if IDE_Link and assigned(Parent) and not (csLoading in ComponentState) then
    Result:=Parent.Height
  else
    Result:=inherited Height+34;
  end;

function TRtcBaseForm.Get_Width: integer;
  begin
  if IDE_Link and assigned(Parent) and not (csLoading in ComponentState) then
    Result:=Parent.Width
  else
    Result:=inherited Width+8;
  end;

procedure TrtcBaseForm.Set_ClientHeight(const Value: integer);
  begin
  if Value<>inherited Height then
    begin
    inherited Height:=Value;
    _Com.Prop('height',Value);
    end;
  end;

procedure TrtcBaseForm.Set_ClientWidth(const Value: integer);
  begin
  if Value<>inherited Width then
    begin
    inherited Width:=Value;
    _Com.Prop('width',Value);
    end;
  end;

procedure TrtcBaseForm.Set_Height(const Value: integer);
  begin
  if Value<>Height then
    begin
    if IDE_Link and assigned(Parent) and not (csLoading in ComponentState) then
      Parent.Height:=Value
    else
      inherited Height:=Value-34;
    inherited Height:=Value;
    _Com.Prop('pheight',Value);
    end;
  end;

procedure TrtcBaseForm.Set_Width(const Value: integer);
  begin
  if Value<>Width then
    begin
    if IDE_Link and assigned(Parent) and not (csLoading in ComponentState) then
      Parent.Width:=Value
    else
      inherited Width:=Value-8;
    _Com.Prop('pwidth',Value);
    end;
  end;

procedure TRtcBaseForm.GetChildren(Proc: TGetChildProc; Root: TComponent);
  var
    I: Integer;
    OwnedComponent: TComponent;
  begin
  inherited GetChildren(Proc, Root);
  if Root = Self then
    for I := 0 to ComponentCount - 1 do
      begin
      OwnedComponent := Components[I];
      if not OwnedComponent.HasParent then Proc(OwnedComponent);
      end;
  end;

function TRtcBaseForm._ClassName: string;
  begin
  Result:='Form';
  end;

procedure TrtcBaseForm.Set_Preview(const Value: boolean);
  begin
  FCom.Preview:=Value;
  end;

function TrtcBaseForm.Get_Preview: boolean;
  begin
  Result:=FCom.Preview;
  end;

procedure TRtcBaseForm.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
  var
    aw,ah:boolean;
  begin
  if (csDesigning in ComponentState) then
    begin
    aw:=AWidth<>ClientWidth;
    ah:=AHeight<>ClientHeight;
    inherited;
    if aw then _Com.Prop('width',ClientWidth);
    if ah then _Com.Prop('height',ClientHeight);
    end;
  end;

procedure TrtcBaseForm._Process(const xName: string; xData: TRtcValue);
  begin
  if xName='x' then
    Close
  else if xName='w' then
    inherited Width:=xData.asInteger
  else if xName='h' then
    inherited Height:=xData.asInteger;
  // Execute Form event "xName" with parameters "xData"
  end;

function TrtcBaseForm.Get_Visible: boolean;
  begin
  Result:=FVisible;
  end;

procedure TrtcBaseForm.Set_Visible(const Value: boolean);
  begin
  if Value<>FVisible then
    begin
    FVisible:=Value;
    if FVisible then
      begin
      _Com.Send('show',nil);
      if assigned(FOnShow) then
        try
          FOnShow(self);
        except
          on E:Exception do
            raise Exception.Create(ClassName+'.onshow - '+E.ClassName+#13#10+E.Message);
        end;
      end
    else
      begin
      _Com.Send('hide',nil);
      if assigned(FOnHide) then
        try
          FOnHide(self);
        except
          on E:Exception do
            raise Exception.Create(ClassName+'.onhide - '+E.ClassName+#13#10+E.Message);
        end;
      end;
    UpdateNow;
    end;
  end;

procedure TrtcBaseForm.Hide;
  begin
  Visible:=False;
  end;

procedure TrtcBaseForm.Show;
  begin
  Visible:=True;
  end;

procedure TrtcBaseForm.Close;
  var
    can:boolean;
    act:TRtcCloseAction;
  begin
  can:=True;
  if assigned(FOnCloseQuery) then
    try
      FOnCloseQuery(self,can);
    except
      on E:Exception do
        raise Exception.Create(ClassName+'.onclosequery -'+E.ClassName+#13#10+E.Message);
    end;

  if can then
    begin
    if FormStyle = fsMDIChild then
      begin
      if biMinimize in BorderIcons then
        act := caMinimize
      else
        act := caNone;
      end
    else
      act := caHide;

    if assigned(FOnClose) then
      try
        FOnClose(self,act);
      except
        on E:Exception do
          raise Exception.Create(ClassName+'.onclose - '+E.ClassName+#13#10+E.Message);
      end;

    if act=caNone then
      UpdateNow
    else if isMain and (act in [caHide,caFree]) then
      _App.Quit
    else if act = caHide then
      begin
      Hide;
      UpdateNow;
      end
    else if act = caMinimize then
      begin
      WindowState := wsMinimized;
      UpdateNow;
      end
    else
      Free;
    end
  else
    UpdateNow;
  end;

function TRtcBaseForm.Get_isMain: boolean;
  begin
  Result:= _App.MainFormID=_ID;
  end;

procedure TRtcBaseForm.Set_isMain(const Value: boolean);
  begin
  if isMain<>Value then
    begin
    if Value then
      _App.MainFormID:=_ID
    else
      _App.MainFormID:=0;
    _Com.Send('ismain',Value);
    end;
  end;

procedure TRtcBaseForm.BringToFront;
  begin
  _Com.Send('tofront',nil);
  end;

procedure TRtcBaseForm.SendToBack;
  begin
  _Com.Send('toback',nil);
  end;

procedure TRtcBaseForm.SetFocus;
  begin
  if (csDesigning in ComponentState) then
    inherited SetFocus
  else
    _Com.Send('setfocus',nil);
  end;

function TRtcBaseForm.Get_Font: TRtcFont;
  begin
  Result:=FFnt;
  end;

procedure TRtcBaseForm.MyFontChange(Sender: TObject);
  begin
  ParentFont:=False;
  if (csDesigning in ComponentState) then
    Refresh;
  end;

procedure TRtcBaseForm.Set_Font(const Value: TRtcFont);
  begin
  FFnt.Assign(Value);
  end;

procedure TRtcBaseForm.Set_IDELink(const Value: boolean);
  begin
  FIDE_Link := Value;
  end;

{ TRtcBaseDataModule }

constructor TRtcBaseDataModule.CreateNew(AOwner: TComponent; Dummy:integer=0);
  begin
  PrepareCom;
  inherited;
  end;

procedure TRtcBaseDataModule.PrepareCom;
  begin
  if not (csDesigning in ComponentState) then
    begin
    fApp:=GetApplication;
    if assigned(fApp) then
      fID:=fApp.Form_Create(self,_ClassName);
    end
  else
    fApp:=nil;

  LastObjID:=0;
  SetLength(fObjs,0);

  FCom:=TRtc_Com.Create(self,self,_ClassName);
  end;

procedure TRtcBaseDataModule.Loaded;
  begin
  inherited;
  FCom.Loaded;
  if assigned(fApp) then fApp.Form_Loaded(fID);
  end;

destructor TRtcBaseDataModule.Destroy;
  var
    c:TObject;
  begin
  if assigned(fApp) then
    fApp.Form_Destroy(fID);

  _Detach;

  c:=FCom;
  inherited;
  if assigned(c) then FreeAndNil(c);
  end;

procedure TRtcBaseDataModule._Detach;
  var
    a:integer;
    c:TRtcCom;
  begin
  FCom.Detach;
  if assigned(fObjs) then
    for a:=length(fObjs)-1 downto 0 do
      if assigned(fObjs[a]) then
        begin
        c:=fObjs[a]._Com;
        if assigned(c) then TRtc_Com(c).Detach;
        fObjs[a]:=nil;
        end;
  LastObjID:=0;
  SetLength(fObjs,0);

  fApp:=nil;
  end;

procedure TRtcBaseDataModule.UpdateNow;
  begin
  if assigned(fApp) then fApp.UpdateNow;
  end;

function TRtcBaseDataModule.New_Object(obj:IrtcfComponent): longint;
  begin
  Inc(LastObjID);
  SetLength(fObjs,LastObjID);
  fObjs[LastObjID-1]:=obj;
  Result:=LastObjID;
  end;

procedure TRtcBaseDataModule.Old_Object(id: Integer);
  begin
  if (id>0) and (id<=LastObjID) then
    fObjs[id-1]:=nil;
  end;

function TRtcBaseDataModule.Get_Object(id: integer):IrtcfComponent;
  begin
  if id=0 then
    Result:=(self as IrtcfComponent)
  else if (id>0) and (LastObjID>=id) then
    Result:=fObjs[id-1]
  else
    Result:=nil;
  end;

function TRtcBaseDataModule._ID: longint;
  begin
  Result:=FID;
  end;

function TRtcBaseDataModule._Owner:TObject;
  begin
  Result:=Owner;
  end;

function TRtcBaseDataModule._Self:TComponent;
  begin
  Result:=Self;
  end;

function TRtcBaseDataModule._App: TRtcBaseApplication;
  begin
  Result:=fApp;
  end;

function TRtcBaseDataModule._Com: TRtcCom;
  begin
  Result:=FCom;
  end;

function TRtcBaseDataModule._ClassName: string;
  begin
  Result:='DataModule';
  end;

{ TRtcBaseApplicationThread }

constructor TRtcBaseApplicationThread.Create;
  begin
  inherited;
  end;

destructor TRtcBaseApplicationThread.Destroy;
  var
    ap:TRtcBaseApplication;
  begin
  if assigned(App) then
    begin
    ap:=App; App:=nil;
    ap.fThr:=nil;

    ap.Form_Data(0,0,'quit',nil);
    ap.UpdateNow;

    ap.OnSend:=nil;
    ap.Free;
    end;
  inherited;
  end;

procedure TRtcBaseApplicationThread.Kill(Job: TObject);
  begin
  inherited;
  end;

function TRtcBaseApplicationThread.Work(Job: TObject): boolean;
  var
    ap:TRtcBaseApplication;
  begin
  Result:=False;
  if Job=Message_AppRun then
    begin
    if assigned(App) then
      begin
      ap:=App;
      ap.MakeActive;
      while ap.Execute do;
      end
    else
      begin
      Result:=True;
      // Free;
      end;
    end
  else if Job=Message_AppQuit then
    begin
    if assigned(App) then
      begin
      ap:=App; App:=nil;
      ap.fThr:=nil;

      ap.Form_Data(0,0,'quit',nil);
      ap.UpdateNow;

      ap.OnSend:=nil;
      FreeAndNil(ap);
      end;
    Result:=True;
    // Free;
    end
  else
    Result:=inherited Work(Job);
  end;

{ TRtcBaseComponent }

constructor TRtcBaseComponent.Create(AOwner: TComponent);
  begin
  FCom:=nil;
  FCom:=TRtc_Com.Create(AOwner,self,_ClassName);
  inherited;
  end;

destructor TRtcBaseComponent.Destroy;
  var
    c:TObject;
  begin
  c:=FCom;
  inherited;
  if assigned(c) then FreeAndNil(c);
  end;

procedure TRtcBaseComponent._Process(const xName: string; xData: TRtcValue);
  begin
  end;

function TRtcBaseComponent._Com: TRtcCom;
  begin
  Result:=FCom;
  end;

procedure TRtcBaseComponent.SetName(const NewName: TComponentName);
  begin
  FCom.Rename(NewName);
  inherited;
  end;

procedure TRtcBaseComponent.UpdateNow;
  begin
  FCom.UpdateNow;
  end;

function TRtcBaseComponent._Self: TComponent;
  begin
  Result:=Self;
  end;

{ TRtcBaseControl }

constructor TRtcBaseControl.Create(AOwner: TComponent);
  begin
  FCom:=nil;
  FCom:=TRtc_Com.Create(AOwner,self,_ClassName);
  inherited;
  FCan := nil;
  end;

destructor TRtcBaseControl.Destroy;
  var
    c:TObject;
  begin
  if assigned(FCan) then FreeAndNil(FCan);
  c:=FCom;
  inherited;
  if assigned(c) then FreeAndNil(c);
  end;

procedure TRtcBaseControl._Process(const xName: string; xData: TRtcValue);
  begin
  end;

function TRtcBaseControl._Com: TRtcCom;
  begin
  Result:=FCom;
  end;

procedure TRtcBaseControl.Paint;
  begin
  // inherited;
  FCom.Paint(Can);
  end;

procedure TRtcBaseControl.WMPaint(var Message: TWMPaint);
  begin
  if Message.DC <> 0 then
    begin
    Can.Lock;
    try
      Can.Handle := Message.DC;
      try
        Paint;
      finally
        Can.Handle := 0;
      end;
    finally
      Can.Unlock;
      end;
    end;
  end;

function TRtcBaseControl.Get_Can: TCanvas;
  begin
  if not assigned(FCan) then
    begin
    FCan := TControlCanvas.Create;
    TControlCanvas(FCan).Control := Self;
    end;
  Result:=FCan;
  end;

procedure TRtcBaseControl.SetName(const NewName: TComponentName);
  var
    ChangeText: Boolean;
  begin
  ChangeText := (csSetCaption in ControlStyle) and
    not (csLoading in ComponentState) and (Name = Text) and
    ((Owner = nil) or not (Owner is TControl) or
    not (csLoading in TControl(Owner).ComponentState));

  FCom.Rename(NewName);
  if ChangeText then Caption:=NewName;
  inherited;
  end;

procedure TRtcBaseControl.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
  var
    al,at,aw,ah:boolean;
  begin
  al:=ALeft<>Left;
  at:=ATop<>Top;
  aw:=AWidth<>Width;
  ah:=AHeight<>Height;
  inherited;
  if al then FCom.Prop('left',Left);
  if at then FCom.Prop('top',Top);
  if aw then FCom.Prop('width',Width);
  if ah then FCom.Prop('height',Height);
  end;

procedure TRtcBaseControl.SetParent(AParent: TWinControl);
  begin
  inherited;
  if assigned(FCom) then FCom.Link('parent',AParent);
  end;

procedure TRtcBaseControl.UpdateNow;
  begin
  FCom.UpdateNow;
  end;

function TRtcBaseControl.IsCaptionStored: Boolean;
  begin
  Result:=True;
  // Result:=inherited IsCaptionStored;
  end;

function TRtcBaseControl.Get_Text: TCaption;
  begin
  Result:=FText;
  end;

procedure TRtcBaseControl.Set_Text(const Value: TCaption);
  begin
  if (Value='') or (FText<>Value) then
    begin
    FText:=Value;
    _Com.Prop('text',Value);
    end;
  end;

function TRtcBaseControl._Self: TComponent;
  begin
  Result:=Self;
  end;

{ TRtcBaseWinControl }

constructor TRtcBaseWinControl.Create(AOwner: TComponent);
  begin
  FCom:=nil;
  FCom:=TRtc_Com.Create(AOwner,self,_ClassName);
  inherited;
  end;

destructor TRtcBaseWinControl.Destroy;
  var
    c:TObject;
  begin
  c:=FCom;
  inherited;
  if assigned(c) then FreeAndNil(c);
  end;

procedure TRtcBaseWinControl._Process(const xName: string; xData: TRtcValue);
  begin
  end;

function TRtcBaseWinControl._Com: TRtcCom;
  begin
  Result:=FCom;
  end;

procedure TRtcBaseWinControl.Paint;
  begin
  inherited;
  FCom.Paint(Can);
  end;

function TRtcBaseWinControl.Get_Can: TCanvas;
  begin
  Result:=inherited Canvas;
  end;

procedure TRtcBaseWinControl.SetName(const NewName: TComponentName);
  var
    ChangeText: Boolean;
  begin
  ChangeText := (csSetCaption in ControlStyle) and
    not (csLoading in ComponentState) and (Name = Text) and
    ((Owner = nil) or not (Owner is TControl) or
    not (csLoading in TControl(Owner).ComponentState));

  FCom.Rename(NewName);
  if ChangeText then Caption:=NewName;
  inherited;
  end;

procedure TRtcBaseWinControl.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
  var
    al,at,aw,ah:boolean;
  begin
  al:=ALeft<>Left;
  at:=ATop<>Top;
  aw:=AWidth<>Width;
  ah:=AHeight<>Height;
  inherited;
  if al then FCom.Prop('left',Left);
  if at then FCom.Prop('top',Top);
  if aw then FCom.Prop('width',Width);
  if ah then FCom.Prop('height',Height);
  end;

procedure TRtcBaseWinControl.SetParent(AParent: TWinControl);
  begin
  inherited;
  if assigned(FCom) then FCom.Link('parent',AParent);
  end;

procedure TRtcBaseWinControl.UpdateNow;
  begin
  FCom.UpdateNow;
  end;

function TRtcBaseWinControl.IsCaptionStored: Boolean;
  begin
  Result:=True;
  // Result:=inherited IsCaptionStored;
  end;

function TRtcBaseWinControl.Get_Text: TCaption;
  begin
  Result:=FText;
  end;

procedure TRtcBaseWinControl.Set_Text(const Value: TCaption);
  begin
  if (Value='') or (FText<>Value) then
    begin
    FText:=Value;
    _Com.Prop('text',Value);
    end;
  end;

function TRtcBaseWinControl._Self: TComponent;
  begin
  Result:=Self;
  end;

initialization
fThrApp:=tObjList.Create(32);
fThrAppCS:=TRtcCritSec.Create;

Message_AppRun:=TRtcBaseApplicationJob.Create;
Message_AppQuit:=TRtcBaseApplicationJob.Create;
finalization
Garbage(fThrApp);
Garbage(fThrAppCS);
Garbage(Message_AppRun);
Garbage(Message_AppQuit);
end.
