{  
  RealThinClient W/A Framework: Client-side functions for setting standard control properties
  Copyright (c) Danijel Tkalcec

  @exclude
}
unit rtcGraphicsCli;

{$include rtcDefs.inc}

{ Some Canvas operations are only available with a player complied using Delphi for Windows.

   Here is the list of Canvas operations which you should avoid using if you want your
   application to look the same on Windows, Linux and MacOSX:

   * Canvas.BrushCopy - will be processed as CopyRect without a transparent color
   * Canvas.PolyBezierTo - will be processed as PolyBezier(arr,false,true);
   * Canvas.TextFlags - will be ignored
   * WMF files - will be ignored or result in an exception, use BMP or ICO instead
}

interface

uses
  rtcTrashcan, rtcLog,
  {$ifdef Windows} Windows, {$endif}
  Classes, SysUtils, Graphics,
  
  rtcInfo;

function Rtcf_ProcessFont(c:TFont; const xName:String; xData:TRtcValue):boolean;
function Rtcf_ProcessPen(c:TPen; const xName:String; xData:TRtcValue):boolean;
function Rtcf_ProcessBrush(c:TBrush; const xName:String; xData:TRtcValue):boolean;
function Rtcf_ProcessCanvas(c:TCanvas; const xName:String; xData:TRtcValue):boolean;

implementation

function Rtcf_ProcessFont(c:TFont; const xName:String; xData:TRtcValue):boolean;
  var
    prop:string;
  procedure SetStyle;
    var
      rec:TRtcRecord;
      fs:TFontStyles;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      fs:=[];
      if rec.asBoolean['b'] then Include(fs,fsBold);
      if rec.asBoolean['i'] then Include(fs,fsItalic);
      if rec.asBoolean['u'] then Include(fs,fsUnderline);
      if rec.asBoolean['s'] then Include(fs,fsStrikeOut);
      c.Style:=fs;
      end;
    end;
  begin
  Result:=True;
  if pos('.',xName)>0 then // remove prefix
    prop:=Copy(xName,Pos('.',xName)+1,length(xName))
  else
    prop:=xName;

  if prop='charset' then      c.Charset:=TFontCharset(xData.asInteger)
  else if prop='color' then   c.Color:=TColor(xData.asInteger)
  else if prop='size' then    c.Size:=xData.asInteger
  else if prop='height' then  c.Height:=xData.asInteger
  else if prop='name' then    c.Name:=xData.asText
  else if prop='pitch' then   c.Pitch:=TFontPitch(xData.asInteger)
  else if prop='size' then    c.Size:=xData.asInteger
  else if prop='style' then   SetStyle
  else
    Result:=False;
  end;

function Rtcf_ProcessPen(c:TPen; const xName:String; xData:TRtcValue):boolean;
  var
    prop:string;
  begin
  Result:=True;
  if pos('.',xName)>0 then // remove prefix
    prop:=Copy(xName,Pos('.',xName)+1,length(xName))
  else
    prop:=xName;

  if prop='color' then        c.Color:=TColor(xData.asInteger)
  else if prop='mode' then    c.Mode:=TPenMode(xData.asInteger)
  else if prop='style' then   c.Style:=TPenStyle(xData.asInteger)
  else if prop='width' then   c.Width:=xData.asInteger
  else
    Result:=False;
  end;

function Rtcf_ProcessBrush(c:TBrush; const xName:String; xData:TRtcValue):boolean;
  var
    prop:string;
  procedure SetBitmap;
    var
      rec:TStream;
    begin
    if xData.isType=rtc_ByteStream then
      begin
      rec:=xData.asByteStream;
      if not assigned(c.Bitmap) then
        c.Bitmap:=TBitmap.Create;
      rec.Position:=0;
      c.Bitmap.LoadFromStream(rec);
      end
    else
      c.Bitmap:=nil;
    end;
  begin
  Result:=True;
  if pos('.',xName)>0 then // remove prefix
    prop:=Copy(xName,Pos('.',xName)+1,length(xName))
  else
    prop:=xName;

  if prop='color' then          c.Color:=TColor(xData.asInteger)
  else if prop='style' then     c.Style:=TBrushStyle(xData.asInteger)
  else if prop='bitmap' then    SetBitmap
  else
    Result:=False;
  end;

function Rtcf_ProcessCanvas(c:TCanvas; const xName:String; xData:TRtcValue):boolean;
  var
    prop:string;
  procedure DoArc;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.Arc(rec.asInteger['x1'],rec.asInteger['y1'],
            rec.asInteger['x2'],rec.asInteger['y2'],
            rec.asInteger['x3'],rec.asInteger['y3'],
            rec.asInteger['x4'],rec.asInteger['y4']);
      end;
    end;
  procedure DoChord;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.Chord(rec.asInteger['x1'],rec.asInteger['y1'],
              rec.asInteger['x2'],rec.asInteger['y2'],
              rec.asInteger['x3'],rec.asInteger['y3'],
              rec.asInteger['x4'],rec.asInteger['y4']);
      end;
    end;
  procedure DoBrushCopy;
    var
      rec:TRtcRecord;
      r:TRtcRecord;
      srec,drec:TRect;
      bmp:TBitmap;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      if (rec.isType['s']=rtc_Record) and
         (rec.isType['d']=rtc_Record) and
         (rec.isType['bmp']=rtc_ByteStream) then
        begin
        r:=rec.asRecord['s'];
        srec:=Rect(r.asInteger['l'],r.asInteger['t'],
                   r.asInteger['r'],r.asInteger['b']);
        r:=rec.asRecord['d'];
        drec:=Rect(r.asInteger['l'],r.asInteger['t'],
                   r.asInteger['r'],r.asInteger['b']);
        bmp:=TBitmap.Create;
        try
          rec.asByteStream['bmp'].Position:=0;
          bmp.LoadFromStream(rec.asByteStream['bmp']);
          {$ifndef FPC}
          c.BrushCopy(drec,bmp,srec,TColor(rec.asInteger['c']));
          {$else}
          c.CopyRect(drec,bmp.Canvas,srec);
          {$endif}
        finally
          bmp.Free;
          end;
        end;
      end;
    end;
  procedure DoCopyRect;
    var
      rec:TRtcRecord;
      r:TRtcRecord;
      srec,drec:TRect;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      if (rec.isType['s']=rtc_Record) and
         (rec.isType['d']=rtc_Record) and
         (rec.isType['bmp']=rtc_ByteStream) then
        begin
        r:=rec.asRecord['s'];
        srec:=Rect(r.asInteger['l'],r.asInteger['t'],
                   r.asInteger['r'],r.asInteger['b']);
        r:=rec.asRecord['d'];
        drec:=Rect(r.asInteger['l'],r.asInteger['t'],
                   r.asInteger['r'],r.asInteger['b']);
        c.CopyRect(drec,c,srec);
        end;
      end;
    end;
  procedure DoDraw;
    var
      rec:TRtcRecord;
      img:TGraphic;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      if rec.isType['img']=rtc_ByteStream then
        begin
        if rec.asText['cls']='bmp' then
          img:=TBitmap.Create
        else if rec.asText['cls']='ico' then
          img:=TIcon.Create
        {$ifndef FPC}
        else if rec.asText['cls']='wmf' then
          img:=TMetafile.Create
        {$endif}
        else
          img:=TGraphicClass(FindClass(rec.asText['cls'])).Create;
        try
          rec.asByteStream['img'].Position:=0;
          img.LoadFromStream(rec.asByteStream['img']);
          c.Draw(rec.asInteger['x'],rec.asInteger['y'],img);
        finally
          img.Free;
          end;
        end;
      end;
    end;
  procedure DoStretchDraw;
    var
      rec:TRtcRecord;
      img:TGraphic;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      if rec.isType['img']=rtc_ByteStream then
        begin
        if rec.asText['cls']='bmp' then
          img:=TBitmap.Create
        else if rec.asText['cls']='ico' then
          img:=TIcon.Create
        {$ifndef FPC}
        else if rec.asText['cls']='wmf' then
          img:=TMetafile.Create
        {$endif}
        else
          img:=TGraphicClass(FindClass(rec.asText['cls'])).Create;
        try
          rec.asByteStream['img'].Position:=0;
          img.LoadFromStream(rec.asByteStream['img']);
          c.StretchDraw(Rect(rec.asInteger['l'],rec.asInteger['t'],
                             rec.asInteger['r'],rec.asInteger['b']), img);
        finally
          img.Free;
          end;
        end;
      end;
    end;
  procedure DoDrawFocusRect;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.DrawFocusRect( Rect(rec.asInteger['l'],rec.asInteger['t'],
                            rec.asInteger['r'],rec.asInteger['b']) );
      end;
    end;
  procedure DoEllipse;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.Ellipse(rec.asInteger['l'], rec.asInteger['t'],
                rec.asInteger['r'],rec.asInteger['b'] );
      end;
    end;
  procedure DoFillRect;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.FillRect( Rect(rec.asInteger['l'],rec.asInteger['t'],
                       rec.asInteger['r'],rec.asInteger['b']) );
      end;
    end;
  procedure DoFloodFill;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.FloodFill(rec.asInteger['x'],rec.asInteger['x'],
                  TColor(rec.asInteger['c']),
                  TFillStyle(rec.asInteger['s']) );
      end;
    end;
  procedure DoFrameRect;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.FrameRect( Rect(rec.asInteger['l'],rec.asInteger['t'],
                       rec.asInteger['r'],rec.asInteger['b']) );
      end;
    end;
  procedure DoLineTo;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.LineTo(rec.asInteger['x'],rec.asInteger['y']);
      end;
    end;
  procedure DoMoveTo;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.MoveTo(rec.asInteger['x'],rec.asInteger['y']);
      end;
    end;
  procedure DoPie;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.Pie(rec.asInteger['x1'],rec.asInteger['y1'],
            rec.asInteger['x2'],rec.asInteger['y2'],
            rec.asInteger['x3'],rec.asInteger['y3'],
            rec.asInteger['x4'],rec.asInteger['y4']);
      end;
    end;
  procedure DoRectangle;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.Rectangle( rec.asInteger['l'],rec.asInteger['t'],
                   rec.asInteger['r'],rec.asInteger['b'] );
      end;
    end;
  procedure DoRoundRect;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.RoundRect( rec.asInteger['x1'],rec.asInteger['y1'],
                   rec.asInteger['x2'],rec.asInteger['y2'],
                   rec.asInteger['x3'],rec.asInteger['y3']);
      end;
    end;
  procedure DoSetPixel;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.Pixels[ rec.asInteger['x'],rec.asInteger['y'] ] := TColor(rec.asInteger['c']);
      end;
    end;
  procedure DoTextOut;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.TextOut(rec.asInteger['x'],rec.asInteger['y'],rec.asText['s']);
      end;
    end;
  procedure DoTextRect;
    var
      rec:TRtcRecord;
    begin
    if xData.isType=rtc_Record then
      begin
      rec:=xData.asRecord;
      c.TextRect( Rect(rec.asInteger['l'],rec.asInteger['t'],
                       rec.asInteger['r'],rec.asInteger['b']),
                  rec.asInteger['x'],rec.asInteger['y'],
                  rec.asText['s']);
      end;
    end;
  procedure DoPolyBezier;
    var
      rec:TRtcDataSet;
      a:integer;
      arr:array of TPoint;
    begin
    if xData.isType=rtc_DataSet then
      begin
      rec:=xData.asDataSet;
      try
        rec.First;
        SetLength(arr,rec.RowCount);
        for a:=0 to length(arr)-1 do
          begin
          arr[a].X:=rec.asInteger['x'];
          arr[a].Y:=rec.asInteger['y'];
          rec.Next;
          end;
        c.PolyBezier(arr);
      finally
        SetLength(arr,0);
        end;
      end;
    end;
  procedure DoPolyBezierTo;
    var
      rec:TRtcDataSet;
      a:integer;
      arr:array of TPoint;
    begin
    if xData.isType=rtc_DataSet then
      begin
      rec:=xData.asDataSet;
      try
        rec.First;
        SetLength(arr,rec.RowCount);
        for a:=0 to length(arr)-1 do
          begin
          arr[a].X:=rec.asInteger['x'];
          arr[a].Y:=rec.asInteger['y'];
          rec.Next;
          end;
        {$ifdef FPC}
        c.PolyBezier(arr,False,True);
        {$else}
        c.PolyBezierTo(arr);
        {$endif}
      finally
        SetLength(arr,0);
        end;
      end;
    end;
  procedure DoPolygon;
    var
      rec:TRtcDataSet;
      a:integer;
      arr:array of TPoint;
    begin
    if xData.isType=rtc_DataSet then
      begin
      rec:=xData.asDataSet;
      try
        rec.First;
        SetLength(arr,rec.RowCount);
        for a:=0 to length(arr)-1 do
          begin
          arr[a].X:=rec.asInteger['x'];
          arr[a].Y:=rec.asInteger['y'];
          rec.Next;
          end;
        c.Polygon(arr);
      finally
        SetLength(arr,0);
        end;
      end;
    end;
  procedure DoPolyLine;
    var
      rec:TRtcDataSet;
      a:integer;
      arr:array of TPoint;
    begin
    if xData.isType=rtc_DataSet then
      begin
      rec:=xData.asDataSet;
      try
        rec.First;
        SetLength(arr,rec.RowCount);
        for a:=0 to length(arr)-1 do
          begin
          arr[a].X:=rec.asInteger['x'];
          arr[a].Y:=rec.asInteger['y'];
          rec.Next;
          end;
        c.PolyLine(arr);
      finally
        SetLength(arr,0);
        end;
      end;
    end;
  begin
  Result:=True;
  if pos('.',xName)>0 then // remove prefix
    prop:=Copy(xName,Pos('.',xName)+1,length(xName))
  else
    prop:=xName;

  if prop='refresh' then
    c.Refresh
  else if prop='copymode' then     c.CopyMode:=xData.asInteger
  {$ifndef FPC}
  else if prop='tflags' then       c.TextFlags:=xData.asInteger
  {$endif}
  else if prop='arc' then          DoArc
  else if prop='chord' then        DoChord
  else if prop='brushcopy' then    DoBrushCopy
  else if prop='copyrect' then     DoCopyRect
  else if prop='draw' then         DoDraw
  else if prop='sdraw' then        DoStretchDraw
  else if prop='focusrect' then    DoDrawFocusRect
  else if prop='ellipce' then      DoEllipse
  else if prop='fillrect' then     DoFillRect
  else if prop='fill' then         DoFloodFill
  else if prop='framerect' then    DoFrameRect
  else if prop='line' then         DoLineTo
  else if prop='move' then         DoMoveTo
  else if prop='pie' then          DoPie
  else if prop='rect' then         DoRectangle
  else if prop='roundrect' then    DoRoundRect
  else if prop='pixel' then        DoSetPixel
  else if prop='text' then         DoTextOut
  else if prop='textrect' then     DoTextRect
  else if prop='polyb' then        DoPolyBezier
  else if prop='polybto' then      DoPolyBezierTo
  else if prop='poly' then         DoPolygon
  else if prop='polyl' then        DoPolyLine
  else if Copy(prop,1,4)='pen.' then    Rtcf_ProcessPen(c.Pen,prop,xData)
  else if Copy(prop,1,6)='brush.' then  Rtcf_ProcessBrush(c.Brush,prop,xData)
  else if Copy(prop,1,5)='font.' then   Rtcf_ProcessFont(c.Font,prop,xData)
  else
    Result:=False;
  end;

end.
