(*
   RealThinClient W/A Framework: Optimized MD5 functions
   Copyright (c) Danijel Tkalcec

   Original from Ronald L. Rivest's RFC 1321

  @exclude
*)
unit rtcMD5;

interface

{$INCLUDE rtcDefs.inc}
{$Q-}{$R-}

{$ifdef WINDOWS}
uses
	Windows;
{$ENDIF}

type
	MD5Count = array[0..1] of DWORD;
	MD5State = array[0..3] of DWORD;
	MD5Block = array[0..15] of DWORD;
	MD5CBits = array[0..7] of byte;
	MD5Digest = array[0..15] of byte;
	MD5Buffer = array[0..63] of byte;
	MD5Context = record
		State: MD5State;
		Count: MD5Count;
		Buffer: MD5Buffer;
	end;

{ Standard MD5 hash as Hex string (length=32)
  4 bits per character using '0'-'9' and 'a'-'f' }
function MD5Str(const M: AnsiString):AnsiString;

{ Standard MD5 hash as Hex GUID
  Example: '<$guid!>' }
function MD5asGUID(const M: AnsiString):AnsiString;

{ RTC-Coded MD5 hash, base32 encoded (length=26)
  Safe for use in HTML and URLs (only contains alpha-numerics)
  5 bits per character using '0'-'9' and 'a'-'v' }
function MD5Code(const M: AnsiString):AnsiString;

{ RTC-Coded Case-sensitive MD5 hash, base64 encoded (length=22)
  Safe for use in HTML and URLs (does not contain reserved HTML characters)
  NOT safe for use in File names (case sensitive)
  6 bits per character using '0'-'9', 'a'-'z', 'A'-'Z', '_' and '$'}
function MD5Code6bit(const M: AnsiString):AnsiString;

{ Binary coded MD5 hash (length=16)
  8 bits per character using #0 - #255 }
function MD5Code8bit(const M: AnsiString):AnsiString;

(* Low-level MD5 functions *)

procedure rtcMD5Init(var Context: MD5Context);
procedure rtcMD5Update(var Context: MD5Context; Input: PAnsiChar; Length: longword);
procedure rtcMD5Final(var Context: MD5Context; var Digest: MD5Digest);

function rtcMD5String(const M: AnsiString): MD5Digest;
function rtcMD5File(const N: String): MD5Digest;
function rtcMD5Print(D: MD5Digest): AnsiString;

function rtcMD5Match(D1, D2: MD5Digest): boolean;

function GetMD5(Buffer:pointer; Length:longword):AnsiString;

implementation

uses
{$ifdef FPC}
  rtcInfo, 
{$ENDIF}
  SysUtils, rtcFastStrings;

var
	PADDING: MD5Buffer = (
		$80, $00, $00, $00, $00, $00, $00, $00,
		$00, $00, $00, $00, $00, $00, $00, $00,
		$00, $00, $00, $00, $00, $00, $00, $00,
		$00, $00, $00, $00, $00, $00, $00, $00,
		$00, $00, $00, $00, $00, $00, $00, $00,
		$00, $00, $00, $00, $00, $00, $00, $00,
		$00, $00, $00, $00, $00, $00, $00, $00,
		$00, $00, $00, $00, $00, $00, $00, $00
	);

// Transform State according to first 64 bytes at Buffer
procedure Transform(Buffer: pointer; var State: MD5State);
  var
  	a, b, c, d: DWORD;
  	Block: ^MD5Block;
  begin
  Block:=Buffer;

	a := State[0];
	b := State[1];
	c := State[2];
	d := State[3];

  inc(a, ((b and c) or ((not b) and d)) + Block[ 0] + $d76aa478); a := (a shl 7) or (a shr (32 - 7)) + b;
	inc(d, ((a and b) or ((not a) and c)) + Block[ 1] + $e8c7b756); d := (d shl 12) or (d shr (32 - 12)) + a;
	inc(c, ((d and a) or ((not d) and b)) + Block[ 2] + $242070db); c := (c shl 17) or (c shr (32 - 17)) + d;
	inc(b, ((c and d) or ((not c) and a)) + Block[ 3] + $c1bdceee); b := (b shl 22) or (b shr (32 - 22)) + c;
	inc(a, ((b and c) or ((not b) and d)) + Block[ 4] + $f57c0faf); a := (a shl 7) or (a shr (32 - 7)) + b;
	inc(d, ((a and b) or ((not a) and c)) + Block[ 5] + $4787c62a); d := (d shl 12) or (d shr (32 - 12)) + a;
	inc(c, ((d and a) or ((not d) and b)) + Block[ 6] + $a8304613); c := (c shl 17) or (c shr (32 - 17)) + d;
	inc(b, ((c and d) or ((not c) and a)) + Block[ 7] + $fd469501); b := (b shl 22) or (b shr (32 - 22)) + c;
	inc(a, ((b and c) or ((not b) and d)) + Block[ 8] + $698098d8); a := (a shl 7) or (a shr (32 - 7)) + b;
	inc(d, ((a and b) or ((not a) and c)) + Block[ 9] + $8b44f7af); d := (d shl 12) or (d shr (32 - 12)) + a;
	inc(c, ((d and a) or ((not d) and b)) + Block[10] + $ffff5bb1); c := (c shl 17) or (c shr (32 - 17)) + d;
	inc(b, ((c and d) or ((not c) and a)) + Block[11] + $895cd7be); b := (b shl 22) or (b shr (32 - 22)) + c;
	inc(a, ((b and c) or ((not b) and d)) + Block[12] + $6b901122); a := (a shl 7) or (a shr (32 - 7)) + b;
	inc(d, ((a and b) or ((not a) and c)) + Block[13] + $fd987193); d := (d shl 12) or (d shr (32 - 12)) + a;
	inc(c, ((d and a) or ((not d) and b)) + Block[14] + $a679438e); c := (c shl 17) or (c shr (32 - 17)) + d;
	inc(b, ((c and d) or ((not c) and a)) + Block[15] + $49b40821); b := (b shl 22) or (b shr (32 - 22)) + c;

	inc(a, ((b and d) or (c and (not d))) + Block[ 1] + $f61e2562); a := (a shl 5) or (a shr (32 - 5)) + b;
	inc(d, ((a and c) or (b and (not c))) + Block[ 6] + $c040b340); d := (d shl 9) or (d shr (32 - 9)) + a;
	inc(c, ((d and b) or (a and (not b))) + Block[11] + $265e5a51); c := (c shl 14) or (c shr (32 - 14)) + d;
	inc(b, ((c and a) or (d and (not a))) + Block[ 0] + $e9b6c7aa); b := (b shl 20) or (b shr (32 - 20)) + c;
	inc(a, ((b and d) or (c and (not d))) + Block[ 5] + $d62f105d); a := (a shl 5) or (a shr (32 - 5)) + b;
	inc(d, ((a and c) or (b and (not c))) + Block[10] + $02441453); d := (d shl 9) or (d shr (32 - 9)) + a;
	inc(c, ((d and b) or (a and (not b))) + Block[15] + $d8a1e681); c := (c shl 14) or (c shr (32 - 14)) + d;
	inc(b, ((c and a) or (d and (not a))) + Block[ 4] + $e7d3fbc8); b := (b shl 20) or (b shr (32 - 20)) + c;
	inc(a, ((b and d) or (c and (not d))) + Block[ 9] + $21e1cde6); a := (a shl 5) or (a shr (32 - 5)) + b;
	inc(d, ((a and c) or (b and (not c))) + Block[14] + $c33707d6); d := (d shl 9) or (d shr (32 - 9)) + a;
	inc(c, ((d and b) or (a and (not b))) + Block[ 3] + $f4d50d87); c := (c shl 14) or (c shr (32 - 14)) + d;
	inc(b, ((c and a) or (d and (not a))) + Block[ 8] + $455a14ed); b := (b shl 20) or (b shr (32 - 20)) + c;
	inc(a, ((b and d) or (c and (not d))) + Block[13] + $a9e3e905); a := (a shl 5) or (a shr (32 - 5)) + b;
	inc(d, ((a and c) or (b and (not c))) + Block[ 2] + $fcefa3f8); d := (d shl 9) or (d shr (32 - 9)) + a;
	inc(c, ((d and b) or (a and (not b))) + Block[ 7] + $676f02d9); c := (c shl 14) or (c shr (32 - 14)) + d;
	inc(b, ((c and a) or (d and (not a))) + Block[12] + $8d2a4c8a); b := (b shl 20) or (b shr (32 - 20)) + c;

	inc(a, (b xor c xor d) + Block[ 5] + $fffa3942);	a := (a shl 4) or (a shr (32 - 4)) + b;
	inc(d, (a xor b xor c) + Block[ 8] + $8771f681);	d := (d shl 11) or (d shr (32 - 11)) + a;
	inc(c, (d xor a xor b) + Block[11] + $6d9d6122);	c := (c shl 16) or (c shr (32 - 16)) + d;
	inc(b, (c xor d xor a) + Block[14] + $fde5380c);	b := (b shl 23) or (b shr (32 - 23)) + c;
	inc(a, (b xor c xor d) + Block[ 1] + $a4beea44);	a := (a shl 4) or (a shr (32 - 4)) + b;
	inc(d, (a xor b xor c) + Block[ 4] + $4bdecfa9);	d := (d shl 11) or (d shr (32 - 11)) + a;
	inc(c, (d xor a xor b) + Block[ 7] + $f6bb4b60);	c := (c shl 16) or (c shr (32 - 16)) + d;
	inc(b, (c xor d xor a) + Block[10] + $bebfbc70);	b := (b shl 23) or (b shr (32 - 23)) + c;
	inc(a, (b xor c xor d) + Block[13] + $289b7ec6);	a := (a shl 4) or (a shr (32 - 4)) + b;
	inc(d, (a xor b xor c) + Block[ 0] + $eaa127fa);	d := (d shl 11) or (d shr (32 - 11)) + a;
	inc(c, (d xor a xor b) + Block[ 3] + $d4ef3085);	c := (c shl 16) or (c shr (32 - 16)) + d;
	inc(b, (c xor d xor a) + Block[ 6] + $04881d05);	b := (b shl 23) or (b shr (32 - 23)) + c;
	inc(a, (b xor c xor d) + Block[ 9] + $d9d4d039);	a := (a shl 4) or (a shr (32 - 4)) + b;
	inc(d, (a xor b xor c) + Block[12] + $e6db99e5);	d := (d shl 11) or (d shr (32 - 11)) + a;
	inc(c, (d xor a xor b) + Block[15] + $1fa27cf8);	c := (c shl 16) or (c shr (32 - 16)) + d;
	inc(b, (c xor d xor a) + Block[ 2] + $c4ac5665);	b := (b shl 23) or (b shr (32 - 23)) + c;

	inc(a, (c xor (b or (not d))) + Block[ 0] + $f4292244); a := (a shl 6) or (a shr (32 - 6)) + b;
	inc(d, (b xor (a or (not c))) + Block[ 7] + $432aff97); d := (d shl 10) or (d shr (32 - 10)) + a;
	inc(c, (a xor (d or (not b))) + Block[14] + $ab9423a7); c := (c shl 15) or (c shr (32 - 15)) + d;
	inc(b, (d xor (c or (not a))) + Block[ 5] + $fc93a039); b := (b shl 21) or (b shr (32 - 21)) + c;
	inc(a, (c xor (b or (not d))) + Block[12] + $655b59c3); a := (a shl 6) or (a shr (32 - 6)) + b;
	inc(d, (b xor (a or (not c))) + Block[ 3] + $8f0ccc92); d := (d shl 10) or (d shr (32 - 10)) + a;
	inc(c, (a xor (d or (not b))) + Block[10] + $ffeff47d); c := (c shl 15) or (c shr (32 - 15)) + d;
	inc(b, (d xor (c or (not a))) + Block[ 1] + $85845dd1); b := (b shl 21) or (b shr (32 - 21)) + c;
	inc(a, (c xor (b or (not d))) + Block[ 8] + $6fa87e4f); a := (a shl 6) or (a shr (32 - 6)) + b;
	inc(d, (b xor (a or (not c))) + Block[15] + $fe2ce6e0); d := (d shl 10) or (d shr (32 - 10)) + a;
	inc(c, (a xor (d or (not b))) + Block[ 6] + $a3014314); c := (c shl 15) or (c shr (32 - 15)) + d;
	inc(b, (d xor (c or (not a))) + Block[13] + $4e0811a1); b := (b shl 21) or (b shr (32 - 21)) + c;
	inc(a, (c xor (b or (not d))) + Block[ 4] + $f7537e82); a := (a shl 6) or (a shr (32 - 6)) + b;
	inc(d, (b xor (a or (not c))) + Block[11] + $bd3af235); d := (d shl 10) or (d shr (32 - 10)) + a;
	inc(c, (a xor (d or (not b))) + Block[ 2] + $2ad7d2bb); c := (c shl 15) or (c shr (32 - 15)) + d;
	inc(b, (d xor (c or (not a))) + Block[ 9] + $eb86d391); b := (b shl 21) or (b shr (32 - 21)) + c;

	inc(State[0], a);
	inc(State[1], b);
	inc(State[2], c);
	inc(State[3], d);
  end;

// -----------------------------------------------------------------------------------------------

// Initialize given Context
procedure rtcMD5Init(var Context: MD5Context);
  begin
	with Context do
    begin
		State[0] := $67452301;
		State[1] := $efcdab89;
		State[2] := $98badcfe;
		State[3] := $10325476;
		Count[0] := 0;
		Count[1] := 0;
    FillChar(Buffer,SizeOf(MD5Buffer),#0);
		// ZeroMemory(@Buffer, SizeOf(MD5Buffer));
  	end;
  end;

// Update given Context to include Length bytes of Input
procedure rtcMD5Update(var Context: MD5Context; Input: PAnsiChar; Length: longword);
  var
    Index: longword;
    PartLen: longword;
    I: longword;
  begin
	with Context do
    begin
		Index := (Count[0] shr 3) and $3f;
		inc(Count[0], Length shl 3);
		if Count[0] < (Length shl 3) then inc(Count[1]);
		inc(Count[1], Length shr 29);
	  end;
	PartLen := 64 - Index;
	if Length >= PartLen then
    begin
    Move(Input[0],Context.Buffer[Index],PartLen);
		// CopyMemory(@Context.Buffer[Index], Input, PartLen);
		Transform(@Context.Buffer, Context.State);
		I := PartLen;
		while I + 63 < Length do
      begin
			Transform(@Input[I], Context.State);
			inc(I, 64);
		  end;
		Index := 0;
	  end
  else
    I := 0;
  Move(Input[I],Context.Buffer[Index],Length-I);
	// CopyMemory(@Context.Buffer[Index], @Input[I], Length - I);
  end;

// Finalize given Context, create Digest and zeroize Context
procedure rtcMD5Final(var Context: MD5Context; var Digest: MD5Digest);
  var
  	Bits: MD5CBits;
  	Index: longword;
  	PadLen: longword;
  begin
  Move(Context.Count,Bits,2*4);
	//CopyMemory(@Bits, @Context.Count, 2*4);
	Index := (Context.Count[0] shr 3) and $3f;
	if Index < 56 then PadLen := 56 - Index else PadLen := 120 - Index;
	rtcMD5Update(Context, @PADDING, PadLen);
	rtcMD5Update(Context, @Bits, 8);
  Move(Context.State,Digest,4*4);
	//CopyMemory(@Digest, @Context.State, 4*4);
  end;

// -----------------------------------------------------------------------------------------------

// Create digest of given Message
function rtcMD5String(const M: AnsiString): MD5Digest;
  var
  	Context: MD5Context;
  begin
	rtcMD5Init(Context);
	rtcMD5Update(Context, @M[1], length(M));
	rtcMD5Final(Context, Result);
  end;

// Create digest of file with given Name
function rtcMD5File(const N: String): MD5Digest;
  var
  	Context: MD5Context;
  {$IFDEF FPC}
    s: string;
  {$ELSE}
  	FileHandle: THandle;
  	MapHandle: THandle;
  	ViewPointer: pointer;
  {$ENDIF}
  begin
  {$IFDEF FPC}
  s:=Read_File(N);
  rtcMD5Init(Context);
  rtcMD5Update(Context, @S[1], length(S));
  rtcMD5Final(Context, Result);
  {$ELSE}
	rtcMD5Init(Context);
	FileHandle := CreateFile(PChar(N), GENERIC_READ, FILE_SHARE_READ or FILE_SHARE_WRITE,
		nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL or FILE_FLAG_SEQUENTIAL_SCAN, 0);
	if FileHandle <> INVALID_HANDLE_VALUE then
    try
  		MapHandle := CreateFileMapping(FileHandle, nil, PAGE_READONLY, 0, 0, nil);
  		if MapHandle <> 0 then
        try
    			ViewPointer := MapViewOfFile(MapHandle, FILE_MAP_READ, 0, 0, 0);
	    		if ViewPointer <> nil then
            try
      				rtcMD5Update(Context, ViewPointer, GetFileSize(FileHandle, nil));
    	  		finally
    		  		UnmapViewOfFile(ViewPointer);
        		end;
        finally
    			CloseHandle(MapHandle);
        end;
  	finally
  		CloseHandle(FileHandle);
  	end;
	rtcMD5Final(Context, Result);
  {$ENDIF}
  end;

// Create hex representation of given Digest
function rtcMD5Print(D: MD5Digest): AnsiString;
  const
	  Digits: array[0..15] of AnsiChar =
		  ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f');
  var
  	I, J: byte;
  begin
	SetLength(Result,32);
  J:=1;
	for I := 0 to 15 do
    begin
    Result[J]:=Digits[(D[I] shr 4) and $0f];
    Result[J+1]:=Digits[D[I] and $0f];
    Inc(J,2);
    end;
  end;

// -----------------------------------------------------------------------------------------------

// Compare two Digests
function rtcMD5Match(D1, D2: MD5Digest): boolean;
  var
  	I: byte;
  begin
  Result:=True;
  for I:=0 to 15 do
    if D1[I]<>D2[I] then
      begin
      Result:=False;
      Break;
      end;
  end;

function MD5Str(const M: AnsiString): AnsiString;
  begin
  Result:=rtcMD5Print(rtcMD5String(M));
  end;

function GetMD5(Buffer:pointer; Length:longword):AnsiString;
  var
  	Context: MD5Context;
    Res: MD5Digest;
  begin
	rtcMD5Init(Context);
	rtcMD5Update(Context, PAnsiChar(Buffer), Length);
	rtcMD5Final(Context, Res);
  Result:=rtcMD5Print(Res);
  end;

function MD5asGUID(const M: AnsiString): AnsiString;
  var
    s:AnsiString;
  begin
  s:=UpperCase(rtcMD5Print(rtcMD5String(M)));
  Result:=Copy(s,1,8)+'-'+Copy(s,9,4)+'-'+Copy(s,13,4)+'-'+Copy(s,17,4)+'-'+Copy(s,21,12);
  end;

function MD5Code(const M: AnsiString):AnsiString;
  var
  	Context: MD5Context;
    Digest: MD5Digest;
    a: integer;
    i: int64;
  const
	  Dig: array[0..31] of AnsiChar =
		  ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
       'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
       'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
       'u', 'v');
  begin
	rtcMD5Init(Context);
	rtcMD5Update(Context, @M[1], length(M));
	rtcMD5Final(Context, Digest);
  SetLength(Result,26);

  Move(Digest[0],i,8);
  for a:=1 to 13 do
    begin
    Result[a]:=Dig[i and $1f];
    i:=i shr 5;
    end;
  Move(Digest[8],i,8);
  for a:=14 to 26 do
    begin
    Result[a]:=Dig[i and $1f];
    i:=i shr 5;
    end;
  end;

function MD5Code6bit(const M: AnsiString):AnsiString;
  var
  	Context: MD5Context;
    Digest: MD5Digest;
    a: integer;
    i: int64;
  const
	  Dig: array[0..63] of AnsiChar =
		  ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
       'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
       'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
       'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
       'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
       'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
       'Y', 'Z', '_', '$');
  begin
	rtcMD5Init(Context);
	rtcMD5Update(Context, @M[1], length(M));
	rtcMD5Final(Context, Digest);
  SetLength(Result,22);

  Move(Digest[0],i,8);
  for a:=1 to 11 do
    begin
    Result[a]:=Dig[i and $3f];
    i:=i shr 6;
    end;
  Move(Digest[8],i,8);
  for a:=12 to 22 do
    begin
    Result[a]:=Dig[i and $3f];
    i:=i shr 6;
    end;
  end;

function MD5Code8bit(const M: AnsiString):AnsiString;
  var
  	Context: MD5Context;
    Digest: MD5Digest;
    a: integer;
  begin
	rtcMD5Init(Context);
	rtcMD5Update(Context, @M[1], length(M));
	rtcMD5Final(Context, Digest);
  SetLength(Result,16);
  
  for a:=1 to 16 do
    Result[a]:=AnsiChar(Digest[a-1]);
  end;

end.
