{
  RealThinClient SDK - UNIX support
  Copyright (c) Danijel Tkalcec
  @exclude
}

unit rtcUnixAsyncSock;

{$INCLUDE rtcDefs.inc}

interface

{ $define NixASockDebug}
uses
  Classes, SysUtils, SyncObjs, Sockets,
  BaseUnix, Unix, rtcUnixSockWrapper, rtcUnixSockUtils, rtcUnixWinSockets,
  rtcUnixWinAPI
  {$ifdef NixASockDebug}
  ,UnixDebug
  {$endif}
  ;

{!note
The socket created by the accept function has the same properties
as the listening socket used to accept it. Consequently, WSAAsyncSelect
events set for the listening socket also apply to the accepted socket.
For example, if a listening socket has WSAAsyncSelect events FD_ACCEPT,
FD_READ, and FD_WRITE, then any socket accepted on that listening socket
will also have FD_ACCEPT, FD_READ, and FD_WRITE events with the same wMsg
value used for messages. If a different wMsg or events are desired,
the application should call WSAAsyncSelect, passing the accepted socket and
the desired new data.
}

type

  { TAsyncWrapper }

  TAsyncWrapper = class(TSocketWrapper)
  private
    fLock       : TCriticalSection;
    fTaskEvents : TAsyncProcSet;
  protected
    procedure SetTaskEvents(AEvents: TAsyncProcSet); virtual;
  public
    isConnecting  : Boolean;
    isConnected   : Boolean; // socket has connected or has been accepted
    isAccepting   : Boolean; // server socket, waiting for incoming connections
    isClosed      : Boolean; // socket has been closed.
    WinHwnd       : Integer;
    wMsg, Event   : LongWord;
    Notifier      : TASyncNotifier;

    AutoUpdate    : Boolean;

    constructor Create;
    destructor Destroy; override;
    function Connect(const Addr: SockAddr; AddrSize: Integer): Integer; override;
    function Send(const Buf; BufSize, Flags: integer): Integer; override;
    function Recv(var Buf; BufSize, Flags: integer): Integer; override;
    function Shutdown(how: Integer): Integer; override;
    function Accept(addr: PSockAddr; addrlen: PInteger): TSocket; override;
    function Listen(backlog: Integer): Integer; override;
    function Close: Integer; override;
    property TaskEvents: TAsyncProcSet read fTaskEvents write SetTaskEvents;

    procedure UpdateManager;
    procedure Lock;
    procedure Unlock;
  end;


  { TWinMessageNotifier }

  TWinMessageNotifier = class(TASyncNotifier)
    procedure Notify(sk: TSocketManager; Event: TAsyncProc; SockError: Word); override;
  end;

  { TASyncManager }

  TASyncManager = class(TObject)
  private
    fPipe       : TPipe;
    fLock       : TCriticalSection;
    fPipeLock   : TCriticalSection;
    fItems      : TList;
    fStopped    : Boolean;
    fWritten    : Boolean;


  protected
    procedure FillDescriptors(var rd, wr, ex: TFDSet; var SelectFD: Integer);

    procedure CheckConnect(man: TSocketManager; a: TAsyncWrapper; const wr: TFDSet);
    procedure CheckAccept(man: TSocketManager; a: TAsyncWrapper; var rd: TFDSet);

    procedure MainLoop;

    procedure DoNotify(sk: TSocketManager; Event: TAsyncProc; SockError: Word); virtual;

    procedure DoStop;
    function GetStop: Boolean;

    procedure Lock;
    procedure Unlock;

  public
    constructor Create;
    destructor Destroy; override;

    procedure AddManager(Async: TSocketManager);
    procedure RemoveManager(Async: TSocketManager);
    procedure UpdateOptions;

  end;


  { TAsyncThread }

  TAsyncThread = class(TThread)
  protected
    procedure Execute; override;
  end;

function GetAsyncManager: TAsyncManager;
function GetNotifier: TASyncNotifier;

//{$ifdef NixASockDebug}
function DebugASyncEventStr(op: TAsyncProcSet): string;
//{$endif}

implementation


const
  ASyncProcStr : array [TAsyncProc] of string = (
    'AP_CONNECT', 'AP_CLOSE', 'AP_WRITE', 'AP_READ', 'AP_ACCEPT');

{$ifdef NixASockDebug}
var
  debugLVL : byte;
{$endif NixASockDebug}

function DebugASyncEventStr(op: TAsyncProcSet): string;
var
  a : TAsyncProc;
begin
  Result := '';
  for a := Low(TAsyncProc) to High(TAsyncProc) do
    if a in op then
      Result := Result + ASyncProcStr[a] + ' ';
end;



var
  manager : TASyncManager;
  thread  : TASyncThread;
  globalnotifier  : TWinMessageNotifier;

const
  //ap_Connect, ap_Clode, ap_Write, ap_Read, ap_Accept
  EventSocketMask : array [TAsyncProc] of Integer =
    (FD_CONNECT, FD_CLOSE, FD_WRITE, FD_READ,FD_ACCEPT);

procedure FinalizeThread;
begin
  try
    if Assigned(rtcUnixAsyncSock.manager) and Assigned(thread) then begin
      Manager.DoStop;
      thread.WaitFor;
      thread.Free;
      manager.Free;
      if Assigned(globalnotifier) then
        globalnotifier.free;
    end;
  except
  end;
end;


function GetNotifier: TASyncNotifier;
begin
  //modify TAsyncNotifier create, if any non Win based notification is to be used
  if not Assigned(globalnotifier) then
    globalnotifier := TWinMessageNotifier.Create;
  Result := globalnotifier;
end;


function GetAsyncManager: TAsyncManager;
begin
  if not Assigned(manager) then begin
    manager := TASyncManager.Create;
    thread := TAsyncThread.Create(false);
  end;
  Result := manager;
end;

{ TAsyncThread }

procedure TAsyncThread.Execute;
begin
  try
    manager.MainLoop;
  except
    {$ifdef NixASockDebug}
    on e: exception do
      DebugLine('error in MainLoop '+ e.message);
    {$endif}
  end;
end;

{ TASyncManager }

constructor TASyncManager.Create;
begin
  inherited Create;
  FpPipe(fPipe);
  fLock := TCriticalSection.Create;
  fPipeLock := TCriticalSection.Create;
  fItems := TList.Create;
end;

destructor TASyncManager.Destroy;
begin
  fItems.Free;
  fLock.Free;
  fPipeLock.Free;
  FpClose(fPipe[0]);
  FpClose(fPipe[1]);
  inherited Destroy;
end;

procedure TASyncManager.AddManager(Async: TSocketManager);
var
  i : Integer;
begin
  Lock;
  try
    i := fItems.IndexOf(ASync);
    if i < 0 then fItems.Add(ASync);
  finally
    Unlock;
  end;
  UpdateOptions;
end;

procedure TASyncManager.RemoveManager(Async: TSocketManager);
var
  i : integer;
begin
  Lock;
  try
    i := fItems.IndexOf(ASync);
    if i < 0 then Exit;
    fItems.Delete(i);
  finally
    Unlock;
  end;
  UpdateOptions;
end;

procedure TASyncManager.UpdateOptions;
var
  buf : Integer;
begin
  fPipeLock.Enter;
  try
    if not fWritten then begin
      buf := 1;
      FpWrite(fPipe[1], buf, sizeof(buf));
      fWritten := true;
    end;
  finally
    fPipeLock.Leave;
  end;
end;


procedure TASyncManager.FillDescriptors(var rd, wr, ex: TFDSet;
  var SelectFD: Integer);
var
  i : Integer;
  a : TAsyncWrapper;
  dbg : string;
begin

  SelectFD := fPipe[0];
  fpFD_ZERO(rd);
  fpFD_ZERO(wr);
  fpFD_ZERO(ex);
  fpFD_SET(fPipe[0], rd);

  Lock;
  try
    for i := 0 to fItems.Count - 1 do begin
      dbg := 'Wait for ';
      if not (TSocketManager(fItems[i]).Wrapper is TAsyncWrapper) then begin
        Continue;
      end;

      a := TAsyncWrapper(TSocketManager(fItems[i]).Wrapper);
      write('locking... ');
      a.Lock;
      try
        write('ed! ');
        if a.isClosed then Continue;

        write('Filling: ', a.Handle, ' ', DebugASyncEventStr(a.TaskEvents));
        write(' ');

        if (not a.isConnected) and a.isConnecting and (ap_Connect in a.TaskEvents) then begin
          fpFD_SET(a.Handle, wr);
          dbg := dbg + 'Connect, ';
        end else if (a.isAccepting) and (ap_Accept in a.TaskEvents) then begin
          {$ifdef NixASockDebug}DebugLine('[AsyncWrapper.Listen] Starting accept for socket = %d', [a.Handle], 0{debugLvl});{$endif}
          fpFD_Set(a.Handle, rd);
          dbg := dbg + 'Accept, ';
        end else if (a.isConnected) then begin
          //must always get read task, so it can be noted if read error appear?
          if (ap_Read in a.TaskEvents) {or (ap_Close in a.TaskEvents)} then begin
            fpFD_SET(a.Handle, rd);
            dbg := dbg + 'Read, ';
          end;
          if (ap_Write in a.TaskEvents) then begin
            fpFD_SET(a.Handle, wr);
            dbg := dbg + 'Write, ';
          end;

          if a.TaskEvents * [ap_Read, ap_Write] = [] then
            dbg := dbg + ' nothing';
        end else begin
          //does nothing
          dbg := dbg + ' nothing';
        end;
        if a.TaskEvents <> [] then begin
          fpFD_SET(a.Handle, ex);
          dbg := dbg + ' and Exception';
        end;

        if a.Handle > SelectFD then SelectFD := a.Handle;
      finally
        write(a.Handle, ': ', dbg);
        a.Unlock;
        //writeln(' unlocked');
      end;
    end; { of for}
  finally
    Unlock;
  end;
  inc(SelectFD);
end;

procedure TASyncManager.CheckConnect(man: TSocketManager; a: TAsyncWrapper; const wr: TFDSet);
var
  err : Integer;
begin
  {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] Calling fpFD_ISSET for Connect', debugLvl);{$endif}
  if fpFD_ISSET(a.Handle, wr) > 0 then begin
    //writeln(' fpgeterrno = ', fpgeterrno, ' for ', a.Handle);
    err := GetSocketError(a.Handle);
    a.isConnecting := false;
    a.isConnected := err = 0;
    {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] check connection for socket = %d; err = %d ', [a.Handle, err], debugLvl);{$endif}

    {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] Sending notificaiton', debugLvl);{$endif}
    DoNotify(man, ap_Connect, err);
    if not a.isConnected then begin
      //writeln('not connected! ', err);
      a.TaskEvents := [ap_Close];
    end else begin
      a.TaskEvents := a.TaskEvents - [ap_Connect];
      //if ap_Write in a.TaskEvents then DoNotify(man, ap_Write, err);
    end;
  end;
end;

procedure TASyncManager.CheckAccept(man: TSocketManager; a: TAsyncWrapper; var rd: TFDSet);
var
  err : Integer;
begin
 {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] Calling fpFD_ISSET for Accept', debugLvl);{$endif}
  if fpFD_ISSET(a.Handle, rd) > 0 then begin
    err := GetSocketError(a.Handle);
    a.TaskEvents := a.TaskEvents - [ap_Accept];
    {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] check accepting for socket = %d; err = %d ', [a.Handle, err], debugLvl);{$endif}

    {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] Sending notificaiton', debugLvl);{$endif}
    DoNotify(man, ap_Accept, err);
    fpFD_CLR(a.Handle, rd)
  end;
end;

procedure TASyncManager.MainLoop;
var
  rd, wr, ex: TFDSet;
  waitfd : integer;
  res : Integer;
  a   : TASyncWrapper;
  i   : Integer;
  err : Integer;
  buf : Integer;
  man : TSocketManager;
  b   : LongWord;
  rcv : Integer;
begin
  while not GetStop do begin
    FillDescriptors(rd, wr, ex, waitfd);
    {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] Selecting Socktes. Count = %d ', [fItems.Count], debugLvl, fItems.Count > 0);{$endif}
    res := fpSelect(waitfd, @rd, @wr, @ex, -1);
    Lock;
    try
      if fpFD_ISSET(fPipe[0], rd) > 0 then begin
        fLock.Enter;
        try
          FpRead(fPipe[0], buf, sizeof(buf));

          fPipeLock.Enter;
          try
            if fWritten then fWritten := false;
          finally
            fPipeLock.Leave;
          end;

          if GetStop then begin
            Exit;
          end;
        finally
          fLock.Leave;
        end;
        dec(res);
      end;
      if res = 0 then Continue;
      {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] Selected socktes count = %d ', [res], debugLvl);{$endif}

      for i := 0 to fItems.Count - 1 do begin
        man := TSocketManager(fItems[i]);

        //a := TSocketManager(fItems[i]);
        if not (man.Wrapper is TAsyncWrapper) then Continue;

        a := TAsyncWrapper(man.Wrapper);

        a.Lock;
        try
          if a.isClosed then Continue; // already notified!

          if a.isAccepting then
            CheckAccept(man, a, rd)
          else if not a.isConnected then begin
            {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] Checking connection', debugLvl);{$endif}
            CheckConnect(man, a, wr);
          end;

          if fpFD_ISSET(a.Handle, ex) > 0 then begin
            //DoNotify(TSocketManager(fItems[i]), ap_Close, GetSocketError(a.Handle));
            a.TaskEvents := [];
            {$ifdef NixASockDebug}DebugLine('!!!! [AsyncMainLoop] Socket %d Exception', [a.Handle], 1);{$endif}
          end;

          if fpFD_ISSET(a.Handle, wr) > 0 then begin
            DoNotify(man, ap_Write, GetSocketError(a.Handle));
            a.TaskEvents := a.TaskEvents - [ap_Write];
          end;

          if fpFD_ISSET(a.Handle, rd) > 0 then begin
            // for some reason. closed socket, did not returned in ex fdset, so
            // before notifying about arrived data, must check if anything is available
            rcv := fprecv(a.Handle, @b, sizeof(b), MSG_PEEK);
            if (rcv > 0) then begin
              //if (ap_Read in a.TaskEvents) then begin
             //   writeln('notifying READ ', a.Handle);
                DoNotify(man, ap_Read, GetSocketError(a.Handle));
              //end else
              //  writeln('CANNOT NOTIFY READ ', a.Handle);
              a.TaskEvents := a.TaskEvents - [ap_Read];
            end else begin
              //writeln('Error in read. socket ', a.Handle, ', Bytes Count (<= 0) = ',rcv, ' eerno = ', fpgeterrno);
              res := GetSocketError(a.Handle);
              if res <> ESysEAGAIN then begin
                // error! socket has been closed (remotely?)
                {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] Socket is closed %d. Notifying err = %d', [a.Handle, res], 1);{$endif}
                {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] Socket is closed %d. Notifying', [a.Handle], debugLvl);{$endif}
                if ap_Close in a.TaskEvents then
                  DoNotify(man, ap_Close, GetSocketError(a.Handle));
                a.TaskEvents := [];
                a.isConnected := false;
                a.isClosed := true;
                {$ifdef NixASockDebug}DebugLine('[AsyncMainLoop] Socket is closed %d. Notification complete', [a.Handle], 1);{$endif}
              end;
            end;
          end;

        finally
          a.Unlock;
        end;
      end; {of for... list of whatched socket managers}


    finally
      Unlock;
    end;

  end;
end;

procedure TASyncManager.DoNotify(sk: TSocketManager; Event: TAsyncProc; SockError: Word);
var
  a   : TAsyncWrapper;
begin
  if not (sk.Wrapper is TAsyncWrapper) then begin
    {$ifdef NixASockDebug}DebugLine('[ASync sock Notification Check] is not ASync wrapper', 1);{$endif}
    Exit;
  end;

  a := TAsyncWrapper(sk.Wrapper);
  if Assigned(a.Notifier) then
    a.Notifier.Notify(sk, Event, SockError)

  else
    {$ifdef NixASockDebug}DebugLine('[ASync sock Notification Check] not Notifier', 1){$endif};
end;

procedure TASyncManager.DoStop;
begin
  fLock.Enter;
  try
    fStopped := true;
  finally
    fLock.Leave;
  end;
  UpdateOptions;
end;

function TASyncManager.GetStop: Boolean;
begin
  fLock.Enter;
  try
    Result := fStopped;
  finally
    fLock.Leave;
  end;
end;

procedure TASyncManager.Lock;
begin
  fLock.Enter;
end;

procedure TASyncManager.Unlock;
begin
  fLock.Leave;
end;

{ TWinMessageNotifier }


procedure TWinMessageNotifier.Notify(sk: TSocketManager; Event: TAsyncProc; SockError: Word);
var
  idx : integer;
  lp : Integer;
begin
  //DebugLine('Posting notification: ' +ASyncProcStr[Event] + ' err = '+IntToStr(SockError));
{
  When one of the nominated network events occurs on the specified socket s,
  the application window hWnd receives message wMsg.

  The wParam parameter identifies the socket on which a network event has occurred.
  The low word of lParam specifies the network event that has occurred.
  The high word of lParam contains any error code.
    The error code be any error as defined in Winsock2.h.
}
  idx := GetIndexFromManager(sk);
  if idx >= 0 then begin
    lp := Integer( (SockError shl 16) ) or Integer(EventSocketMask[Event]);
    {$ifdef NixASockDebug}DebugLine('[ASync sock Notification Check] Posting message ' + ASyncProcStr[Event], 1);{$endif}
    //writeln('**** debug. Notify TAsyncWrapper = ', Integer(sk.Wrapper));
    PostMessage(TAsyncWrapper(sk.Wrapper).WinHwnd, TAsyncWrapper(sk.Wrapper).wMsg, idx,  lp );
   end else
    {$ifdef NixASockDebug}DebugLine('[ASync sock Notification Check] not index for Socket manager', 1){$endif};
end;


{ TAsyncWrapper }

procedure TAsyncWrapper.Lock;
begin
  //write('TAsyncWrapper.Lock');
  fLock.Enter;
end;

procedure TAsyncWrapper.Unlock;
begin
  //writeln('TAsyncWrapper.UnLock');
  fLock.Leave;
end;

procedure TAsyncWrapper.SetTaskEvents(AEvents: TAsyncProcSet);
begin
  Lock;
  try
    fTaskEvents := AEvents;
  finally
    Unlock;
  end;
end;

constructor TAsyncWrapper.Create;
begin
  inherited Create;
  fLock:= TCriticalSection.Create;
  AutoUpdate := true;
end;

destructor TAsyncWrapper.Destroy;
begin
  fLock.Free;
  inherited Destroy;
end;

function TAsyncWrapper.Connect(const Addr: SockAddr; AddrSize: Integer): Integer;
begin
  Lock;
  try
    isConnected := false;
    isConnecting := true;
    Include(fTaskEvents, ap_Connect);
    Result:=inherited Connect(Addr, AddrSize);
    //writeln('Async Connect result = ', Result, ' ', fpgeterrno);
    if AutoUpdate then UpdateManager;
  finally
    Unlock;
  end;
end;

function TAsyncWrapper.Send(const Buf; BufSize, Flags: integer): Integer;
begin
  Lock;
  try
    Include(fTaskEvents, ap_Write);
    Result:=inherited Send(Buf, BufSize, Flags);
    if AutoUpdate then UpdateManager;
  finally
    Unlock;
  end;
end;

function TAsyncWrapper.Recv(var Buf; BufSize, Flags: integer): Integer;
begin
  Lock;
  try
    Include(fTaskEvents, ap_Read);
    Result:=inherited Recv(Buf, BufSize, Flags);
    if AutoUpdate then UpdateManager;
  finally
    Unlock;
  end;
end;

function TAsyncWrapper.Shutdown(how: Integer): Integer;
begin
  Lock;
  try
    TaskEvents := [ap_Close];
    Result:=inherited Shutdown(how);
    UpdateManager;
  finally
    Unlock;
  end;
  //if AutoUpdate then GetAsyncManager.UpdateOptions; // the socket must callback by itself!
end;

function TAsyncWrapper.Accept(addr: PSockAddr; addrlen: PInteger): TSocket;
begin
  Lock;
  try
    include(fTaskEvents, ap_Accept);
    isAccepting:=true;

    Result:=inherited Accept(addr, addrlen);
    if AutoUpdate then UpdateManager;
  finally
    Unlock;
  end;
end;

function TAsyncWrapper.Listen(backlog: Integer): Integer;
begin
  Lock;
  try
    include(fTaskEvents, ap_Accept);
    isAccepting:= true;

    Result:=inherited Listen(backlog);
    {$ifdef NixASockDebug}DebugLine('[AsyncWrapper.Listen] Listern result = %d', [Result], debugLvl);{$endif}

    if AutoUpdate then UpdateManager;
  finally
    Unlock;
  end;
end;

function TAsyncWrapper.Close: Integer;
begin
  Lock;
  try
    fTaskEvents := [];
    Result := inherited Close;
    UpdateManager;
  finally
    Unlock;
  end;
end;

procedure TAsyncWrapper.UpdateManager;
begin
  GetAsyncManager.UpdateOptions;
end;

initialization
  {$ifdef NixASockDebug}
  debugLVL := RegisterDebugLVL;
  {$endif}

finalization
  FinalizeThread;
  {$ifdef NixASockDebug}
  DebugLine('...rtcUnixASockSock finalized');
  {$endif}


end.

