{
  RealThinClient SDK - UNIX support
  Copyright (c) Danijel Tkalcec
  @exclude
}

unit rtcUnixConsoleMsgQueue;

interface

uses
  Classes, SysUtils, SyncObjs;

type
  TMessage = class(TObject);

  TEventProc = procedure (Sender: TObject; msg: TMessage) of object;
  
  { TQueueTarget }

  TQueueTarget = class(TObject)
    Proc  : TEventProc;
    constructor Create(AProc: TEventProc);
  end;
  
  { TMessageInQueue }

  TMessageInQueue = class(TObject)
    msg     : TMessage;
    Target  : Integer;
    constructor Create(atarget: Integer; amsg: TMessage);
  end;
  
  { TMessageQueue }

  TMessageQueue = class(TObject)
  private
    fTargets    : TList;
    fQueue      : TList;
    fCanceled   : Boolean;
    
    fWait       : TEvent;
    fLock       : TCriticalSection;
  protected
    procedure ZeroTarget(Sender: TObject; msg: TMessage); virtual;
    procedure RemoveMessgesToTarget(id: Integer);
    
    procedure HandleMessage(Target: Integer; msg: TMessage; FreeMsg: boolean = false);
    
    procedure Lock;
    procedure Unlock;
  public
    OnPreHandle     : TEventProc;
    OnNoTargetEvent : TEventProc;
    
    constructor Create;
    destructor Destroy; override;
    
    function AddTarget(AProc: TEventProc): Integer;
    procedure ReleaseTarget(id: Integer);
    
    function PutMessage(Target: Integer; Msg: TMessage): Boolean;
    function PeekMessage(var Target: Integer; var Msg: TMessage): Boolean;
    procedure ProcNextEvent;

    function WaitForEvents(TimeOut: LongWord): Boolean;
    
    procedure CancelQueue;
  end;
  
implementation

{ TMessageQueue }

procedure TMessageQueue.ZeroTarget(Sender: TObject; msg: TMessage);
begin
  if Assigned(OnNOTargetEvent) then OnNoTargetEvent(Self, msg);
end;

procedure TMessageQueue.RemoveMessgesToTarget(id: Integer);
var
  i : integer;
  m : TMessageInQueue;
begin
  Lock;
  try
    for i := fQueue.Count - 1 downto 0 do begin
      m := TMessageInQueue(fQueue[i]);
      if Assigned(m) and (m.Target = id) then begin
        m.Free;
        fQueue.Delete(i);
      end;
    end;
  finally
    Unlock;
  end;
end;

procedure TMessageQueue.HandleMessage(Target: Integer; msg: TMessage; FreeMsg: boolean = false);
var
  qt    : TQueueTarget;
  proc  : TEventProc;
begin
  Lock;
  try
    if (Target < 0) or (Target >= fTargets.Count) then Exit;
    qt := TQueueTarget(fTargets[Target]);
    if Assigned(qt) then proc := qt.Proc
    else proc := nil;
  finally
    Unlock;
  end;
  if not Assigned(proc) then Exit;
  
  try
    proc(Self, msg);
  except
  end;
  
  if FreeMsg then msg.Free;
end;

procedure TMessageQueue.Lock;
begin
  fLock.Enter;
end;

procedure TMessageQueue.Unlock;
begin
  fLock.Leave;
end;

constructor TMessageQueue.Create;
begin
  inherited Create;
  fTargets := TList.Create;
  fTargets.Add( TQueueTarget.Create(ZeroTarget) );
  fQueue := TList.Create;
  fWait := TEvent.Create(nil, true, false, '');
  fLock := TCriticalSection.Create;
end;

destructor TMessageQueue.Destroy;
begin
  fLock.Free;
  fWait.Free;
  fTargets.Free;
  fQueue.Free;
  inherited Destroy;
end;

function TMessageQueue.AddTarget(AProc: TEventProc): Integer;
var
  i   : integer;
  trg : TQueueTarget;
begin
  Result := - 1;
  for i := 1 to fTargets.Count - 1 do
    if not Assigned(fTargets[i]) then begin
      Result := i;
      Break;
    end;
  if Result < 0 then
    Result := fTargets.Add ( TQueueTarget.Create(AProc))
  else begin
    trg := TQueueTarget.Create(AProc);
    fTargets[Result]:=trg;
  end;
end;

procedure TMessageQueue.ReleaseTarget(id: Integer);
begin
  if (id <= 0) or (id>=fTargets.Count) then Exit;
  RemoveMessgesToTarget(id);
  TQueueTarget(fTargets[id]).Free;
  fTargets[id] := nil;
end;

function TMessageQueue.PutMessage(Target: Integer; Msg: TMessage): Boolean;
var
  m : TMessageInQueue;
begin
  Result := not fCanceled;
  if not Result then Exit;
  
  m := TMessageInQueue.Create(Target, msg);
  Lock;
  try
    fQueue.Add(m);
  finally
    Unlock;
  end;
  fWait.SetEvent;
end;

function TMessageQueue.PeekMessage(var Target: Integer; var Msg: TMessage): Boolean;
var
  m  : TMessageInQueue;
begin
  if fQueue.Count <= 0 then begin
    Result := false;
    Exit;
  end;
  
  Lock;
  try
    m := TMessageInQueue(fQueue[0]);
    Target := m.Target;
    Msg := m.msg;
  finally
    Unlock;
  end;
end;

procedure TMessageQueue.ProcNextEvent;
var
  m  : TMessageInQueue;
begin
  if fQueue.Count <= 0 then Exit;

  Lock;
  try
    m := TMessageInQueue(fQueue[0]);
  finally
    Unlock;
  end;

  try
    HandleMessage(m.Target, m.msg);
  except
  end;

  Lock;
  try
    fQueue.Delete(0);
  finally
    Unlock;
    m.msg.Free;
    m.Free;
  end;
end;

function TMessageQueue.WaitForEvents(TimeOut: LongWord): Boolean;
begin
  Result := fWait.WaitFor(TimeOut) = wrSignaled;
  Lock;
  try
    Result := (fQueue.Count > 0) and not fCanceled;
  finally
    Unlock;
  end;
end;

procedure TMessageQueue.CancelQueue;
begin
  Lock;
  try
    fCanceled := true;
    fWait.SetEvent;
  finally
    Unlock;
  end;
end;

{ TMessageInQueue }

constructor TMessageInQueue.Create(atarget: Integer; amsg: TMessage);
begin
  msg:=amsg;
  Target:= atarget;
end;

{ TQueueTarget }

constructor TQueueTarget.Create(AProc: TEventProc);
begin
  Proc := AProc;
end;

end.

