{
  RealThinClient SDK - UNIX support
  Copyright (c) Danijel Tkalcec
  @exclude
}

unit rtcUnixMsgQueue;
// this is base module to organize custom message queue

{$include rtcDefs.inc}

interface

uses
{$IFDEF UNIX}
  cthreads,
{$ENDIF}
  Classes, SysUtils, SyncObjs;

type
  TEventMessage = class(TObject);

  TEventProc = procedure (Sender: TObject; msg: TEventMessage) of object;
  
  TMessageQueue = class;
  
  TQueueResult = (qr_Success, qr_False, qr_QueueClosed);

  { TQueueTarget }

  TQueueTarget = class(TObject)
    fProc   : TEventProc;
    fOwner  : TMessageQueue;
  public
    constructor Create(AOwner: TMessageQueue);
    property Owner: TMessageQueue read fOwner;
    property Proc: TEventProc read fProc write fProc;
  end;
  
  { TMessageInQueue }

  TMessageInQueue = class(TObject)
    msg     : TEventMessage;
    Target  : TQueueTarget;
    OwnMsg  : Boolean;
    constructor Create(atarget: TQueueTarget; amsg: TEventMessage; AOwnMsg: Boolean = true);
    destructor Destroy; override;
  end;
  
  { TMessageQueue }

  TMessageQueue = class(TObject)
  private
    fTargets    : TList;
    fQueue      : TList;
    fCanceled   : Boolean;
    fProcRest   : Boolean;
    
    fWait       : TEvent;
    fLock       : TCriticalSection;
  protected
    procedure RemoveMessgesToTarget(atarget : TQueueTarget);
    
    procedure HandleMessage(Target: TQueueTarget; msg: TEventMessage; FreeMsg: boolean = false); virtual;
    function AllocTarget(AProc: TEventPRoc): TQueueTarget; virtual;
    
    procedure Lock;
    procedure Unlock;


    function OpenStatus(ForRead: Boolean) : TQueueResult;
  public
    OnPreHandle     : TEventProc;
    OnNoTargetEvent : TEventProc;
    
    constructor Create;
    destructor Destroy; override;
    
    function AddTarget(AProc: TEventProc): TQueueTarget;
    procedure ReleaseTarget(var ATarget: TQueueTarget);
    
    function PutMessage(Target: TQueueTarget; Msg: TEventMessage): TQueueResult;

    function PeekMessage(var Target: TQueueTarget; var Msg: TEventMessage; RemoveFromQueue: Boolean): TQueueResult;

    procedure ProcNextEvent;
    function RemoveNextEvent: Boolean;
    procedure ProcCustomEvent(Target: TQueueTarget; msg: TEventMessage; FreeMsg: Boolean);

    function WaitForEvents(TimeOut: Integer): TQueueResult;
    
    procedure CancelQueue( DeletePending: Boolean);
  end;
  
implementation

{ TMessageQueue }

procedure TMessageQueue.RemoveMessgesToTarget(atarget : TQueueTarget);
var
  i : integer;
  m : TMessageInQueue;
begin
  Lock;
  try
    for i := fQueue.Count - 1 downto 0 do begin
      m := TMessageInQueue(fQueue[i]);
      if Assigned(m) and (m.Target = atarget) then begin
        m.Free;
        fQueue.Delete(i);
      end;
    end;
  finally
    Unlock;
  end;
end;

procedure TMessageQueue.HandleMessage(Target: TQueueTarget; msg: TEventMessage; FreeMsg: boolean = false);
var
  proc  : TEventProc;
begin
  Lock;
  try
    if not Assigned(msg) or not Assigned(target) then
      target := TQueueTarget(fTargets[0]);
    proc := Target.Proc;
  finally
    Unlock;
  end;

  try
    proc(Self, msg);
  except
  end;
  if FreeMsg then msg.Free;
end;

function TMessageQueue.AllocTarget(AProc: TEventProc): TQueueTarget;
begin
  Result := TQueueTarget.Create(Self);
  Result.Proc := AProc;
end;

procedure TMessageQueue.Lock;
begin
  fLock.Enter;
end;

procedure TMessageQueue.Unlock;
begin
  fLock.Leave;
end;

function TMessageQueue.OpenStatus(ForRead: Boolean): TQueueResult;
begin
  Lock;
  try
    if fCanceled then begin
      Result := qr_QueueClosed;
      if fQueue.Count > 0 then Result := qr_Success;
    end else
      Result := qr_Success;
  finally
    Unlock;
  end;
end;

constructor TMessageQueue.Create;
begin
  inherited Create;
  fTargets := TList.Create;
  fQueue := TList.Create;
  fWait := TEvent.Create(nil, true, false, '');
  fWait.ResetEvent;

  fLock := TCriticalSection.Create;
end;

destructor TMessageQueue.Destroy;
begin
  fLock.Free;
  fWait.Free;
  fTargets.Free;
  fQueue.Free;
  inherited Destroy;
end;

function TMessageQueue.AddTarget(AProc: TEventProc): TQueueTarget;
begin
  Lock;
  try
    if OpenStatus(false) <> qr_Success then begin
      Result := nil;
      Exit;
    end;

    Result := AllocTarget(AProc);
    fTargets.Add(Result);
  finally
    Unlock;
  end;
end;

procedure TMessageQueue.ReleaseTarget(var atarget: TQueueTarget);
var
  i   : integer;
begin
  Lock;
  try
    i := fTargets.IndexOf(atarget);
    if i < 0 then Exit;
    RemoveMessgesToTarget(atarget);
    atarget.Free;
    atarget := nil;
    fTargets.Delete(i);
  finally
    Unlock;
  end;
end;

function TMessageQueue.PutMessage(Target: TQueueTarget;Msg: TEventMessage): TQueueResult;
var
  m : TMessageInQueue;
begin
  if not Assigned(msg) then begin
    Result := qr_False;
    Exit;
  end;

  Result := OpenStatus(false);
  if Result <> qr_Success then Exit;
  Lock;
  try
    Result := OpenStatus(false);
    if Result <> qr_Success then Exit;

    m := TMessageInQueue.Create(Target, msg);
    fQueue.Add(m);
    fWait.SetEvent;
  finally
    Unlock;
  end;
end;

function TMessageQueue.PeekMessage(var Target: TQueueTarget; var Msg: TEventMessage;
  RemoveFromQueue: Boolean): TQueueResult;
var
  m  : TMessageInQueue;
begin
  Lock;
  try
    if fQueue.Count <= 0 then begin
      Result := qr_False;
      if fCanceled
        then Result := qr_QueueClosed
        else fWait.ResetEvent;
      Exit;
    end;
  
    Result := qr_Success;
    m := TMessageInQueue(fQueue[0]);
    Target := m.Target;
    Msg := m.msg;


    if RemoveFromQueue then begin
      m.OwnMsg := false;
      RemoveNextEvent;
    end;
  finally
    Unlock;
  end;
end;

function TMessageQueue.RemoveNextEvent: Boolean;
begin
  Lock;
  try
    Result := fQueue.Count > 0;
    if not Result then Exit;

    TMessageInQueue(fQueue[0]).OwnMsg:=false;
    TMessageInQueue(fQueue[0]).Free;
    fQueue.Delete(0);
  finally
    Unlock;
  end;

end;

procedure TMessageQueue.ProcNextEvent;
var
  m  : TEventMessage;
  t  : TQueueTarget;
begin
  Lock;
  try
    if fQueue.Count <= 0 then Exit;
    if not (PeekMessage(t, m, true) = qr_Success) then Exit;
  finally
    Unlock;
  end;
  ProcCustomEvent(t, m, true);
end;

procedure TMessageQueue.ProcCustomEvent(Target: TQueueTarget; msg: TEventMessage; FreeMsg: Boolean);
begin
  HandleMessage(Target, msg, FreeMsg);
end;

function TMessageQueue.WaitForEvents(TimeOut: Integer): TQueueResult;
begin
  Result := OpenStatus(true);
  if Result <> qr_Success then Exit;

  if fWait.WaitFor(TimeOut) = wrSignaled then begin
    Lock;
    try
      Result := OpenStatus(true);
    finally
      Unlock;
    end;
  end else
    Result := qr_False;
end;

procedure TMessageQueue.CancelQueue(DeletePending: Boolean);
var
  i : integer;
begin
  Lock;
  try
    fCanceled := true;
    fProcRest := not DeletePending;
    if DeletePending then begin
      for i := 0 to fQueue.Count - 1 do
        TMessageInQueue(fQueue[i]).Free;
      fQueue.Clear;
    end;

    fWait.SetEvent;
  finally
    Unlock;
  end;
end;

{ TMessageInQueue }

constructor TMessageInQueue.Create(atarget: TQueueTarget; amsg: TEventMessage; AOwnMsg : Boolean);
begin
  msg:=amsg;
  Target:= atarget;
  OwnMsg := AOwnMsg;
end;

destructor TMessageInQueue.Destroy;
begin
  if OwnMsg then msg.Free;
  inherited Destroy;
end;

{ TQueueTarget }

constructor TQueueTarget.Create(AOwner: TMessageQueue);
begin
  fOwner := AOwner;
end;

end.

