{
  RealThinClient SDK - UNIX support
  Copyright (c) Danijel Tkalcec
  @exclude
}

unit rtcUnixSockWrapper;


interface

uses
  Sockets, BaseUnix, Unix, rtcUnixSockUtils, SyncObjs;
  
type
  { TSocketWrapper }

  TSocketWrapper = class(TObject)
  public
    Handle  : Integer; // actuall system socket
    SocketID  : Integer; // socket index in the socket managers pool
    constructor Create;
    function Connect(const Addr: SockAddr; AddrSize: Integer): Integer; virtual;
    function Send(const Buf; BufSize, Flags: integer): Integer; virtual;

    function SendTo(var Buf; len, flags: Integer; var addrto: TSockAddr; tolen: Integer): Integer; virtual;

    function Recv(var Buf; BufSize, Flags: integer): Integer; virtual;
    function RecvFrom(var Buf; len, flags: Integer; var from: TSockAddr; var fromlen: Integer): Integer; virtual;
    function Shutdown(how: Integer): Integer; virtual;

    function Close: Integer; virtual;

    function Listen(backlog: Integer): Integer; virtual;
    function Bind(var addr: TSockAddr; namelen: Integer): Integer; virtual;
    function Accept(addr: PSockAddr; addrlen: PInteger): TSocket;virtual;

    function IoctlSocket(cmd: DWORD; var arg: LongWord): Integer; virtual;


    function SetSockOpt(level, optname: Integer; optval: PChar; optlen: Integer): Integer; virtual;
    function GetSockOpt(level, optname: Integer; optval: PChar; var optlen: Integer): Integer; virtual;

  end;
  
  { TAsyncWrapper }

  TAsyncProc = (ap_Connect, ap_Close, ap_Write, ap_Read, ap_Accept);
  TAsyncProcSet = set of TAsyncProc;

  TSocketManager = class;

  TASyncNotifier = class(TObject)
  public
    procedure Notify(sk: TSocketManager; Event: TAsyncProc; SockError: Word); virtual; abstract;
  end;

  { TSocketManager }

  TSocketManager = class(TObject)
  private
    Index : Integer;
  protected
    fHandle  : Integer;
    fWrapper  : TSocketWrapper;
    procedure SetWrapper(AWrapper: TSocketWrapper);
    procedure SetHandle(AHandle: Integer);
  public
    isListen  : Boolean;
    isBusy  : Boolean;
    property Wrapper : TSocketWrapper read fWrapper write SetWrapper;
    property Handle  : integer read fHandle write SetHandle;
  end;

function CreatePoolManager: TSocketManager;
procedure ReleasePoolManager(Manager: TSocketManager);

function GetManagerFromIndex(s: Integer): TSocketManager;
function GetIndexFromManager(man: TSocketManager): Integer;

implementation

var
  ManagersPool : array of TSocketManager;
  MPLock   : TCriticalSection;


function GetIndexFromManager(man: TSocketManager): Integer;
begin
  if not Assigned(man) then Result := -1
  else Result := man.Index+1;
end;

function GetManagerFromIndex(s: Integer): TSocketManager;
var
  i   : Integer;
begin
  MPLock.Enter;
  try
    i := Integer(s-1);
    if (i < 0) or (i >= length(ManagersPool)) then
      Result := nil
    else begin
      Result := ManagersPool[i];
      if Assigned(Result) and (not Result.isBusy) then
        Result := nil;
    end;
  finally
    MPLock.Leave;
  end;
end;

function CreatePoolManager: TSocketManager;
var
  i   : Integer;
begin
  MPLock.Enter;
  try
    Result := nil;
    for i := 0 to length(ManagersPool) - 1 do
      if Assigned(ManagersPool[i]) and (not ManagersPool[i].isBusy) then begin
        Result := ManagersPool[i];
        Break;
      end else if not Assigned(ManagersPool[i]) then begin
        Result := TSocketManager.Create;
        Result.Index := i;
        ManagersPool[i] := Result;
        Break;
      end;
    if not Assigned(Result) then begin
      i := length(ManagersPool);
      if i = 0
        then SetLength(ManagersPool, 16)
        else SetLength(ManagersPool, i*2);
      Result := TSocketManager.Create;
      Result.Index := i;
      ManagersPool[i] := Result;
    end;
    Result.isBusy := true;

  finally
    MPLock.Leave;
  end;
end;

procedure ReleasePoolManager(Manager: TSocketManager);
begin
  MPLock.Enter;
  try
    Manager.isBusy := false;
    Manager.Wrapper.Free;
  finally
    MPLock.Leave;
  end;
end;

procedure InitPool;
begin
  MPLock := TCriticalSection.Create;
end;

procedure ClearPool;
var
  i   : integer;
begin
  for i := 0 to length(ManagersPool) - 1 do
    if Assigned(ManagersPool[i]) then
      ManagersPool[i].Free;
  MPLock.Free;
end;


{ TSocketWrapper }

constructor TSocketWrapper.Create;
begin
  inherited Create;
  Handle := -1;
end;

function TSocketWrapper.Connect(const Addr: SockAddr; AddrSize: Integer): integer;
begin
  Result := fpconnect(Handle, @Addr, AddrSize);
end;

function TSocketWrapper.Send(const Buf; BufSize, Flags: integer): Integer;
begin
  Result := fpsend(Handle, @buf, BufSize, Flags);
end;

function TSocketWrapper.Recv(var Buf; BufSize, Flags: integer): Integer;
begin
  Result := fprecv(Handle, @Buf, BufSize, Flags);
end;


function TSocketWrapper.SendTo(var Buf; len, flags: Integer;
  var addrto: TSockAddr; tolen: Integer): Integer;
begin
  Result := fpsendto(Handle, @Buf, len, flags, @addrto, tolen);
end;

function TSocketWrapper.RecvFrom(var Buf; len, flags: Integer;
  var from: TSockAddr; var fromlen: Integer): Integer;
begin
  Result := fprecvfrom(Handle, @buf, len, flags, @from, @fromlen);
end;

function TSocketWrapper.Shutdown(how: Integer): Integer;
begin
  Result := fpshutdown(Handle, how);
end;

function TSocketWrapper.Close: Integer;
begin
  Result := CloseSocket(Handle);
end;

function TSocketWrapper.Listen(backlog: Integer): Integer;
begin
  Result := fplisten(Handle, backlog);
end;

function TSocketWrapper.Bind(var addr: TSockAddr; namelen: Integer): Integer;
begin
  Result := fpbind(Handle, @addr, namelen);
end;

function TSocketWrapper.Accept(addr: PSockAddr; addrlen: PInteger): TSocket;
begin
  Result := fpaccept(Handle, addr, PSockLen(addrlen));
end;

function TSocketWrapper.IoctlSocket(cmd: DWORD; var arg: LongWord): Integer;
begin
  Result := FpIOCtl(Handle, cmd, @arg);
end;


function TSocketWrapper.SetSockOpt(level, optname: Integer; optval: PChar;
  optlen: Integer): Integer;
begin
  Result := fpsetsockopt(Handle, level, optname, optval, optlen);
end;

function TSocketWrapper.GetSockOpt(level, optname: Integer; optval: PChar;
  var optlen: Integer): Integer;
begin
  Result := fpgetsockopt(Handle, level, optname, optval, @optlen);
end;


{ TSocketManager }

procedure TSocketManager.SetWrapper(AWrapper: TSocketWrapper);
begin
  fWrapper := AWrapper;
  if Assigned(fWrapper) then fWrapper.Handle := fHandle;
end;

procedure TSocketManager.SetHandle(AHandle: Integer);
begin
  fHandle := AHandle;
  if Assigned(fWrapper) then
    fWrapper.Handle := fHandle;
end;

initialization
  InitPool;

finalization
  ClearPool;

end.

