{
  RealThinClient SDK - UNIX support
  Copyright (c) Danijel Tkalcec
  @exclude
}

unit rtcUnixWinMsgQueue;
// this is Windows-like message queue, derived from nixMsgQueue

{$include rtcDefs.inc}

interface

uses
{$IFDEF UNIX}
  cthreads,
{$ENDIF}
  Classes, SysUtils, SyncObjs, rtcUnixMsgQueue, rtcUnixWinCommon;

type
  TWinProc = procedure (Whnd: Integer; MsgID, wParam, lParam: LongWord) of object;

  { TWinMessage }

  TWinMessage = class(TEventMessage)
    MsgID: LongWord;
    wParam, lParam:LongWord;
    Time  : LongWord;
    constructor Create(AID, AWParam, ALParam: LongWord);
  end;

  { TWinQueueTarget }

  TWinQueueTarget = class(TQueueTarget)
  public
    WinProc : TWinProc;
    TagObj  : TObject;
    procedure WinTargetProc(Sender: TObject; msg: TEventMessage);
  end;

  { TWinMessageQueue }

  TWinMessageQueue = class(TMessageQueue)
  protected
    function AllocTarget(AProc: TEventPRoc): TQueueTarget; override;
  public
    ThreadID: Integer;
    DefaultTargetHandle: Integer;
    constructor Create;
    destructor Destroy; override;
  end;

function winmsgCreateTarget(WinProc: TWinProc): Integer;
procedure winmsgCloseTarget(Handle: Integer);

function winmsgGetTagObject(Handle: Integer):TObject;
procedure winmsgSetTagObject(Handle: Integer; obj: TObject);


function winmsgPostThreadMessage(idThread: LongWord; msg: LongWord; wParam, lParam: LongWord): Boolean;
function winmsgPostMessage(Hwnd: Integer; msg: LongWord; wParam, lParam: LongWord): Boolean;

function winmsgPeekMessage(var msg: TMsgStruct; hwnd: Integer; RemoveFromQueue: Boolean): Boolean;
function winmsgGetMessage(var msg: TMsgStruct; hwnd: Integer): Boolean;
function winmsgDispatchMessage(const msg: TMsgStruct): Boolean;

procedure winmsgMainLoop;

//?? remove from interface
function GetThreadQueue(forced: Boolean = true): TMessageQueue;
procedure RemoveThreadQueue;
procedure CloseThreadQueue;

implementation

var
  // using threadvar instead?
  Queues      : TList;
  QueuesLock  : TCriticalSection;
  TrgPool     : TList;
  TrgPoolLock : TCriticalSection;

{function HandleToTarget(idx: Integer): TQueueTarget; forward;
function TargetToHandle(Target: TQueueTarget): Integer;forward;
procedure ReleaseHandle(Target: TQueueTarget); forward;}

function TargetToHandle(Target: TQueueTarget): Integer;
var
  i     : integer;
  empt  : Integer;
  t     : TQueueTarget;
begin
  TrgPoolLock.Enter;
  try
    //Result := TrgPool.IndexOf(Target);
    empt := -1;
    for i := 0 to TrgPool.Count - 1 do begin
      t := TQueueTarget( TrgPool[i] );
      if not Assigned(t) and (empt < 0) then
        empt := i
      else if t = Target then begin
        Result := i+1;
        Exit;
      end;
    end;

    if empt < 0 then
      Result := TrgPool.Add(Target)
    else begin
      TrgPool[empt] := Target;
      Result := empt;
    end;
    inc(Result);

  finally
    TrgPoolLock.Leave;
  end;
end;

function HandleToTarget(idx: Integer): TQueueTarget;
begin
  TrgPoolLock.Enter;
  try
    dec(idx);

    if (idx < 0) or (idx >= TrgPool.Count) then Result := nil
    else Result := TQueueTarget(TrgPool[idx]);
  finally
    TrgPoolLock.Leave;
  end;
end;

procedure ReleaseHandle(Target: TQueueTarget);
var
  idx : Integer;
begin
  TrgPoolLock.Enter;
  try
    idx := TrgPool.IndexOf(Target);
    if (idx >= 0) and (idx < TrgPool.Count) then
      TrgPool[idx] := nil;
  finally
    TrgPoolLock.Leave;
  end;
end;


function GetThreadQueueAndTarget(Hwnd: Integer; var Queue: TMessageQueue; var Target: TQueueTarget): Boolean;
begin
  Result := hwnd > 0;
  if not Result then Exit;

  Target := HandleToTarget(hwnd);
  Result := Assigned(Target);
  if not Result then Exit;

  Result := GetThreadQueue(false) = Target.Owner;
  if not Result then Exit;
  Queue := Target.Owner;
end;

function winmsgDispatchMessage(const msg: TMsgStruct): Boolean;
var
  trg : TQueueTarget;
  wm  : TWinMessage;
  q   : TMessageQueue;
begin
  Result := false;
  if msg.Wnd <> 0 then begin
    trg := HandleToTarget(msg.Wnd);
    if (not Assigned(trg)) then Exit;
  end else begin
    q := GetThreadQueue(false);
    if not Assigned(q) then Exit;
    trg := HandleToTarget( TWinMessageQueue(q).DefaultTargetHandle );
  end;

  wm := TWinMessage.Create(msg.message, msg.wParam, msg.lParam);
  try
    wm.Time := msg.time;
    Result := true;
    trg.Owner.ProcCustomEvent(trg, wm, true);
  finally
  end;

end;

function winmsgGetMessage(var msg: TMsgStruct; hwnd: Integer): Boolean;
var
  q : TMessageQueue;
  t : TQueueTarget;
  stop  : Boolean;
begin
  if hwnd > 0 then begin
    Result := GetThreadQueueAndTarget(hwnd, q, t);

    if not Result then Exit;
  end else begin
    // queue must be created by calling PeekThreadMessage, first!
    q := nil;
    t := nil;
  end;

  stop := false;
  while not stop do begin
    Result := winmsgPeekMessage(msg, hwnd, true);
    if not Result  then begin
      if not Assigned(q) then begin
        q := GetThreadQueue(false);
        if not assigned(q) then Exit;
      end;
      if q.WaitForEvents( Integer(INFINITE)) = qr_QueueClosed then begin
        Result := false;
        Exit;
      end;
    end else
      Stop := true;
  end;
end;

function winmsgPeekMessage(var msg: TMsgStruct; hwnd: Integer; RemoveFromQueue: Boolean): Boolean;
var
  t : TQueueTarget;   // wanted target
  q : TMessageQueue;
  m : TEventMessage;
  wm  : TWinMessage;
  tm  : TQueueTarget; // message target
  queuemsg : Boolean; // process queue messge only
  stop     : Boolean;
begin
  Result := false;
  if hwnd > 0 then begin
    queuemsg := false;
    if not GetThreadQueueAndTarget(hwnd, q, t) then Exit;
  end else begin
    queuemsg := hwnd < 0;
    q := GetThreadQueue(true);
    t := nil;
  end;

  stop := false;
  while not stop do begin
    Result := q.PeekMessage(tm, m, false) = qr_Success;

    if not Result then begin
      stop := true;
    end else if not (m is TWinMessage) then
      q.ProcNextEvent
    else begin
      wm := TWinMessage(m);


      if Assigned(tm) then begin // there's target for next queue
        msg.wnd := TargetToHandle(tm);
        if not Assigned(t) and not (queuemsg) then begin // PeekMsg(msg, 0, ..)
          Result := true;
          stop := true;
        end else if (queuemsg) or (t <> tm) then begin // PeekMsg(msg, -1, ..)
          q.ProcNextEvent;
        end else begin                    // PeekMsg(msg, hwd, ..)
          msg.Wnd := TargetToHandle(tm);
          stop := true;
        end;
      end else if Assigned(t) then
        q.ProcNextEvent
      else begin
        msg.Wnd := 0;
        stop := true;
      end;

      if stop and Assigned(wm) then begin
        msg.time := wm.Time;
        msg.message := wm.MsgID;
        msg.wParam := wm.wParam;
        msg.lParam := wm.lParam;
        msg.pt.x := 0;
        msg.pt.y := 0;
        if RemoveFromQueue then begin
          q.RemoveNextEvent;
        end;
        Result := true;
        wm.Free;
      end else if not Assigned(wm) then
        Result := false;
    end;

  end; {of while}

end;

  
function FindMsgQueue(threadid: Integer): Integer;
var
  i  : Integer;
begin
  QueuesLock.Enter;
  try
    Result := -1;
    for i := 0 to Queues.Count - 1 do
      if TWinMessageQueue(Queues[i]).ThreadID = threadid then begin
        Result := i;
        Exit;
      end;
  finally
    QueuesLock.Leave;
  end;
end;

procedure winmsgMainLoop;
var
  msg : TMsgStruct;
begin
  try
    FillChar(msg, sizeof(msg), 0);
    while winmsgGetMessage(msg, 0) do begin
      winmsgDispatchMessage(msg);
    end;
  except
  end;
end;

function GetThreadQueue(forced: Boolean): TMessageQueue;
var
  i   : integer;
  id  : integer;
  q   : TWinMessageQueue;
begin
  Result := nil;
  id := Integer(GetCurrentThreadID);
  QueuesLock.Enter;
  try
    i := FindMsgQueue(id);
    if i < 0 then begin
      if not forced then begin
        Exit;
      end;
      q := TWinMessageQueue.Create;
      q.ThreadID := id;
      Result := q;
      Queues.Add(Result);
    end else
      Result := TMessageQueue(Queues[i]);
  finally
    QueuesLock.Leave;
  end;
end;

procedure RemoveThreadQueue;
var
  id  : Integer;
  i   : Integer;
begin
  id := Integer(GetCurrentThreadID);
  QueuesLock.Enter;
  try
    i := FindMsgQueue(id);
    if i < 0 then Exit;
    TWinMessageQueue(Queues[i]).Free;
    Queues.Delete(i);
  finally
    QueuesLock.Leave;
  end;
end;

{function PostWinMessage(Queue: TMessageQueue; Target: TQueueTarget; msg: LongWord; wParam, lParam: LongWord): Boolean;
begin
  Result := Assigned(Queue);
  Queue.PutMessage(Target, TWinMessage.Create(msg, wParam, lParam));
end;}


function winmsgPostThreadMessage(idThread: LongWord; msg: LongWord; wParam, lParam: LongWord): Boolean;
var
  q : TMessageQueue;
  idx : integer;
begin
  QueuesLock.Enter;
  try
    idx := FindMsgQueue(idThread);
    Result := idx >= 0;
    if not Result then Exit;

    q := TMessageQueue(Queues[idx]);
    Result := Assigned(q) and (q.PutMessage(nil, TWinMessage.Create(msg, wParam, lParam))=qr_Success);
  finally
    QueuesLock.Leave;
  end;
end;

function winmsgPostMessage(Hwnd: Integer; msg: LongWord; wParam, lParam: LongWord): Boolean;
var
  trg : TQueueTarget;
  q   : TMessageQueue;
begin
  trg := HandleToTarget(Hwnd);
  Result := Assigned(trg);
  if not Result then Exit;

  q := trg.Owner;
  Result := Assigned(q) and (q.PutMessage(trg, TWinMessage.Create(msg, wParam, lParam))=qr_Success);
end;

function winmsgGetTagObject(Handle: Integer):TObject;
var
  trg : TWinQueueTarget;
begin
  trg := TWinQueueTarget(HandleToTarget(Handle));
  if not Assigned(trg) then Exit;
  Result := trg.TagObj;
end;

procedure winmsgSetTagObject(Handle: Integer; obj: TObject);
var
  trg : TWinQueueTarget;
begin
  trg := TWinQueueTarget(HandleToTarget(Handle));
  if not Assigned(trg) then Exit;
  trg.TagObj := obj;
end;

function winmsgCreateTarget(WinProc: TWinProc): Integer;
var
  q : TMessageQueue;
  wt  : TWinQueueTarget;
begin
  Result := 0;
  q := GetThreadQueue(true);
  if not Assigned(q) or not (q is TWinMessageQueue) then Exit;

  wt := TWinQueueTarget(q.AddTarget(nil));
  wt.Proc := wt.WinTargetProc;
  wt.WinProc := WinProc;
  Result := TargetToHandle(wt);
end;

procedure winmsgCloseTarget(Handle: Integer);
var
  trg : TQueueTarget;
begin
  trg := HandleToTarget(Handle);
  if not Assigned(trg) then Exit;

  ReleaseHandle(trg);
  trg.Owner.ReleaseTarget(trg);
end;

procedure CloseThreadQueue;
var
  q : TMessageQueue;
begin
  q:= GetThreadQueue(false);
  if not Assigned(q) then Exit;
  q.CancelQueue(false);
end;


procedure InitQueues;
begin
  Queues := TList.Create;
  QueuesLock := TCriticalSection.Create;
  TrgPool := TList.Create;
  TrgPoolLock := TCriticalSection.Create;

end;

procedure ReleaseQueues;
begin
  Queues.Free;
  QueuesLock.Free;
  TrgPool.Free;
  TrgPoolLock.Free;
end;

{ TWinMessage }

constructor TWinMessage.Create(AID, AWParam, ALParam: LongWord);
begin
  MsgID := AID;
  WParam := AWParam;
  LParam := ALParam;
end;

{ TWinQueueTarget }

procedure TWinQueueTarget.WinTargetProc(Sender: TObject; msg:TEventMessage);
var
  win : TWinMessage;
begin
  if not (msg is TEventMessage) then Exit;
  win := TWinMessage(msg);
  if ASsigned(winProc) then
    WinProc(TargetToHandle(Self), win.MsgID, win.wParam, win.lParam);
end;

{ TWinMessageQueue }

function TWinMessageQueue.AllocTarget(AProc: TEventPRoc): TQueueTarget;
begin
  Result:=TWinQueueTarget.Create(Self);
  Result.Proc := TWinQueueTarget(Result).WinTargetProc;
end;

constructor TWinMessageQueue.Create;
var
  trg  : TWinQueueTarget;
begin
  inherited Create;
  trg := AllocTarget(nil) as TWinQueueTarget;
  trg.Proc := trg.WinTargetProc;
  DefaultTargetHandle := TargetToHandle(trg);
end;

destructor TWinMessageQueue.Destroy;
begin
  inherited Destroy;
end;


initialization
  InitQueues;

finalization
  ReleaseQueues;
  
end.

