{
  "HTTP Client Provider (WinSock)" - Copyright (c) Danijel Tkalcec
  @html(<br>)

  Using TRtcWSockClientProvider to implement a HTTP Client provider

  @exclude
}
unit rtcWSockHttpCliProv;

{$INCLUDE rtcDefs.inc}

interface

uses
  Classes,
  SysUtils,

  rtcInfo,
  rtcLog,
  rtcConn,
  rtcConnProv,

  rtcFastStrings,
  rtcWSockCliProv;

type
  TRtcWSockHttpClientProvider = class(TRtcWSockClientProvider)
  private
    FOnInvalidResponse:TRtcEvent;

    FMaxHeaderSize:integer;
    FMaxResponseSize:integer;

    FRequest:TRtcClientRequest;
    FResponse:TRtcClientResponse;

    FResponseBuffer:TRtcHugeString;

    FResponseWaiting:boolean;

    ReqComplete:boolean; // internal Request.Complete indicator (to avoid problems with changing Request objects)

    FChunked:boolean;
    FChunkState:byte;

    FResponseLine:boolean; // response line received
    InBuffer:AnsiString; // data received, including HTTP header (header will be stripped when read)
    FHaveResponse:boolean; // response header accepted, receiving request data.
    LenToRead:int64; // number of bytes left to read from last Request

    LenToWrite:int64; // number of bytes to write out using inherited Write()
    LenToSend:int64; // number of bytes left to send out (DataOut event)
    FHeaderOut:boolean;
    
    FFixupRequest: TRtcClientRequestFixup;

  protected
    procedure ClearResponse;

    procedure TriggerConnect; override;
    procedure TriggerConnectLost; override;
    procedure TriggerDataReceived; override;
    procedure TriggerDataSent; override;
    procedure TriggerDataOut; override;

    procedure TriggerInvalidResponse; virtual;

  public
    constructor Create; override;
    destructor Destroy; override;

    procedure SetTriggerInvalidResponse(Event:TRtcEvent);

    procedure WriteHeader(SendNow:boolean=True); overload;
    procedure WriteHeader(const Header_Text:AnsiString; SendNow:boolean=True); overload;

    procedure Write(const ResultData:AnsiString; SendNow:boolean=True); override;

    // On DataReceived, read server response body using this:
    function Read:AnsiString; override;

    property Request:TRtcClientRequest read FRequest write FRequest;
    property Response:TRtcClientResponse read FResponse write FResponse;

    // Max. allowed size of the first (status) line in response header
    property MaxResponseSize:integer read FMaxResponseSize write FMaxResponseSize;
    // Max. allowed size of the complete response Header
    property MaxHeaderSize:integer read FMaxHeaderSize write FMaxHeaderSize;

    property FixupRequest:TRtcClientRequestFixup read FFixupRequest write FFixupRequest;
    end;

implementation

const
  CRLF:AnsiString = #13#10;
  END_MARK:AnsiString = #13#10#13#10;

{ TRtcWSockHttpClientProvider }

procedure TRtcWSockHttpClientProvider.SetTriggerInvalidResponse(Event: TRtcEvent);
  begin
  FOnInvalidResponse:=Event;
  end;

procedure TRtcWSockHttpClientProvider.TriggerInvalidResponse;
  begin
  if assigned(FOnInvalidResponse) then
    FOnInvalidResponse;
  end;

constructor TRtcWSockHttpClientProvider.Create;
  begin
  inherited;
  FResponseBuffer:=TRtcHugeString.Create;

  InBuffer:='';
  LenToWrite:=0;
  LenToSend:=0;
  FHeaderOut:=False;
  FResponseLine:=False;
  ReqComplete:=False;

  FFixupRequest:=nil;
  end;

destructor TRtcWSockHttpClientProvider.Destroy;
  begin
  FResponseBuffer.Free;

  InBuffer:='';
  LenToWrite:=0;
  LenToSend:=0;
  FResponseLine:=False;
  FHeaderOut:=False;
  FResponseWaiting:=False;

  inherited;
  end;

procedure TRtcWSockHttpClientProvider.ClearResponse;
  begin
  FResponseBuffer.Clear;

  FResponseLine:=False;
  FResponse.Clear;
  LenToRead:=-1;
  end;

procedure TRtcWSockHttpClientProvider.TriggerConnect;
  begin
  Request.Init;

  FResponseBuffer.Clear;

  InBuffer:='';
  LenToWrite:=0;
  LenToSend:=0;
  FHeaderOut:=False;
  FResponseLine:=False;
  FResponseWaiting:=False;
  FHaveResponse:=False;
  FChunked:=False;
  FChunkState:=0;

  ClearResponse;

  inherited;
  end;

procedure TRtcWSockHttpClientProvider.TriggerConnectLost;
  begin
  if FHaveResponse then // Processing a response ...
    begin
    if not FChunked and (LenToRead=-1) then // No content-length and not chunked
      begin
      LenToRead:=0;
      Response.Done:=True;
      Request.Active:=False;
      FHaveResponse:=False; // get ready for next request
      FResponseLine:=False;
      FHeaderOut:=False;
      FChunked:=False;
      FChunkState:=0;

      ReqComplete:=False; // DataReceived events have to wait until a new request has been sent out

      inherited TriggerDataReceived;
      end;
    end;
  inherited;
  end;

procedure TRtcWSockHttpClientProvider.TriggerDataReceived;
  var
    InBuf,
    s,
    StatusLine,
    HeadStr:AnsiString;
    HeadLen,
    MyPos:integer;

  function HexToInt(s:AnsiString):integer;
    var
      i,len:integer;
      c:AnsiChar;
    begin
    Result:=0;
    len:=length(s);
    i:=1;
    while len>0 do
      begin
      c:=s[len];
      if c in ['1'..'9'] then
        Result:=Result+i*(Ord(c)-Ord('0'))
      else if s[len] in ['A'..'F'] then
        Result:=Result+i*(Ord(c)-Ord('A')+10)
      else if s[len] in ['a'..'f'] then
        Result:=Result+i*(Ord(c)-Ord('a')+10);
      i:=i*16;Dec(len);
      end;
    end;

  procedure ResponseError;
    begin
    ReqComplete:=False; // no more reading, please!
    FResponseLine:=False;

    TriggerInvalidResponse;
    end;

  begin
  if not ReqComplete then
    begin
    if assigned(CryptPlugin) then
      begin
      // Read AnsiString from buffer
      InBuffer:=InBuffer + inherited Read;

      if InBuffer='' then
        begin
        FResponseWaiting:=True;
        Exit;
        end
      else
        FResponseWaiting:=False;
      end
    else
      begin
      FResponseWaiting:=True;
      Exit;
      end;
    end
  else
    FResponseWaiting:=False;

  // Read AnsiString from buffer
  InBuffer:=InBuffer + inherited Read;

  repeat
    if not FHaveResponse then // Don't have the header yet ...
      begin
      if not FResponseLine then
        begin
        if length(InBuffer)>5 then
          InBuf:=UpperCase(Copy(InBuffer,1,5))
        else
          InBuf:=UpperCase(InBuffer);
        // Accept streaming data as response
        if ((length(InBuf)=5) and (InBuf<>'HTTP/')) or
           ((length(InBuf)=4) and (InBuf<>'HTTP')) or
           ((length(InBuf)=3) and (InBuf<>'HTT')) or
           ((length(InBuf)=2) and (InBuf<>'HT')) or
           ((length(InBuf)=1) and (InBuf<>'H')) then
          begin
          ClearResponse;

          Response.Receiving:=True;
          Response.Started:=True;

          FHaveResponse:=True;
          FResponseLine:=True;
          LenToRead:=-1; // Unlimited length (streaming data until disconnected)

          Continue;
          end;

        MyPos:=PosEx2(CRLF,InBuffer);
        if (MaxResponseSize>0) and
           ( (MyPos>MaxResponseSize+1) or
             ((MyPos<=0) and (length(InBuffer)>MaxResponseSize+length(CRLF))) ) then
          begin
          ClearResponse;

          ResponseError;
          Exit;
          end
        else if (MyPos>0) then
          begin
          ClearResponse;

          StatusLine:=Copy(InBuffer,1,MyPos-1);
          Delete(InBuffer,1,MyPos+length(CRLF)-1);

          if UpperCase(Copy(StatusLine,1,5))<>'HTTP/' then
            begin
            ResponseError;
            Exit;
            end;

          Response.Receiving:=True;
          Response.Started:=True;

          { Our line probably looks like this:
            HTTP/1.1 200 OK }
          MyPos:=PosEx(' ',StatusLine); // first space before StatusCode
          if MyPos<=0 then
            begin
            ResponseError;
            Exit;
            end;
          Delete(StatusLine,1,MyPos); // remove 'HTTP/1.1 '

          MyPos:=PosEx(' ',StatusLine); // space after StatusCode
          if MyPos<=0 then // no Status Text
            begin
            s:=StatusLine;
            StatusLine:='';
            end
          else
            begin
            s:=Copy(StatusLine,1,MyPos-1); // StatusCode
            Delete(StatusLine,1,MyPos); // StatusText
            end;

          if s<>'' then
            begin
            try
              Response.StatusCode:=Str2Int(s);
              Response.StatusText:=StatusLine;
            except
              // if there is something wrong with this, just ignore the exception
              end;
            end;

          FResponseLine:=True;
          end;
        end;

      if FResponseLine then
        begin
        // See if we can get the whole header ...
        HeadLen:=Pos(CRLF, InBuffer);
        if HeadLen<>1 then
          HeadLen:=Pos(END_MARK, InBuffer);

        if HeadLen=1 then
          begin
          // Delete CRLF from the body
          Delete(InBuffer,1,2);

          if Response.StatusCode=100 then
            begin // special handling of the "100:Continuing" Http status code
            FResponseLine:=False;
            Continue;
            end;

          // No Header: disconnect closes the response.
          Request.Close:=True;

          if (Request.Method='HEAD') or
             (Response.StatusCode=204) or
             (Response.StatusCode=304) or
             ( (Response.StatusCode>=100) and (Response.StatusCode<=199) ) then
            begin
            FChunked:=False;
            LenToRead:=0;
            end;

          FHaveResponse:=True;
          end
        else if (MaxHeaderSize>0) and
           ( (HeadLen>MaxHeaderSize) or
             ((HeadLen<=0) and (length(InBuffer)>MaxHeaderSize+length(END_MARK))) ) then
          begin
          ResponseError;
          Exit;
          end
        else if HeadLen>0 then
          begin
          // Separate header from the body
          HeadStr:=Copy(InBuffer, 1, HeadLen+length(END_MARK)-1);
          Delete(InBuffer,1,HeadLen+length(END_MARK)-1);

          if Response.StatusCode=100 then
            begin // special handling of the "100:Continuing" Http status code
            FResponseLine:=False;
            Continue;
            end;

          FHaveResponse:=True;

          // Scan for all header attributes ...
          MyPos:=Pos(CRLF, HeadStr);
          while (MyPos>1) do // at least 1 character inside line
            begin
            StatusLine:=Copy(HeadStr,1,MyPos-1);
            Delete(HeadStr,1,MyPos+Length(CRLF)-1);

            MyPos:=PosEx(':',StatusLine);
            if MyPos>0 then
              begin
              s:=UpperCase(Trim(Copy(StatusLine,1,MyPos-1)));
              Delete(StatusLine,1,MyPos);
              StatusLine:=Trim(StatusLine);

              if s='TRANSFER-ENCODING' then
                begin
                if UpperCase(StatusLine)='CHUNKED' then
                  begin
                  FChunked:=True;
                  FChunkState:=0;
                  end;
                end
              else if s='CONTENT-LENGTH' then
                begin
                LenToRead:=Str2Int64Def(StatusLine,0);

                Response.ContentLength:=LenToRead;
                end
              else if s='CONNECTION' then
                begin
                if UpperCase(Trim(StatusLine))='CLOSE' then
                  Request.Close:=True
                else if UpperCase(Trim(StatusLine))='KEEP-ALIVE' then
                  Request.Close:=False;
                end;

              Response[s]:=StatusLine;
              end;

            MyPos:=Pos(CRLF, HeadStr);
            end;

          if LenToRead=-1 then
            Request.Close:=True;

          if (Request.Method='HEAD') or
             (Response.StatusCode=204) or
             (Response.StatusCode=304) or
             ( (Response.StatusCode>=100) and (Response.StatusCode<=199) ) then
            begin
            FChunked:=False;
            LenToRead:=0;
            end;

          StatusLine:='';
          HeadStr:='';
          end;
        end;
      end;

    if FHaveResponse then // Processing a response ...
      begin
      if FChunked then // Read data as chunks
        begin
        if (FChunkState=0) and (InBuffer<>'') then // 1.step = read chunk size
          begin
          MyPos:=Pos(CRLF,InBuffer);
          if MyPos>0 then
            begin
            StatusLine:=Trim(Copy(InBuffer,1,MyPos-1));
            Delete(InBuffer,1,MyPos+1);

            LenToRead:=HexToInt(StatusLine);

            FChunkState:=1; // ready to read data
            end;
          end;

        if (FChunkState=1) and (InBuffer<>'') then // 2.step = read chunk data
          begin
          if (LenToRead>length(InBuffer)) then // need more than we have
            begin
            Response.ContentIn:=Response.ContentIn+length(InBuffer);

            if LenToRead>0 then
              Dec(LenToRead, length(InBuffer));

            FResponseBuffer.Add(InBuffer);
            InBuffer:='';

            inherited TriggerDataReceived;

            Response.Started:=False;
            end
          else
            begin
            if LenToRead>0 then
              begin
              Response.ContentIn:=Response.ContentIn+LenToRead;

              FResponseBuffer.Add(Copy(InBuffer,1,LenToRead));

              Delete(InBuffer,1,LenToRead);
              LenToRead:=0;
              FChunkState:=2; // this is not the last chunk, ready to read CRLF
              end
            else
              FChunkState:=3; // this was last chunk, ready to read CRLF
            end;
          end;

        if (FChunkState>=2) and (length(InBuffer)>=2) then // 3.step = close chunk
          begin
          LenToRead:=-1;
          Delete(InBuffer,1,2); // Delete CRLF

          if FChunkState=2 then
            begin
            FChunkState:=0;
            end
          else
            begin
            Response.Done:=True;
            Request.Active:=False;
            FHaveResponse:=False; // get ready for next request
            FChunked:=False;
            FChunkState:=0;
            FResponseLine:=False;
            FHeaderOut:=False;

            ReqComplete:=False; // DataReceived events have to wait until a new request has been sent out
            end;

          inherited TriggerDataReceived;

          Response.Started:=False;
          end;
        end
      else // Read data as stream or with predefined length
        begin
        if (LenToRead>0) or (LenToRead=-1) then
          begin
          if (LenToRead>length(InBuffer)) or
             (LenToRead=-1) then // need more than we have
            begin
            Response.ContentIn:=Response.ContentIn+length(InBuffer);

            if LenToRead>0 then
              Dec(LenToRead, length(InBuffer));

            FResponseBuffer.Add(InBuffer);

            InBuffer:='';
            end
          else
            begin
            Response.ContentIn:=Response.ContentIn+LenToRead;

            FResponseBuffer.Add(Copy(InBuffer,1,LenToRead));

            Delete(InBuffer,1,LenToRead);

            LenToRead:=0;
            Response.Done:=True;
            Request.Active:=False;
            FHaveResponse:=False; // get ready for next request
            FChunked:=False;
            FResponseLine:=False;
            FHeaderOut:=False;

            ReqComplete:=False; // DataReceived events have to wait until a new request has been sent out
            end;
          end
        else
          begin
          Response.Done:=True;
          Request.Active:=False;
          FHaveResponse:=False; // get ready for next request
          FChunked:=False;
          FResponseLine:=False;
          FHeaderOut:=False;

          ReqComplete:=False; // DataReceived events have to wait until a new request has been sent out
          end;

        inherited TriggerDataReceived;

        Response.Started:=False;
        end;
      end
    else
      Break; // Failing to fetch a header will break the loop.

    until (InBuffer='') or not ReqComplete;
  end;

procedure TRtcWSockHttpClientProvider.WriteHeader(SendNow:boolean=True);
  var
    s:AnsiString;
  begin
  if FHeaderOut then
    raise Exception.Create('Last header intercepted with new header, before data sent out.');

  FixupRequest.Fixup(Request);

  s:=Request.Method+' '+Request.URI+' HTTP/1.1'+CRLF+
     Request.HeaderText;

  if Request.Close then s:=s+'Connection: close'+CRLF;

  s:=s+CRLF;

  Request.Started:=True;
  Request.Active:=True;

  LenToWrite:=Request.ContentLength;
  LenToSend:=length(s) + Request.ContentLength;

  FHeaderOut:=True;
  inherited Write(s, SendNow or (LenToWrite<=0));
  end;

procedure TRtcWSockHttpClientProvider.WriteHeader(const Header_Text:AnsiString; SendNow:boolean=True);
  var
    s:AnsiString;
  begin
  if FHeaderOut then
    raise Exception.Create('Last header intercepted with new header, before data sent out.');

  if Header_Text<>'' then
    Request.HeaderText:=Header_Text;

  FixupRequest.Fixup(Request);

  s:=Request.Method+' '+Request.URI+' HTTP/1.1'+CRLF +
     Request.HeaderText;

  if Request.Close then s:=s+'Connection: close'+CRLF;

  s:=s+CRLF;

  Request.Started:=True;
  Request.Active:=True;

  LenToWrite:=Request.ContentLength;
  LenToSend:=length(s) + Request.ContentLength;

  FHeaderOut:=True;
  inherited Write(s, SendNow or (LenToWrite<=0));
  end;

procedure TRtcWSockHttpClientProvider.Write(const ResultData: AnsiString; SendNow:boolean=True);
  begin
  if length(ResultData)=0 then Exit;

  if not FHeaderOut then
    raise Exception.Create('Trying to send Data without Header. Call WriteHeader before Write.');

  if length(ResultData)>LenToWrite then
    raise Exception.Create('Trying to send more Data out than specified in Header.');

  Dec(LenToWrite, length(ResultData));

  Request.ContentOut:=Request.ContentOut + length(ResultData);

  inherited Write(ResultData, SendNow);
  end;

function TRtcWSockHttpClientProvider.Read: AnsiString;
  begin
  if FResponseBuffer.Size>0 then
    begin
    Result:=FResponseBuffer.Get;
    FResponseBuffer.Clear;
    end
  else
    Result:='';
  end;

procedure TRtcWSockHttpClientProvider.TriggerDataSent;
  begin
  if Request.Active then
    Request.Started:=False;

  inherited TriggerDataSent;

  if FResponseWaiting then
    if ReqComplete then
      TriggerDataReceived;
  end;

procedure TRtcWSockHttpClientProvider.TriggerDataOut;
  begin
  if not ReqComplete then
    if assigned(Request) and Request.Active then
      begin
      Dec(LenToSend, DataOut);
      ReqComplete := LenToSend<=0;
      Request.Complete := ReqComplete;
      end;

  inherited TriggerDataOut;
  end;

end.
