unit rtcWafSrv;

interface

{$include rtcDefs.inc}

{.$DEFINE RTC_QUICK_LOCK}

uses
  rtcTrashcan,
  Classes, SysUtils,

  rtcFunction, rtcDataSrv, rtcSrvModule, rtcInfo, rtcConn,
  rtcHttpSrv, rtcSyncObjs, memStrObjList,

  rtcLink, rtcBase, rtcForms, rtcLog;

type
  TRtcWafErrorHandler=procedure(const E:Exception) of object;
  TRtcWafSetBaseFormClass=procedure(var Form:TRtcBaseFormClass) of object;

  TRtcWafServerModule=class(TRtcBaseServerModule)
  private
    FnPut: TRtcFunction;
    FnGet: TRtcFunction;
    FnRun: TRtcFunction;
    FnStop: TRtcFunction;
    FnShow: TRtcFunction;

    FnGroup: TRtcFunctionGroup;

    FErrorHandler: TRtcWafErrorHandler;
    FOnStart: TRtcBaseEvent;

  protected
    procedure FnRunExecute(Sender: TRtcConnection; Param: TRtcFunctionInfo; Result: TRtcValue);
    procedure FnStopExecute(Sender: TRtcConnection; Param: TRtcFunctionInfo; Result: TRtcValue);
    procedure FnPutExecute(Sender: TRtcConnection; Param: TRtcFunctionInfo; Result: TRtcValue);
    procedure FnGetExecute(Sender: TRtcConnection; Param: TRtcFunctionInfo; Result: TRtcValue);
    procedure FnShowExecute(Sender: TRtcConnection; Param: TRtcFunctionInfo; Result: TRtcValue);

  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;

  published

    { This event has to be implemented for every RTC Web Application.
      From this event, you need to create, inizialize and show your
      Main Form and any DataModules required by your Main Form.
      If you do not create and show your Main Form here
      (using the ".Show" method of the Form), your Main Form will not become
      visible on the Client side and the user will not be able to use it. }
    property OnAppStart:TRtcBaseEvent read FOnStart write FOnStart;
    { This event does not necessarily have to be implemented,
      but when implemented, it will be triggered any time an exception gets out unhandled.
      You can either create a form to display an error message to the Clinent, log the
      error message (and the associated exception class), or simply ignore all
      exceptions and the user will not be notified about any unhandled exceptions. }
    property OnAppError:TRtcWafErrorHandler read FErrorHandler write FErrorHandler;

  published
  // properties and events from the parent class which need to be published

    {$IFDEF COMPRESS}
    { Use this property to define what compression level you want to use when sending
      data from Server to client. Default Compression value is "cNone" (no compression).
      You can use different compression levels between client and server (for example,
      fast or no compression from client and max from server). If your server has to
      work with clients which don't support compression, you have to use "cNone". }
    property Compression;
    {$ENDIF}

    {  Use this property to define what Data Formats you want to accept for this ServerModule.
       Since this is a set, you can choose one supported format or all supported formats. }
    property DataFormats;

    { Set this property to a value other than 0 if you want to enable automatic
      Encryption for clients which have their EncryptionKey option activated.
      EncryptionKey value defines the length on the encryption key, in bytes.
      One byte represents encryption strength of 8 bits. To have 256-bit encryption,
      set EncryptionKey=32. @html(<br><br>)

      The final encryption key is combined from a key received by the client
      and a key generated by this ServerModule. When EncryptionKey is >0 by the
      ClientModule doing the call AND by this ServerModule, Encryption handshake will
      be done automatically by those two components, so that the user only has to set
      the values and use the components as if there is no encryption.

      If ServerModule's EncryptionKey property is 0 (zero), server will not allow data to be
      encrypted and all communication with all clients will flow without encryption.
      Clients which have ForceEncryption set to True, will not work if the server
      doesn't want to support encryption. }
    property EncryptionKey;

    { If you need a 100% secure connection, define a Secure Key AnsiString
      (combination of letters, numbers and special characters) for each
      ServerModule/ClientModule pair, additionally to the EncryptionKey value.
      ClientModule will be able to communicate with the ServerModule ONLY if
      they both use the same SecureKey. Default value for the SecureKey is
      an empty AnsiString (means: no secure key). @html(<br><br>)

      SecureKey will be used in the encryption initialisation handshake,
      to decrypt the first key combination received by the ClientModule.
      Since all other data packages are already sent using some kind of encryption,
      by defining a SecureKey, you encrypt the only key part which would have
      been sent out without special encryption. }
    property SecureKey;

    { Setting this property to TRUE will tell the ServerModule to work with
      Clients ONLY if they requested to use encryption. If AutoEncryptKey is > 0 and
      Client doesn't request encryption, function calls will not be processed
      and any request coming from the client will be rejected, until
      client requests and initializes encryption. }
    property ForceEncryption;

    { Using this property, you define how long a session will live (in seconds)
      when there are no requests from this client and the session was
      created by a call from ClientModule that has its AutoSessions property enabled.
      The higher this value, the longer a session will stay valid when there
      are no requests coming from the client for which the session was created.
      This value will be used only after the Client has sent a valid request
      which produces a valid response from the server. Before that, a default
      Session Live time of @Link(RTC_SESSION_TIMEOUT) seconds will be used. @html(<br><br>)

      Session Live counter is reset each time a new request is received from the same client,
      so that this parameter only removes sessions which are inactive longer than
      AutoSessionsLive seconds. To keep your server from nasty clients creating tons of
      sessions and leaving them inactive, keep this property under 600 seconds,
      even if you want your session to stay alive for a long time. You do not have to
      overexagurate this value, since every session consumes memory and client sessions which are not
      needed will ONLY THEN be removed from memory when this AutoSessionsLive timeout expires. }
    property AutoSessionsLive;

    { When AutoSessions are used, you can define what client data you want to use to lock the
      Sessions to this client and deny access to session data from other clients. }
    property AutoSessionsLock;

    { If ModuleHost is specified, then Request.Host will be compared to the ModuleHost
      property to decide if this request should be processed by this ServerModule. @html(<br>)
      If your DataServer has to serve more different hosts, while your ServerModule
      is not supposed to react to requests from all those hosts, you can assign the
      host name to which this ServerModule belongs to. If ModuleHost is left blank,
      then this ServerModule will respond to any request asking for this servermodule's
      ModuleFileName, regardless of the HOST header. @html(<br><br>)

      To process all requests for a domain and all of its sub-domains, enter domain name ONLY
      (example: "realthinclient.com" for "realthinclient.com" and any sub-domain like
      "www.realthinclient.com", "mymod.myapp.realthinclient.com", etc). @html(<br>)
      To limit the requests to a sub-domain and its sub-sub-domains, enter the name
      of the highest sub-domain (example: "myapp.realthinclient.com" for
      "myapp.realthinclient.com" and any of its sub-domains like
      "mymod.myapp.realthinclient.com"). @html(<br>)
      To process ONLY requests pointed exactly to ONE HOST, add "." in front of your
      host name (example: ".realthinclient.com" will ONLY react to requests with
      "realthinclient.com" in their HOST header). }
    property ModuleHost;
    { This property will be compared to Request.FileName to decide if the
      request we just received was pointed to this ServerModule. Any request asking
      for this FileName will be processed by this ServerModule component.
      Since parameters are passed to the server module through request's Content
      body (rather than headers), we do not need to check the request for anything
      else than it's FileName to know if the request is directed to this module. }
    property ModuleFileName;

    { Event to be triggered after a new Session was opened.
      This event is not mandatory, but it can be used for session logging. }
    property OnSessionOpen;
    { Event to be triggered before an existing Session was to be closed.
      This event is not mandatory, but it can be used for session logging. }
    property OnSessionClose;
    { Event to be triggered when the Server starts listening on the connection.
      This event is not mandatory, but is can be used for component initialization. }
    property OnListenStart;
    { Event to be triggered when the Server stops listening on the connection.
      This event is not mandatory, but it can be used for component deinitialization. }
    property OnListenStop;
    end;

implementation

var
  AppCS:TRtcCritSec;
  AppList:TStrObjList;

type
  TRtcWafApplication=class(TRtcApplication)
  private
    CS,CS2:TRtcCritSec;
    EV:TRtcEvent;
    FOnError: TRtcWafErrorHandler;

  public
    Closing:boolean;
    LockCount:integer;
    ToSend:TRtcArray;
    Sleeper:TRtcDelayedCall;

    constructor CreateNew(const ID:AnsiString); virtual;
    destructor Destroy; override;

    // App has prepared Data which has to be sent to the Client
    procedure ClientExecute(const Data:TRtcValueObject);
    // App exception error handler
    procedure ErrorHandler(const E:Exception);

    procedure Lock;
    procedure UnLock;

    // Data received from Client for local execution
    procedure Execute(Data:TRtcArray);

    function GetToSend(var dc:TRtcDelayedCall):TRtcArray;

    property OnError:TRtcWafErrorHandler read FOnError write FOnError;
    end;

  TRtcApplicationSession=class(TRtcObject)
    AppID:AnsiString;
    procedure Kill; override;
    end;

{ Global Functions }

procedure PrepareApps;
  begin
  AppCS:=TRtcCritSec.Create;
  AppList:=tStrObjList.Create(32);
  end;

procedure ReleaseApps;
  begin
  Garbage(AppCS);
  Garbage(AppList);
  end;

procedure AppStore(const app:TRtcWafApplication; const ID:AnsiString);
  begin
  AppCS.Enter;
  try
    AppList.insert(ID,app);
  finally
    AppCS.Leave;
    end;
  end;

{$IFDEF RTC_QUICK_LOCK}
function AppLock(const ID:AnsiString):TRtcWafApplication;
  var
    o:TObject;
  begin
  AppCS.Enter;
  try
    o:=AppList.search(ID);
    if assigned(o) then
      Result:=TRtcWafApplication(o)
    else
      begin
      Result:=nil;
      AppCS.Leave;
      end;
  except
    AppCS.Leave;
    raise;
    end;
  end;
procedure AppUnLock(var app:TRtcWafApplication);
  begin
  if assigned(app) then
    begin
    app:=nil;
    AppCS.Leave;
    end;
  end;
{$ELSE}
function AppLock(const ID:AnsiString):TRtcWafApplication;
  var
    o:TObject;
  begin
  AppCS.Enter;
  try
    o:=AppList.search(ID);
    if assigned(o) then
      begin
      Result:=TRtcWafApplication(o);
      Inc(Result.LockCount);
      end
    else
      Result:=nil;
  finally
    AppCS.Leave;
    end;
  if assigned(Result) then
    Result.Lock;
  end;
procedure AppUnLock(var app:TRtcWafApplication);
  var
    temp:TRtcWafApplication;
  begin
  if assigned(app) then
    begin
    app.UnLock;
    temp:=app; app:=nil;

    AppCS.Enter;
    try
      Dec(temp.LockCount);
      if temp.Closing and (temp.LockCount=0) then
        temp.EV.SetEvent;
    finally
      AppCS.Leave;
      end;
    end;
  end;
{$ENDIF}

procedure AppRemove(const ID:AnsiString);
  var
    o:TObject;
    app:TRtcWafApplication;
  begin
  app:=nil;
  AppCS.Enter;
  try
    o:=AppList.search(ID);
    if assigned(o) then
      begin
      app:=TRtcWafApplication(o);
      app.Closing:=True;
      AppList.remove(id);
      if app.LockCount>0 then
        app.EV.ResetEvent // need to wait for LockCount to drop to ZERO
      else
        app.EV.SetEvent; // nothing to wait for
      end
    else
      raise Exception.Create('Application destroyed before destructor ('+String(ID)+')!');
  finally
    AppCS.Leave;
    end;
  if assigned(app) then
    app.EV.WaitFor(WAIT_INFINITE);
  end;

function AppQuit(const ID:AnsiString):boolean;
  var
    App:TRtcWafApplication;
  begin
  App:=AppLock(ID);
  if assigned(App) then
    try
      Result:=App.Quit
    finally
      AppUnLock(App);
    end
  else
    Result:=False;
  end;

{ TRtcWafApplication }

constructor TRtcWafApplication.CreateNew(const ID:AnsiString);
  begin
  inherited Create(nil);

  UserID:=ID;

  // Log('CreateNew', String(UserID));

  CS:=TRtcCritSec.Create;
  CS2:=TRtcCritSec.Create;
  EV:=TRtcEvent.Create(True,False);
  Sleeper:=nil;
  ToSend:=nil;

  MaxUpdateDelay:=10; // keep max 10 updates in cache before sending them out
  OnSend:=ClientExecute;
  OnException:=ErrorHandler;

  UseClientCache:=True;

  AppStore(self,UserID);
  end;

destructor TRtcWafApplication.Destroy;
  begin
  // Log('Destroy', String(UserID));

  AppRemove(UserID); // will wait for all locks to be released and then destroy the App object

  OnSend:=nil;
  OnException:=nil;

  if assigned(Sleeper) then
    begin
    Sleeper.WakeUp;
    Sleeper:=nil;
    end;

  if assigned(ToSend) then FreeAndNil(ToSend);

  try
    FreeAndNil(CS);
    FreeAndNil(CS2);
    FreeAndNil(EV);
  except
    on E:Exception do
      Log('TRtcWafApplication.Destroy:',E);
    end;

  try
    inherited;
  except
    on E:Exception do
      Log('TRtcWafApplication.Destroy inherited:',E);
    end;
  end;

procedure TRtcWafApplication.ErrorHandler(const E: Exception);
  begin
  if assigned(FOnError) then
    FOnError(E);
  { with TErrorForm.Create(nil) do
      begin
      Caption:=E.ClassName;
      eMessage.Caption:=E.Message;
      Show;
      end; }
  end;

procedure TRtcWafApplication.Execute(Data: TRtcArray);
  var
    a:integer;
    o:TRtcValueObject;
  begin
  // Log('IN : '+Data.toCode, String(UserID));

  // Called from AppExecute: Connection B (Put) wants to prepare data for execution
  EnterExecute;
  try
    for a:=0 to Data.Count-1 do
      begin
      o:=Data.asObject[a];
      Data.asObject[a]:=nil;
      ToExecute(o);
      end;
  finally
    LeaveExecute;
    end;
  end;

procedure TRtcWafApplication.ClientExecute(const Data: TRtcValueObject);
  begin
  // Called directly by RtcApplication object to prepare data for the Client
  CS2.Enter;
  try
    if not assigned(ToSend) then
      ToSend:=TRtcArray.Create;
    ToSend.asObject[ToSend.Count]:=Data;
    if assigned(Sleeper) then
      begin
      // Log('WakeUp!', String(UserID));
      Sleeper.WakeUp;
      Sleeper:=nil;
      end;
  finally
    CS2.Leave;
    end;
  end;

function TRtcWafApplication.GetToSend(var dc:TRtcDelayedCall): TRtcArray;
  begin
  // Connection A wants to pick up data for the Client
  CS2.Enter;
  try
    Result:=ToSend;
    ToSend:=nil;
    if Result=nil then
      begin
      // Log('Sleeping ...', String(UserID));
      Sleeper:=dc;
      end
    else
      begin
      // Log('OUT: '+Result.toCode, String(UserID));
      if assigned(dc) then
        begin
        CancelDelayedCall(dc);
        dc:=nil;
        end;
      end;
  finally
    CS2.Leave;
    end;
  end;

procedure TRtcWafApplication.Lock;
  begin
  CS.Enter;
  end;

procedure TRtcWafApplication.UnLock;
  begin
  CS.Leave;
  end;

{ TRtcApplicationSession }

procedure TRtcApplicationSession.Kill;
  begin
  // Session expired
  // Log('Session Closed',String(AppID));
  AppQuit(AppID);
  AppID:='';
  Free;
  end;

{ TRtcWafServer }

constructor TRtcWafServerModule.Create(AOwner: TComponent);
  begin
  inherited;

  FnGroup:=TRtcFunctionGroup.Create(nil);

  FnPut:=TRtcFunction.Create(nil);
  FnPut.Group:=FnGroup;
  FnPut.FunctionName:='put';
  FnPut.OnExecute:=FnPutExecute;

  FnGet:=TRtcFunction.Create(nil);
  FnGet.Group:=FnGroup;
  FnGet.FunctionName:='get';
  FnGet.OnExecute:=FnGetExecute;

  FnRun:=TRtcFunction.Create(nil);
  FnRun.Group:=FnGroup;
  FnRun.FunctionName:='run';
  FnRun.OnExecute:=FnRunExecute;

  FnStop:=TRtcFunction.Create(nil);
  FnStop.Group:=FnGroup;
  FnStop.FunctionName:='stop';
  FnStop.OnExecute:=FnStopExecute;

  FnShow:=TRtcFunction.Create(nil);
  FnShow.Group:=FnGroup;
  FnShow.FunctionName:='show';
  FnShow.OnExecute:=FnShowExecute;

  FunctionGroup:=FnGroup;
  AutoSessions:=True;
  end;

destructor TRtcWafServerModule.Destroy;
  begin
  FunctionGroup:=nil;

  FnPut.Free;
  FnGet.Free;
  FnRun.Free;
  FnStop.Free;
  FnShow.Free;
  FnGroup.Free;
  inherited;
  end;

procedure TRtcWafServerModule.FnRunExecute(Sender: TRtcConnection; Param: TRtcFunctionInfo; Result: TRtcValue);
  var
    Ses:TRtcServerSession;
    AppSess:TRtcApplicationSession;
    App:TRtcWafApplication;
  begin
  // Connection A
  Ses:=TRtcDataServer(Sender).Session;
  if not assigned(Ses) then
    raise Exception.Create('AutoSessions needs to be TRUE');

  if assigned(Ses.Obj['app']) then
    begin
    // We need an Application instance
    AppSess:=TRtcApplicationSession(Ses.Obj['app']);
    App:=AppLock(AppSess.AppID);
    if assigned(App) then
      try
        App.FreeAllForms(True);
      finally
        AppUnLock(App);
      end
    else
      begin
      App:=TRtcWafApplication.CreateNew(Ses.ID);
      App.OnError:=FErrorHandler;
      end;
    end
  else
    begin
    // We need an Application instance
    AppSess:=TRtcApplicationSession.Create;
    App:=TRtcWafApplication.CreateNew(Ses.ID);
    App.OnError:=FErrorHandler;

    AppSess.AppID:=Ses.ID;
    Ses.Obj['app']:=AppSess;
    end;

  Result.asString:=Ses.ID;
  end;

procedure TRtcWafServerModule.FnGetExecute(Sender: TRtcConnection; Param: TRtcFunctionInfo; Result: TRtcValue);
  var
    Ses:TRtcServerSession;
    AppSess:TRtcApplicationSession;
    App:TRtcWafApplication;
    dc:TRtcDelayedCall;
    data:TRtcValueObject;
  begin
  // Connection A
  Ses:=TRtcDataServer(Sender).Session;

  if not assigned(Ses) then // No session
    Result.asException:='QUIT'
  else if not assigned(Ses.Obj['app']) then // Session invalid
    Result.asException:='QUIT'
  else
    begin
    if not Ses.asBoolean['delay'] then
      dc:=PrepareDelayedCall(20000,Param,FnGetExecute)
    else
      begin
      Ses.asBoolean['delay']:=False;
      dc:=nil;
      end;

    AppSess:=TRtcApplicationSession(Ses.Obj['app']);
    App:=AppLock(AppSess.AppID);
    if not assigned(App) then
      begin
      if assigned(dc) then
        begin
        CancelDelayedCall(dc);
        dc:=nil;
        end;
      AppSess.Kill;
      Ses.Obj['app']:=nil;
      Ses.Close;
      Result.asException:='QUIT';
      end
    else
      begin
      try
        data:=App.GetToSend(dc);
      finally
        AppUnLock(App);
        end;

      if assigned(data) then
        Result.asObject:=data
      else if assigned(dc) then
        begin
        // No data waiting, we were not sleeping yet
        Ses.asBoolean['delay']:=True;
        PostDelayedCall(dc);
        end;
      end;
    // else: No data waiting, we were waken up from sleep, just return NULL
    end;
  end;

procedure TRtcWafServerModule.FnShowExecute(Sender: TRtcConnection; Param: TRtcFunctionInfo; Result: TRtcValue);
  var
    App:TRtcWafApplication;
  begin
  if assigned(FOnStart) then
    begin
    App:=AppLock(Param.asString['app']);
    if assigned(App) then
      try
        Result.asBoolean:=App.ExecuteEvent(FOnStart);
      finally
        AppUnLock(App);
      end
    else
      Result.asBoolean:=False;
    end
  else
    Result.asBoolean:=False;
  end;

procedure TRtcWafServerModule.FnPutExecute(Sender: TRtcConnection; Param: TRtcFunctionInfo; Result: TRtcValue);
  var
    App:TRtcWafApplication;
  begin
  App:=AppLock(Param.asString['app']);
  if assigned(App) then
    try
      App.Execute(Param.asArray['data']);
      Result.asBoolean:=True;
    finally
      AppUnLock(App);
    end
  else
    Result.asBoolean:=False;
  end;

procedure TRtcWafServerModule.FnStopExecute(Sender: TRtcConnection; Param: TRtcFunctionInfo; Result: TRtcValue);
  begin
  Result.asBoolean:=AppQuit(Param.asString['app']);
  end;

initialization
  PrepareApps;
finalization
  ReleaseApps;
end.
