{
  RealThinClient SDK - UNIX support
  Copyright (c) Danijel Tkalcec
  @exclude
}

unit rtcunixThreadTimer;
// unix thread based timer. Used by console Window Emulation

{.$define DebugTimer}

interface

uses
  BaseUnix, Unix,
  SyncObjs, Classes;

type
  TTimerEvent = procedure (Sender: TObject) of object;

type
  { TThreadedTimer }

  TThreadedTimer = class(TObject)
  private
    fInterval : Integer;
    fLock     : TCriticalSection;
    fEnabled  : Boolean;

    fLeftVal  : Integer;
    fOnTimer  : TTimerEvent;
    fAdded    : Boolean;
  protected
    procedure Lock;
    procedure Unlock;

    function GetInterval: Integer;
    procedure SetInterval(const AValue: Integer);
    function GetEnabled: Boolean;
    procedure SetEnabled(AEnabled: Boolean);

    function GetOnTimer: TTimerEvent;
    procedure SetOnTimer(AEvent: TTimerEvent);

    procedure ReleaseControlThread;

    procedure DoTimer; virtual;
  public
    Tag : TObject;
    constructor Create;
    destructor Destroy; override;

    property Interval: Integer read GetInterval write SetInterval;
    property Enabled: Boolean read GetEnabled write SetEnabled;

    // event is called from separate thread.
    property OnTimer : TTimerEvent read GetOnTimer write SetOnTimer;
  end;

function GetTickCount: LongWord;
procedure Sleep(mls: Integer);

implementation

procedure Sleep(mls: Integer);
var
  tm : timespec;
begin
  if mls <= 0 then Exit;
  tm.tv_nsec:= (mls mod 1000) * 1000*1000;
  tm.tv_sec := mls div 1000;
  FpNanoSleep(@tm, nil);
end;

{ debug functions }
{$ifdef DebugTimer}
procedure DebugTimer(const s: string);
begin
//  writeln(s);
end;

function IntToStr(i: Integer): string;
begin
  Str(i, Result);
end;
{$endif}

type
  { TControlThread }

  TControlThread = class(TThread)
  private
    fEvent    : TEvent;
    fLock     : TCriticalSection;
    fTimers   : TList;
    fDoStop   : Boolean;
  protected
    procedure Execute; override;

    procedure TimersLoop;

    // returns true, if Event is timeOut, if Event is signaled returns false
    function WaitPeriod(mlsWaitTime: Integer; var MustStop: Boolean): Boolean;
    procedure SignalTimers(delta: Integer);
    procedure GetNextWaitTime(var NextWaitTime: Integer; ExtraDelta: Integer);

    procedure Lock;
    procedure Unlock;

    procedure StopThread;

    procedure ReleaseTimers;

    procedure Update;
  public
    constructor Create(ASuspended: Boolean);
    destructor Destroy; override;

    procedure Add(Timer: TThreadedTimer);
    procedure Remove(Timer: TThreadedTimer);
  end;

var
  GlobalTimerThread : TControlThread = nil;
  Releasing         : Boolean = false;

function ControlThread: TControlThread;
begin
  if not Assigned(GlobalTimerThread) and not (Releasing) then
    GlobalTimerThread := TControlThread.Create(false);
  Result := GlobalTimerThread;
end;

function GetTickCount: LongWord;
var
  v : TTimeVal;
  i : Int64;
begin
  fpgettimeofday(@v, nil);
  i := Int64(v.tv_sec) * 1000 + Int64(v.tv_usec div 1000);
  Result := LongWord(i and $FFFFFFFF);
end;


{ TControlThread }

procedure TControlThread.Execute;
begin
  try
    TimersLoop;
  except
  end;
end;

procedure TControlThread.TimersLoop;
var
  back      : Integer;
  nexttime  : Integer;
  delta     : Integer;
  dtime     : Integer;
  Stop      : Boolean;
  timeout   : Boolean;
begin
  GetNextWaitTime(nexttime, 0);
  Stop := false;

  while not Stop do begin
    {$ifdef DebugTimer}DebugTimer('[ControlThread] starting loop');{$endif}
    back := GetTickCount;
    if nexttime <> 0 then begin
      {$ifdef DebugTimer}DebugTimer('[ControlThread] waiting for period: '+ IntToStr(nexttime));{$endif}
      timeout := WaitPeriod(nexttime, Stop)

    end;
    Lock;
    try
      dtime := GetTickCount;
      delta := dtime - back;

      // work around for Linux TEvent.Wait
      // TEvent fires with zero time, if wait time is less than 1000 Mls
      if timeout and (delta = 0) and (nexttime > 0) then begin
        Sleep(10);
        dtime := GetTickCount;
        delta := dtime - back;
      end;

      if nexttime >= 0 then // no need to signal, if no timers are waiting
        SignalTimers(delta);

      // timer events compensate
      delta := GetTickCount - dtime;
      {$ifdef DebugTimer}DebugTimer('[ControlThread] getting next wait time ');{$endif}
      GetNextWaitTime(nexttime, delta);
      {$ifdef DebugTimer}DebugTimer('[ControlThread] next wait time = '+ IntToStr(nexttime));{$endif}
    finally
      unlock;
    end;

  end;
end;

function TControlThread.WaitPeriod(mlsWaitTime: Integer; var MustStop: Boolean): Boolean;
begin
  Result := fEvent.WaitFor(mlsWaitTime) = wrTimeout;
  fEvent.ResetEvent;
  MustStop := fDoStop;
end;


function Min(a, b: Integer): Integer;
begin
  if a < b then Result := a else Result := b;
end;

// todo: delta must (might) increase with every t.DoTimer call
// and so it should be managed here as well

procedure TControlThread.SignalTimers(delta: Integer);
var
  i : Integer;
  t : TThreadedTimer;
  m : Integer;
begin
  for i := 0 to fTimers.Count - 1 do begin
    t := TThreadedTimer(fTimers[i]);

    t.Lock;
    try
      if not t.fEnabled then Continue;

      dec(t.fLeftVal, delta);
      {$ifdef DebugTimer}DebugTimer('[ControlThread] Left Time: '+IntToStr(t.fLeftVal) + '; delta = '+IntToStr(delta));{$endif}
      if t.fLeftVal <= 0 then begin
        t.DoTimer; // todo: growing delta here!
        if t.Interval > 0 then begin
          inc(t.fLeftVal, t.Interval);
          while (t.fLeftVal <= 0) do begin // if a lot of time missed
            t.fLeftVal := t.fLeftVal + t.fInterval;
            t.DoTimer;
          end;
        end;
      end;
    finally
      t.Unlock;
    end;

  end;
end;

procedure TControlThread.GetNextWaitTime(var NextWaitTime: Integer; ExtraDelta: Integer);
var
  t : TThreadedTimer;
  i : integer;
begin
  NextWaitTime := -1;
  {$ifdef DebugTimer}
  DebugTimer('[ControlThread] ExtraDelta = '+IntToStR(ExtraDelta));
  DebugTimer('[ControlThread] timers count = '+IntToStr(fTimers.Count));
  {$endif}
  for i := 0 to fTimers.Count - 1 do begin
    t := TThreadedTimer(fTimers[i]);
    t.Lock;
    try
      if not t.fEnabled then Continue;
      {$ifdef DebugTimer}
      DebugTimer('[ControlThread] time '+ IntToStR(Integer(t)));
      {$endif}
      dec(t.fLeftVal, ExtraDelta);
      if t.fLeftVal < 0 then
        NextWaitTime := 0
      else if (NextWaitTime < 0) or (t.fLeftVal < NextWaitTime) then
        NextWaitTime := t.fLeftVal;
    finally
      t.Unlock;
    end;
  end;
end;

procedure TControlThread.Lock;
begin
  fLock.Enter;
end;

procedure TControlThread.Unlock;
begin
  fLock.Leave;
end;

procedure TControlThread.StopThread;
begin
  fDoStop := true;
  fEvent.SetEvent;
  WaitFor;
end;

procedure TControlThread.ReleaseTimers;
var
  i : integer;
begin
  Lock;
  try
    for i := 0 to fTimers.Count - 1 do TThreadedTimer(FTimers[i]).ReleaseControlThread;
  finally
    Unlock;
  end;
end;

procedure TControlThread.Update;
begin
  fEvent.SetEvent;
end;

constructor TControlThread.Create(ASuspended: Boolean);
begin
  inherited Create(ASuspended);
  fLock := TCriticalSection.Create;
  fEvent := TEvent.Create(nil, true, false, '');
  FTimers := TList.Create;
end;

destructor TControlThread.Destroy;
begin
  StopThread;
  ReleaseTimers;
  FTimers.Free;
  fEvent.Free;
  fLock.Free;
  inherited Destroy;
end;

procedure TControlThread.Add(Timer: TThreadedTimer);
begin
  if not Assigned(Timer) then Exit;
  Lock;
  try
    fTimers.Add(Timer);
    {$ifdef DebugTimer}DebugTimer('[ControlThread] adding timer and upating');{$endif}
    fEvent.SetEvent;
  finally
    Unlock;
  end;
end;

procedure TControlThread.Remove(Timer: TThreadedTimer);
var
  i : integer;
begin
  if not Assigned(Timer) then Exit;
  Lock;
  try
    i := fTimers.IndexOf(Timer);
    if i < 0 then Exit;
    fTimers.Delete(i);
    fEvent.SetEvent;
  finally
    Unlock;
  end;

end;

{ TThreadedTimer }

constructor TThreadedTimer.Create;
begin
  inherited;
  fLock := TCriticalSection.Create;
  fInterval := 1000;
end;

destructor TThreadedTimer.Destroy;
var
  th : TControlThread;
begin
  Enabled := false;
  if fAdded then begin
    th := ControlThread;
    if Assigned(th) then th.Remove(Self);
  end;
  fLock.Free;
  inherited Destroy;
end;

procedure TThreadedTimer.Lock;
begin
  fLock.Enter;
end;

procedure TThreadedTimer.Unlock;
begin
  fLock.Leave;
end;

procedure TThreadedTimer.DoTimer;
begin
  try
    // must be locked
    {$ifdef DebugTimer}DebugTimer('[TThreadedTimer] signal');{$endif}
    if Assigned(fOnTimer) then fOnTimer(Self);
  except
  end;
end;

function TThreadedTimer.GetInterval: Integer;
begin
  Lock;
  try
    Result := fInterval;
  finally
    Unlock;
  end;
end;

procedure TThreadedTimer.SetInterval(const AValue: Integer);
begin
  Lock;
  try
    fInterval := AValue;
  finally
    Unlock;
  end;
end;

function TThreadedTimer.GetEnabled: Boolean;
begin
  Lock;
  try
    Result := fEnabled;
  finally
    Unlock;
  end;
end;

procedure TThreadedTimer.SetEnabled(AEnabled: Boolean);
var
  th : TControlThread;
begin
  Lock;
  try
    if AEnabled and (fInterval <= 0) then AEnabled := false;
    if fEnabled = AEnabled then Exit;

    fEnabled := AEnabled;
    if fEnabled then fLeftVal := fInterval;

    if not fAdded then begin
      th := ControlThread;
      if Assigned(th) then th.Add(Self);
      fAdded := true;
    end else if fAdded then
      ControlThread.Update;
  finally
    Unlock;
  end;
end;

function TThreadedTimer.GetOnTimer: TTimerEvent;
begin
  lock;
  try
    Result := fOnTimer;
  finally
    unlock;
  end;
end;

procedure TThreadedTimer.SetOnTimer(AEvent: TTimerEvent);
begin
  lock;
  try
    fOnTimer := AEvent;
  finally
    Unlock;
  end;
end;

procedure TThreadedTimer.ReleaseControlThread;
begin
  Lock;
  try
    fAdded := false;
  finally
    Unlock;
  end;
end;


procedure ReleaseTimersThread;
var
  th  : TControlThread;
begin
  if Assigned(GlobalTimerThread) then begin
    Releasing := true;
    th :=  GlobalTimerThread;
    GlobalTimerThread := nil;
    th.Free;
  end;
end;

initialization

finalization
  ReleaseTimersThread;

end.

