{
  @exclude

  @html(<b>)
  StreamSec SSL Plug-in
  @html(</b>)
  - Copyright (c) Henrick Hellstrom & Danijel Tkalcec
  @html(<br><br>)

  This unit defines Plug-ins for using StreamSec 2.1 SSL components 
  with rtcHttpClient and rtcHttpServer for SSL encryption/decryption.
}
unit rtcSSecPlugin;

interface

uses
  Classes, SysUtils,

  // RTC SDK
  rtcPlugins,

  // StreamSec Tools 2.1
  SecUtils,
  ASN1,
  Tls,
  TlsClass,
  TlsConst,
  StreamSecII,
  TlsInternalServer;

type      
  {$IFNDEF UNICODE}
  TOctetStringStream = TStringStream;
  {$ENDIF}

  // Basic StreamSec Plug-in implementation
  TRtcSSecPlugin = class(TRtcCryptPlugin)
  private
    FTLSServer: TCustomTLSInternalServer;
    FBuffer: string;
    procedure SetTLSServer(const Value: TCustomTLSInternalServer);
  protected
    procedure CheckTLSServer; virtual; abstract;
    procedure Notification(AComponent: TComponent; Operation: TOperation);
      override;
  public
    { Called before we do a graceful disconnect, in case some data has to be sent out. }
    procedure BeforeDisconnect(var ConnCryptObj:TObject; ID:TRtcConnID; var OutData:AnsiString); override;

    { Called after a connection was closed. }
    procedure AfterDisconnect(var ConnCryptObj:TObject; ID:TRtcConnID); override;

    { Called when data arrived.
      InData = data received from recipient (decode that!)
      OutData = data prepared by "decoder" for sending back to recipient (encoded data)
      OutPlainText = decrypted input data (for use by application) }
    procedure DataReceived(var ConnCryptObj:TObject; ID:TRtcConnID; const InData:AnsiString; var OutData:AnsiString; var OutPlainText:AnsiString); override;

    { Called when data needs to be sent.
      InData = application data which we want to be encoded for sending
      OutData = encoded data which should be sent out }
    procedure DataToSend(var ConnCryptObj:TObject; ID:TRtcConnID; const InData:AnsiString; var OutData:AnsiString); override;
  published
    property TLSServer: TCustomTLSInternalServer read FTLSServer write SetTLSServer;
    end;

  { StreamSec Plugin for Server-side SSL encryption/decryption }
  TRtcSSecServerPlugin = class(TRtcSSecPlugin)
  protected
    procedure CheckTLSServer; override;
  public
    { Called after a new connection was established.
      OutData = data which has to be sent out immediately. }
    procedure AfterConnect(var ConnCryptObj:TObject; ID:TRtcConnID; var OutData:AnsiString); override;

  end;

  { StreamSec Plugin for client-side SSL encryption/decryption }
  TRtcSSecClientPlugin = class(TRtcSSecPlugin)
  protected
    procedure CheckTLSServer; override;
  public
    { Called after a new connection was established.
      OutData = data which has to be sent out immediately. }
    procedure AfterConnect(var ConnCryptObj:TObject; ID:TRtcConnID; var OutData:AnsiString); override;
  end;

  TRtcSSecConnCryptObject = class
  private
    fTLSLayer: TCustomTLS_ContentLayer;
    function GetSecurityParams: PTLSSecurityParams;
    function GetBulkCipherAlgorithm: TBulkCipherAlgorithm;
    function GetCipherType: TCipherType;
    function GetClientCert: TASN1Struct;
    function GetCompressionAlgorithm: TCompressionMethod;
    function GetEncrypted: Boolean;
    function GetEntity: TConnectionEnd;
    function GetIsExportable: Boolean;
    function GetKeyMaterialLength: Word;
    function GetKeySize: Word;
    function GetMacAlgorithm: THashAlgorithm;
    function GetServerCert: TASN1Struct;
  public
    property Encrypted: Boolean read GetEncrypted;
    property ServerCert: TASN1Struct read GetServerCert;
    property ClientCert: TASN1Struct read GetClientCert;
    property Entity: TConnectionEnd read GetEntity;
    property BulkCipherAlgorithm: TBulkCipherAlgorithm read GetBulkCipherAlgorithm;
    property CipherType: TCipherType read GetCipherType;
    property KeySize: Word read GetKeySize;
    property KeyMaterialLength: Word read GetKeyMaterialLength;
    property IsExportable: Boolean read GetIsExportable;
    property MacAlgorithm: THashAlgorithm read GetMacAlgorithm;
    property CompressionAlgorithm: TCompressionMethod read GetCompressionAlgorithm;
  end;

implementation

function RtcConnIDToPointer(ID: TRtcConnID): Pointer;
begin
  Result := Pointer(ID);
end;

{ TRtcSSecPlugin }

procedure TRtcSSecPlugin.AfterDisconnect(var ConnCryptObj: TObject;
  ID: TRtcConnID);
var
  lDummy: AnsiString;
begin
  BeforeDisconnect(ConnCryptObj,ID,lDummy);
end;

procedure TRtcSSecPlugin.BeforeDisconnect(var ConnCryptObj: TObject;
  ID: TRtcConnID; var OutData: AnsiString);
var
  lObj: TRtcSSecConnCryptObject;
  lRes: TOctetStringStream;
begin
  if Assigned(ConnCryptObj) then begin
    lObj := ConnCryptObj as TRtcSSecConnCryptObject;
    if Assigned(lObj.fTLSLayer) then begin
      lRes := TOctetStringStream.Create('');
      try
        lObj.fTLSLayer.Close(lRes);
        OutData := lRes.DataString;
      finally
        lRes.Free;
      end;
      lObj.fTLSLayer.Release;
      FreeAndNil(ConnCryptObj);
    end;
  end;
end;

procedure TRtcSSecPlugin.DataReceived(var ConnCryptObj: TObject; ID: TRtcConnID;
  const InData: AnsiString; var OutData, OutPlainText: AnsiString);
var
  lObj: TRtcSSecConnCryptObject;
  lIn, lOut, lPT: TOctetStringStream;
  Client: TCustomTLS_ContentLayer;
begin
  if Assigned(ConnCryptObj) then begin
    lObj := ConnCryptObj as TRtcSSecConnCryptObject;
    Client := lObj.fTLSLayer;
    lIn := TOctetStringStream.Create(InData);
    lOut := TOctetStringStream.Create('');
    lPT := TOctetStringStream.Create('');
    try
      if not Assigned(Client) then begin
        if lIn.Size > 0 then begin
          Assert(Assigned(FTLSServer),'TLSServer must be assigned');
          Client := FTLSServer.TLSAddServerSession;
          Client.Unmanaged := True;
          Client.UserData := RtcConnIDToPointer(ID);
          lObj.fTLSLayer := Client;
          Client.Accept(lIn,lOut);
        end;
      end;

      while lIn.Size>lIn.Position do
        Client.DecodeData(lIn,lPT,lOut);

      if (FBuffer <> '') and Assigned(Client) and Client.Encrypted then begin
        lIn.Size := 0;
        lIn.WriteString(FBuffer);
        lIn.Position := 0;
        FBuffer := '';
        while lIn.Size>lIn.Position do
          Client.EncodeData(lIn,lOut);
      end;
      OutData := lOut.DataString;
      OutPlainText := lPT.DataString;
    finally
      lIn.Free;
      lOut.Free;
      lPT.Free;
    end;
  end;
end;

procedure TRtcSSecPlugin.DataToSend(var ConnCryptObj: TObject; ID: TRtcConnID;
  const InData: AnsiString; var OutData: AnsiString);
var
  lObj: TRtcSSecConnCryptObject;
  lIn, lOut: TOctetStringStream;
  Client: TCustomTLS_ContentLayer;
begin
  if Assigned(ConnCryptObj) then begin
    lObj := ConnCryptObj as TRtcSSecConnCryptObject;
    Client := lObj.fTLSLayer;
    if not Assigned(Client) then
      raise Exception.Create('Connection is closed');
    if Client.Encrypted then begin
      lIn := TStringStream.Create(FBuffer + InData);
      lOut := TOctetStringStream.Create('');
      try
        FBuffer := '';
        while lIn.Size > lIn.Position do
          Client.EncodeData(lIn,lOut);

        OutData := lOut.DataString;
      finally
        lIn.Free;
        lOut.Free;
      end;
    end else begin
      FBuffer := FBuffer + InData;
      OutData := '';
    end;
  end;
end;

procedure TRtcSSecPlugin.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if Operation = opRemove then begin
    if AComponent = FTLSServer then
      FTLSServer := nil;
  end;
end;

procedure TRtcSSecPlugin.SetTLSServer(const Value: TCustomTLSInternalServer);
begin
  if Assigned(FTLSServer) then
    FTLSServer.RemoveFreeNotification(Self);
  FTLSServer := Value;
  CheckTLSServer;
  if Assigned(FTLSServer) then
    FTLSServer.FreeNotification(Self);
end;

{ TRtcSSecServerPlugin }

type
  TTLSServerHack = class(TCustomTLSInternalServer);

procedure TRtcSSecServerPlugin.AfterConnect(var ConnCryptObj: TObject;
  ID: TRtcConnID; var OutData: AnsiString);
begin
  Assert(Assigned(FTLSServer),'TLSServer must be assigned');
  OutData := '';
  ConnCryptObj := TRtcSSecConnCryptObject.Create;
end;

procedure TRtcSSecServerPlugin.CheckTLSServer;
begin
  if Assigned(FTLSServer) then
    if FTLSServer.ClientOrServer = cosClientSide then begin
      FTLSServer := nil;
      raise Exception.Create('ClientOrServer must be set to cosServerSide');
    end;
end;

{ TRtcSSecClientPlugin }

procedure TRtcSSecClientPlugin.AfterConnect(var ConnCryptObj: TObject;
  ID: TRtcConnID; var OutData: AnsiString);
var
  lRes: TOctetStringStream;
  Client: TCustomTLS_ContentLayer;
begin
  Assert(Assigned(FTLSServer),'TLSServer must be assigned');
  ConnCryptObj := TRtcSSecConnCryptObject.Create;
  lRes := TOctetStringStream.Create('');
  try
    Client := FTLSServer.TLSAddClientSession;
    Client.Unmanaged := True;
    Client.UserData := RtcConnIDToPointer(ID);
    TRtcSSecConnCryptObject(ConnCryptObj).fTLSLayer := Client;
    Client.Connect(lRes);
    OutData := lRes.DataString;
  finally
    lRes.Free;
  end;
end;

procedure TRtcSSecClientPlugin.CheckTLSServer;
begin
  if Assigned(FTLSServer) then
    if FTLSServer.ClientOrServer = cosServerSide then begin
      FTLSServer := nil;
      raise Exception.Create('ClientOrServer must be set to cosClientSide');
    end;
end;

{ TRtcSSecConnCryptObject }

function TRtcSSecConnCryptObject.GetBulkCipherAlgorithm: TBulkCipherAlgorithm;
begin
  if GetEncrypted then
    Result := GetSecurityParams.bulk_cipher_algorithm
  else
    Result := bcaNull;
end;

function TRtcSSecConnCryptObject.GetCipherType: TCipherType;
begin
  if GetEncrypted then
    Result := GetSecurityParams.cipher_type
  else
    Result := ctStream;
end;

function TRtcSSecConnCryptObject.GetClientCert: TASN1Struct;
var
  lCert: PASN1Struct;
begin
  Result := nil;
  if Assigned(fTLSLayer) then begin
    lCert := fTLSLayer.ClientCertificate;
    if Assigned(lCert) then
      Result := lCert^;
  end;
end;

function TRtcSSecConnCryptObject.GetCompressionAlgorithm: TCompressionMethod;
begin
  if GetEncrypted then
    Result := GetSecurityParams.compression_algorithm
  else
    Result := cmNull;
end;

function TRtcSSecConnCryptObject.GetEncrypted: Boolean;
begin
  Result := False;
  if Assigned(fTLSLayer) then
    if fTLSLayer.Active then
      if fTLSLayer.Encrypted then
        Result := True;
end;

function TRtcSSecConnCryptObject.GetEntity: TConnectionEnd;
begin
  if GetEncrypted then
    Result := GetSecurityParams.entity
  else
    Result := ceServer;
end;

function TRtcSSecConnCryptObject.GetIsExportable: Boolean;
begin
  if GetEncrypted then
    Result := GetSecurityParams.is_exportable
  else
    Result := True;
end;

function TRtcSSecConnCryptObject.GetKeyMaterialLength: Word;
begin
  if GetEncrypted then
    Result := GetSecurityParams.key_material_length
  else
    Result := 0;
end;

function TRtcSSecConnCryptObject.GetKeySize: Word;
begin
  if GetEncrypted then
    Result := GetSecurityParams.key_size
  else
    Result := 0;
end;

function TRtcSSecConnCryptObject.GetMacAlgorithm: THashAlgorithm;
begin
  if GetEncrypted then
    Result := GetSecurityParams.mac_algorithm
  else
    Result := haNull;
end;

function TRtcSSecConnCryptObject.GetSecurityParams: PTLSSecurityParams;
begin
  if Assigned(fTLSLayer) then
    Result := fTLSLayer.Context
  else
    Result := nil;
end;

function TRtcSSecConnCryptObject.GetServerCert: TASN1Struct;
var
  lCert: PASN1Struct;
begin
  Result := nil;
  if Assigned(fTLSLayer) then begin
    lCert := fTLSLayer.ServerCertificate;
    if Assigned(lCert) then
      Result := lCert^;
  end;
end;

end.


