{ 
  RealThinClient W/A Framework: Caching functions
  Copyright (c) Danijel Tkalcec

  @exclude
}
unit rtcCache;

{$include rtcDefs.inc}

interface

uses
  SysUtils, rtcTrashcan,
  rtcInfo, rtcMD5, rtcCrypt, rtcSyncObjs;

var
  { Sub-Folder inside AppFileName's directory where all CACHE files will be stored.
    If you want CACHE files to be created in the same folder as AppFile (EXE/DLL),
    set CACHE_FOLDER to an empty string before you create the first TRtcApplication instance. }
  CACHE_FOLDER:string='CACHE';

  { Full path to the CACHE folder. If you leave this variable empty (default),
    it will be initialized automatically by using the AppFileName and CACHE_FOLDER
    variables, immediately before the first CACHE file needs to be written.
    If you want your CACHE files written to a specific folder by using full path,
    you can do it by setting this variable before creating the first TRtcApplication instance.
    RTC_CACHE_FOLDER should ALWAYS end with '\' }
  RTC_CACHE_FOLDER:string='';

  { Secure Key to be used for encrypting cache files. }
  RTC_CACHE_KEY:AnsiString='RTC SDK 3.0';

  { Minimim number of bytes a data chunk should have to make caching feasible.
    For chunks of a smaller size, caching will NOT be used even if it was requested. }
  RTC_CACHE_MINSIZE:integer=128;

  { Hash function used by RTC Cache to generate a unique ID from data.
    By default, this points to the MD5Code function from the rtcMD5 unit,
    which generates a hash code using MD5 and returns the value in custom
    RTC 5-bit encoding (consumes 26 characters instead of the standard 32).

    While the MD5 technique is very fast and will be adequate for most applications,
    there are reasons to consider using functions with a stronger encryption:

    1. There are techniques for generating MD5 collisions,
       meaning two different inputs that hash to the same value.

    2. Under some circumstances, an attacker might use this flaw for attacks.
       Such attacks are /only/ possible if the attacker gets to choose the file initially,
       i.e. the attacker has to be the one to create and upload the file

       An attack that exploits MD5 collisions also has to have a mean of providing
       two different versions of a chunk. For instance, systems with two related servers
       might be vulnerable. Typically, attacks of this kind targets file sharing networks.

    If you require higher security than provided by the standard MD5 algorithms,
    you should use a commercial security package from StreamSec. }
  rtcCache_Hash:function(const Data: AnsiString): String;

// Write "Data" to FileName in local cache
function rtcCache_Write(var Data:AnsiString; const FileName:string):boolean;

// Read data from FileName in local cache
function rtcCache_Read(const FileName:string):AnsiString;

implementation

var
  fCacheCS:TRtcCritSec;

procedure NeedCacheFolder;
  begin
  if RTC_CACHE_FOLDER='' then
    begin
    RTC_CACHE_FOLDER:=ExtractFilePath(AppFileName);
    if Copy(RTC_CACHE_FOLDER,length(RTC_CACHE_FOLDER),1)<>'\' then
      RTC_CACHE_FOLDER:=RTC_CACHE_FOLDER+'\';

    if CACHE_FOLDER<>'' then
      begin
      RTC_CACHE_FOLDER:=RTC_CACHE_FOLDER+CACHE_FOLDER;
      if Copy(RTC_CACHE_FOLDER,length(RTC_CACHE_FOLDER),1)<>'\' then
        RTC_CACHE_FOLDER:=RTC_CACHE_FOLDER+'\';
      end;

    if not DirectoryExists(RTC_CACHE_FOLDER) then
      if not CreateDir(RTC_CACHE_FOLDER) then
        begin
        RTC_CACHE_FOLDER:=GetTempDirectory;
        if Copy(RTC_CACHE_FOLDER,length(RTC_CACHE_FOLDER),1)<>'\' then
          RTC_CACHE_FOLDER:=RTC_CACHE_FOLDER+'\';

        RTC_CACHE_FOLDER:=RTC_CACHE_FOLDER+CACHE_FOLDER;
        if not DirectoryExists(RTC_CACHE_FOLDER) then
          CreateDir(RTC_CACHE_FOLDER);

        if Copy(RTC_CACHE_FOLDER,length(RTC_CACHE_FOLDER),1)<>'\' then
          RTC_CACHE_FOLDER:=RTC_CACHE_FOLDER+'\';
        end;
    end;
  end;

function rtcCache_Write(var Data:AnsiString; const FileName:string):boolean;
  begin
  NeedCacheFolder;
  fCacheCS.Enter;
  try
    if not File_Exists(RTC_CACHE_FOLDER+FileName) then
      begin
      Crypt(Data,RTC_CACHE_KEY);
      Result:=Write_File(RTC_CACHE_FOLDER+FileName, Data);
      end
    else
      Result:=True;
  finally
    fCacheCS.Leave;
    end;
  end;

function rtcCache_Read(const FileName:string):AnsiString;
  begin
  NeedCacheFolder;
  Result:=Read_File(RTC_CACHE_FOLDER+FileName);
  DeCrypt(Result,RTC_CACHE_KEY);
  end;

initialization
fCacheCS:=TRtcCritSec.Create;
rtcCache_Hash:=@MD5Code;
finalization
Garbage(fCacheCS);
end.
